/* $Id: adjobj.h,v 1.11 2007/12/05 15:03:46 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_ADJOBJ_INCLUDED
#define DEP_ADJOBJ_INCLUDED

#include "depdef.h"
#include "upari.h"
#include "lspars.h"
#include "drvsub.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{

/** Adjustment object class.
 *  Base class for all objects treated in the least squares adjustment
 *  e.g. exterior orientation, object points and additional parameters.
 *  @author Dirk Stallmann
 */
class AdjObject
{
	friend void corr_table(ostream& os, const Matrix<double>& qxx,
		const AdjObject& ao1, const AdjObject& ao2, double lim);
	friend void corr_matrix(ostream& os, const Matrix<double>& qxx,
		const AdjObject& ao1, const AdjObject& ao2, double lim);
	friend void covar_matrix(ostream& os, const DMatrix& qxx,
		const AdjObject& ao1, const AdjObject& ao2, double sigma0);

public:
	/** constructor.
	 *  Use the argument only when the number of unknown parameters
	 *  is known. Otherwise use the read-method to define
	 *  the number of parameters.
	 *
	 *  Defaults: number of parameters if unknown, no approximations exist.
	 *
	 *  @param sz number of unknown parameters
	 */
	AdjObject(size_t sz=0);

	/// virtual destructor
	virtual ~AdjObject();

	/**@name set and get adjustment object ID */
	//@{
		/// set object ID
		void set_id(const string& s) { _id = s; }

		/// get object ID 
		string get_id() const { return _id; }
	//@}

	/**@name set and check inuse flag */
	//@{
		/// set inuse flag
		void inuse(bool flg) { _inuse = flg; }

		/// test inuse flag
		bool is_inuse() const { return _inuse == true; }

		/// test inuse flag
		bool is_unused() const { return _inuse == false; }
	//@}

	/**@name set and check approximation flag
	 *  Flag to determine (all) parameters are approximate values
	 *  or not.
	 *  Please note: Parameter values are initilized with zero.
	 *  And zero might be (e.g. additional parameters) a valid
	 *  approximate value. Typically approximate values are read
	 *  from file using the method read(). The flag will be set in
	 *  this function to determine the read parameters are approximate
	 *  values or not. If not an special methods will be used to
	 *  calculate good approximations.
	 *  The default state of
	 *  adjustment objects is "no approximations exist".
	 *  During the adjustment the flag is not changed because
	 *  adjusted (updated) parameters are used as new approximate values
	 *  in the next iteration step.
	 */
	//@{
		/// set approximation flag
		void set_approx(bool flag) { _approx_flag = flag; }

		/// check approximation flag
		bool is_approx() const { return _approx_flag == true; }
	//@}
	
	/**@name set and get index to normal equations functions */
	//@{
		/// set index
		void set_index(int k) { _index = k; }

		/// get index
		int get_index() const { return _index; }
	//@}

	/// get number of unknown parameters
	virtual size_t num_unk() const { return _size; }

	/// get number of constraints (default 0)
	virtual size_t num_con() const { return 0; }

	/**@name functions to get unknown parameter by index.
	 *  Only available after setting the unknown parameter
	 *  (upar) information using function set\_upari.
	 *  Without check on vaild index and existence of
	 *  upar-information!
	 *
	 *  @param i index of unknown parameter
	 */
	//@{
		/// get inital value
		double ival(size_t i) const { return _upari[i].ival; }
	
		/// get unknown parameter
		double par(size_t i) const { return *(_upari[i].par); }
	
		/// get std. dev. of parameter
		double sdev(size_t i) const { return *(_upari[i].sdev); }
	
		/// get parameter status
		u_state state(size_t i) const { return _upari[i].state; }

		/// get string of parameter status
		string state_str(size_t i) const;

		/// get weight
		double pii(size_t i) const { return _upari[i].pii; }

		/// calculate change: change := adjusted - initial
		double change(size_t i) const
		{
			return ((par(i) - ival(i)));
		}
	
		/// calculate ratio: ratio := change / std.dev.
		double ratio(size_t i) const
		{
			return (state(i) == eli) ? (0.) : (fabs((change(i)) / sdev(i)));
		}
		
		/// get name of unknown parameter
		string uname_idx(size_t i) const
		{
			return uname(i - _index);
		}
	//@}

	/// count number of unknowns by state
	size_t num(u_state state) const;

	/// print unknown parameters
	void print_info(ostream& os) const;

	/// stochastic weighting of the unknowns to the normal matrix
	void stoch_weighting(Matrix<double>& atpa,
		Vector<double>& atpl, double& ltpl) const;

	/** add weighted constraints.
	 *  Simply do nothing because per default no constraints exist.
	 */
	virtual void add_constraints(Matrix<double>& atpa,
		Vector<double>& atpl, double& ltpl) const {}

	/** update unknown parameters
	 *  @return number of updated unknowns
	 */
	size_t update_unknowns(const Vector<double>& dx);

	/** update std. devs. of unknowns
	 *  @return number of updated unknowns
	 */
	size_t update_sdev(double sigma0, const Matrix<double>& qxx);

	/**@name virtual functions
	 *  Specific functions for each derived class.
	 */
	//@{
		/// get adjustment object class name (long description)
		virtual string class_str() const;
		
		/// get (sub-)type name (long description)
		virtual string type_str() const;

		/// get (sub-)type name (keyword)
		virtual string key_str() const;

		/// get name of unknown parameter i
		virtual string uname(size_t i) const;

		/// set unknown parameter information (upari)
		virtual bool set_upari() = 0;

		/** print changes of unknowns
		 *  This method is used to output the parameters
		 *  during the iteration steps.
		 *  The following values are printed:
		 *  param. name, state, approx. value, change, adjusted value.
		 *  UPARI required!
		 */
		virtual void print_changes(ostream& os, const Vector<double>& dx) const;

		/** print unknown parameters
		 *  This method is used to output the parameters
		 *  after the iteration steps.
		 *  The following values are printed:
		 *  param. name, state, initial value, adjusted value, change, std.dev.
		 *  A test on significance is applied when the test flag
		 *  is set.
		 *  UPARI required!
		 *
		 *  @param os output stream
		 */
		virtual void print_unknowns(ostream& os) const;

		/** print precision of unknowns
		 *  This method is used to output the parameters
		 *  after calculation of precision values.
		 *  Also a test on significance is applied.
		 *  The following values are printed:
		 *  param. name, state, initial value, adjusted value, change,
		 *  std.dev., ratio.
		 *  UPARI required!
		 */
		virtual void print_precision(ostream& os) const;

		// Change this function later!
		// Default: Check against the general convergence limit.
		// For a special check or when nothing to check overwrite
		// this function in the derived class!

		/** check on convergence.
		 *  Default function: Nothing to check, ignore the parameters.
		 *  Simply return -1.
		 * @param dx solution vector
		 * @return -1: , >= 0:
		 */
		virtual int check_convergence(const Vector<double>& dx) const
		{
			return -1;
		}

		/// sum of absolute changes
		virtual void sum_changes(const Vector<double>& dx,
			double& sum, size_t& num) const;
	//@}

	/**@name definitions of standard deviations
	 * @param s standard deviation
	 */
	//@{
		/// get string of parameter status from std. dev.
		string state_str(double s) const;

		/// check std. devs. is valid
		bool is_valid(double s) const { return (s > 0.); }
	
		/// check on real std. dev.
		bool is_rsdev(double s) const {
			return (s >= lspars._smin && s <= lspars._smax); }

		/// check on observed parameter
		bool is_obs(double s) const {
			return (s >= lspars._smin && s <= lspars._smax); }

		/// check on eliminated parameter
		bool is_eli(double s) const { return (s < lspars._smin); }

		/// check on unknown parameter
		bool is_unk(double s) const { return (s > lspars._smax); }
	//@}
	
	/** check on significance
	 * @param t test value
	 * @return t-test result
	 */
	bool chk_sign(double t) const { return (t > lspars._t_quantil); }
	
	// print standard deviation
	void print_sdev(ostream& os, const Form& fmt, double s, double sc=1.) const;

protected:
	/// object ID
	string		_id;

	/// inuse flag (maybe replace this by object status!!!)
	bool		_inuse;

	/// approximation flag
	bool		_approx_flag;

	/// index to normal equations
	int			_index;

	/// number of unknown parameters
	size_t		_size;

	/** unknown parameter information (upari).
	 *  Information is set by the constructor and/or the read-method
	 */
	Upari*		_upari;

protected:
	/// create table
	bool create();

	/// set table entry
	void set(size_t i, double* par, double* sdev);
	
public:
	/** definition of least-squares parameters (effects all adjustment
	 * objects).
	 */
	static LSParams		lspars;
};


// non-member functions ...

/// print correlations in table form
void corr_table(ostream& os, const Matrix<double>& qxx,
	const AdjObject& ao1, const AdjObject& ao2, double lim);

/** print correlations in table form.
 *  Attention: Correlations between parameters of adjustment object
 *  appears twice in the output!
 */
void corr_matrix(ostream& os, const Matrix<double>& qxx,
	const AdjObject& ao1, const AdjObject& ao2, double lim);

/// print matrix of covariance
void covar_matrix(ostream& os, const DMatrix& qxx,
	const AdjObject& ao1, const AdjObject& ao2, double sigma0);

} // namespace DEP

#endif
