/**********************************************************************/
//
// FILE:    adscam.cpp
//          ADS camera class functions.
// CREATED: 15 Feb 04  (ds)
// CHANGES: see include file
// REFS:    dito 
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <fstream>
#include "adscam.h"

namespace DEP
{

// enum, keywords

ADSCamera::map_t ADSCamera::_map[] =
{
	{ UNKNOWN, "UNKNOWN" },
	{ ADS_CALIBRATION_FILE, "ADS_CALIBRATION_FILE" },
	{ CALIBRATION_SOURCE, "CALIBRATION_SOURCE" },
	{ CAMERA_NAME, "CAMERA_NAME" },
	{ SENSOR_LINE, "SENSOR_LINE" },
	{ CALIBRATION_DATE, "CALIBRATION_DATE" },
	{ IMU_MISALIGNMENT, "IMU_MISALIGNMENT" },
	{ FOCAL_LENGTH_MM, "FOCAL_LENGTH_MM" },
	{ NUM_PIXELS, "NUM_PIXELS" },
	{ PIXEL_SIZE, "PIXEL_SIZE" },
	{ START_XY, "START_XY" },
	{ END_XY, "END_XY" },
	{ PAV_Z_OFFSET, "PAV_Z_OFFSET" },
	{ RADIOMETRIC_GAIN, "RADIOMETRIC_GAIN" },
	{ IRRADIANCE_GAIN, "IRRADIANCE_GAIN" },
	{ LEFTSIDE_GAIN, "LEFTSIDE_GAIN" },
	{ END_OF_MAP, 0 }
};


// constructor

ADSCamera::ADSCamera() : PBCameraCalib(),
	_calib_version(),
	_calib_source(),
	_camera_name(),
	_sensor_line(),
	_calib_date()
{
	// empty
}


// get string of camera type

string ADSCamera::type_str() const
{
	ostringstream oss;
	oss << "ADS40";
	return oss.str();
}


// get keyword name

string ADSCamera::key_str() const
{
	ostringstream oss;
	oss << "ads40-camera";
	return oss.str();
}


// read camera parameter

bool ADSCamera::read(istream& is)
{	
	size_t i = 0;
	string line;
	size_t num_pix = 0;
	bool eod_found = false; // end of calibration data found

	while (getline(is, line))
		{
		//cout << "line: " << line << endl;

		string keyword;
		istringstream iss(line);
		iss >> keyword;
		//cout << "keyword: " << keyword << endl;

		ekey key = code(keyword.c_str());
		switch (key)
			{
			case ADS_CALIBRATION_FILE:
				iss >> _calib_version;
				break;

			case CALIBRATION_SOURCE:
				_calib_source = strip(line);
				break;

			case CAMERA_NAME:
				_camera_name = strip(line);
				break;

			case SENSOR_LINE:
				_sensor_line = strip(line);
				break;

			case CALIBRATION_DATE:
				_calib_date = strip(line);
				break;
/*
			//case IMU_MISALIGNMENT:
				//iss >> _phi >> _sdev_phi;
				//iss >> _omega >> _sdev_omega;
				//iss >> _kappa >> _sdev_kappa;

				degree or radian???
				_phi /= rho_deg;
				_sdev_phi /= rho_deg;
				_omega /= rho_deg;
				_sdev_omega /= rho_deg;
				_kappa /= rho_deg;
				_sdev_kappa /= rho_deg;

				//break;
*/
			case FOCAL_LENGTH_MM:
				iss >> _princ_dist;
				_princ_dist *= 1e-3; // convert from mm to m
				break;

			case NUM_PIXELS:
				iss >> _nc;
				break;

			case PIXEL_SIZE:
				iss >> _sc;
				_sc *= 1e-3; // convert from mm/pixel to m/pixel
				_sr = _sc;
				break;

			case START_XY:
				if (_xcal != 0) delete[] _xcal;
				if (_ycal != 0) delete[] _ycal;

				_xcal = new double[_nc];
				_ycal = new double[_nc];

				if (_xcal == 0 || _ycal == 0)
					{
					cout << "ERROR: Memory allocation for "
						"pixel data failed.\n";
					return false;
					}

				for (i = 0; i < _nc; i++)
					{
					//getline(is, line); cout << line << endl;
					//iss.seekg(ios_base::beg);
					is >> _xcal[i] >> _ycal[i];
					_xcal[i] *= 1e-3; // convert from mm to m
					_ycal[i] *= 1e-3;
					num_pix++;
					}
					getline(is, line); // go to behind EOL
				break;

			case END_XY:
				eod_found = true;

				// check number of pixels
				if (num_pix != _nc)
					{
					cout << "ERROR: Check number of pixels failed.\n";
					return false;
					}

				break;

			case IMU_MISALIGNMENT:
			case PAV_Z_OFFSET:
			case RADIOMETRIC_GAIN:
			case IRRADIANCE_GAIN:
			case LEFTSIDE_GAIN:
				//cout << "WARNING: Keyword known but values ignored: "
				//	<< keyword << endl;
				break;

			case UNKNOWN:
			default:
				cout << "WARNING: Keyword unknown: "
					<< keyword << endl;
				break;
			}
		}

	 if (!eod_found)
	 	{
		cout << "ERROR: End of calibration data (END_XY) not found.\n";
		return false;
		}

	// calc. ICS location
	_c0 = (_nc - 1) / 2.;
	//_r0 = (_nr - 1) / 2.;

	// calc. image format
	_size_x = _nr * _sr;
	_size_y = _nc * _sc;

	return true;
}


// write camera parameter

void ADSCamera::write(ostream& os) const
{
	assert(false);
}


// print camera parameter

void ADSCamera::print(ostream& os) const
{
	PBCameraCalib::print(os);

	os << "Calibration file version :  " << _calib_version << endl;
	os << "Calibration source       :  " << _calib_source << endl;
	os << "Camera name              :  " << _camera_name << endl;
	os << "Sensor line              :  " << _sensor_line << endl;
	os << "Calibration date         :  " << _calib_date << endl;
/*
	Form int5; int5.width(5);
	Form fix4(4); fix4.fixed().width(10);
	os << "Calibration pixel position" << endl;
	int k = 0;
	os << "  pixel [" << int5(k) << "] "
		<< "  x = " << fix4(_xcal[k] * 1000.) << " mm"
		<< "  y = " << fix4(_ycal[k] * 1000.) << " mm" << endl;
	k = _nc - 1;
	os << "  pixel [" << int5(k) << "] "
		<< "  x = " << fix4(_xcal[k] * 1000.) << " mm"
		<< "  y = " << fix4(_ycal[k] * 1000.) << " mm" << endl;
*/
	os << endl;
}


// get code

ADSCamera::ekey ADSCamera::code(const char* name)
{
	for (int i = 0; _map[i].name != 0; i++)
		{
		if (strcmp(name, _map[i].name) == 0)
			return (ekey)_map[i].code;
		}

	return (ekey)_map[0].code;
}

} // namespace DEP
