/**********************************************************************/
//
// FILE:    adsorient.h
//          ADS exterior orienation parameters class.
// CREATED: 10 Feb 04  (ds)
// CHANGES:
// REFS:    - ADS40 Info Kit, April 23, 2002.
//          - Orientation Date File (ODF) Format, Sept. 2000,
//            revised Dec. 2001.
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#ifndef DEP_ADSORIENT_INCLUDED
#define DEP_ADSORIENT_INCLUDED

#include "oridata.h"

using namespace std;
using namespace GenLib2;

namespace DEP
{
/** Orientation data set class.
 *  This class holds the exterior orientation data for each scanline
 *  from the ADS scanner camera. The orientation data is read from
 *  a file in the Orientation Data File (ODF) format.
 *
 *  Attention: What is the relation between the absolute time
 *  and the base time or time tags?
 *
 *  References:
 *    - ADS40 Info Kit, April 23, 2002.
 *    - Orientation Date File (ODF) Format, Sept. 2000, revised Dec. 2001.
 *
 *  @author Dirk Stallmann
 */
class ADSOrientation : public OrientationData
{
public:
	/// units
	enum
	{
		METERS = 0,
		FEET,
		RADIANS,
		DEGREES
	};

	/// rotation sequences
	enum
	{
		OPK
	};

	/// ODF record
	typedef struct
	{
		int					t;
		int					xo;
		int					yo;
		int					zo;
		int					omega;
		int					phi;
		int					kappa;
		unsigned short		sdev_xo;
		unsigned short		sdev_yo;
		unsigned short		sdev_zo;
		unsigned short		sdev_omega;
		unsigned short		sdev_phi;
		unsigned short		sdev_kappa;
	} ODF;

	/// constructor
	ADSOrientation();

	/**@name read/write/print orientation data set */
	//@{
		/// read
		bool read(const char* fname);

		/// write
		void write(ostream& os) const;
	//@}
	
	/// get file format
	string fmt() const;

	/// read ODF header
	bool read_header(istream& is);

	/// print ODF header
	void print_header(ostream& os) const;
	
	/** get ODF header data
	 */
	//@{
		/// get number of orientation data sets
		int get_numsets() const { return _numsets; }
	
		/// get base time
		double get_basetime() const { return double(_basetime); }
	
		/// get time precision
		double get_tprec() const { return double(_tprec); }

		/// get precision position
		double get_pprec() const { return double(_pprec); }
	
		/// get precision angle
		double get_aprec() const { return double(_aprec); }
	
		/// get base X value
		double get_basex() const { return double(_basex); }

		/// get base Y value
		double get_basey() const { return double(_basey); }

		/// get base Z value
		double get_basez() const { return double(_basez); }

		/// get precision of std. dev. position
		double get_prec_sdev_pos() const { return double(_prec_sdev_pos); }

		/// get precision of std. dev. angles
		double get_prec_sdev_ang() const { return double(_prec_sdev_ang); }
	//@}
	
private:
	/** ODF header information
	 */
	//@{
		char	_version[16];
		int		_major;
		int		_minor;
		char	_datasource[32];
		char	_projectname[64];
		char	_stripname[64];
		int		_numsets;
		int		_posunits;
		int		_angularunits;
		int		_rotationseq;
		int		_abstime;			// abolute time
		char	_comments[128];
		int		_basetime;			// base time
		int		_basex;				// base X value
		int		_basey;				// base Y value
		int		_basez;				// base Z value

		// version 1.1 extension
		int		_pprec;				// precision position
		int		_aprec;				// precision angle
		int		_tprec;				// time precision
		int		_prec_sdev_pos;		// precision std. dev. position
		int		_prec_sdev_ang;		// precision std. dev. angles
	//@}
};

} // namespace DEP

#endif

