/* $Id: bunadj.cpp,v 1.33.2.1 2008/02/06 17:06:27 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <fstream>
#include "timer.h"

#include "osform.h"

#include "deperr.h"
#include "depvec.h"
#include "bunadj.h"
#include "adjintnat.h"
#include "osform.h"

#ifdef GENLIB2_HAVE_LAPACK
	#include "adjintlap.h"
#endif

#include "fiodep.h"
#include "camera.h"
#include "scanner.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "extori.h"
#include "apset.h"
#include "poscor.h"
#include "attcor.h"
#include "oridata.h"
#include "imumis.h"
#include "lconeqn.h"
#include "cholesky.h"
//#include "octransf.h"
//#include "cgicfg.h"
#include "eoquaternion.h"
#include "dlt.h"


namespace DEP
{

// erase all images, cameras, etc.

void BundleAdjustment::clear()
{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		delete *image;

	_images.clear();

	vector<Camera*>::const_iterator camera;
	for (camera = _cameras.begin(); camera != _cameras.end(); ++camera)
		delete *camera;

	_cameras.clear();

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		delete *op;

	_obj_points.clear();

	vector<Station*>::const_iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		delete *stat;

	_stations.clear();

	vector<APSet*>::const_iterator ap;
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		delete *ap;

	_ap_sets.clear();

	vector<PosCorrection*>::const_iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		delete *pcor;

	_pos_corrs.clear();

	vector<AttCorrection*>::const_iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		delete *acor;

	_att_corrs.clear();
	
	{
	vector<IMUMisalignment*>::const_iterator it;
	for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
		delete *it;
	
	_imu_mis.clear();
	}
	
	{
	vector<OrientationData*>::const_iterator it;
	for (it = _ori_data.begin(); it != _ori_data.end(); ++it)
		delete *it;

	_ori_data.clear();
	}
}


// constructor

BundleAdjustment::BundleAdjustment() :
	_pars(),
	_images(),
	_cameras(),
	_obj_points(),
	_stations(),
	_ap_sets(),
	_pos_corrs(),
	_att_corrs(),
	_imu_mis(0),
	_ori_data(0),
	//_oc_transf(0),
	_list(),
	_atpa(),
	_atpl(),
	_dec_vec(),
	_dx(),
	_qxx(),
	_ltpl(0.),
	_vtpv(0.),
	_sigma0(0.),
	_f(0),
	_adj_interface(0)
{
	// empty
}


// destructor

BundleAdjustment::~BundleAdjustment()
{
	clear();
	delete _adj_interface;
}


// read configuration file

bool BundleAdjustment::read_config(istream& cfg_in)
{
	FileIO fio;
	string keyword, fname, id;
	bool rtn = false;	// return value

	while (cfg_in >> keyword)
		{
		// check keyword
		FileIO::ecode type = fio.code(keyword.c_str());
		if (type == FileIO::UNKNOWN)
			{
			cout << "ERROR: Unknown keyword: '" << keyword << "'\n";
			return false;
			}
		else if (type == FileIO::STOP_DEP)
			break;
		
		// read filename
		cfg_in >> fname;
		/*
		// determine file type again
		if (fio.file_type(fname.c_str(), type) != true)
			{
			cout << "ERROR: Cannot determine file type of: '"
				<< fname << "'\n";
				return false;
			}
		*/
		
		// open file
		ifstream in(fname.c_str());
		if (!in)
			{
			cout << "ERROR: Cannot open input file: " << fname << endl;
			return false;
			}
		
		// apply specific read-function
		switch (type)
			{
			case FileIO::PARAMS:
				if ((rtn = _pars.read(in)) != true)
					cout << "ERROR: Reading params./options file failed.\n";
				break;

			case FileIO::LS_PARAMS:
				if ((rtn = AdjObject::lspars.read(in)) != true)
					cout << "ERROR: Reading LS-parameters file failed.\n";
				break;

			case FileIO::NETWORK_DESIGN_PARAMS:
				if ((rtn = _pn_pars.read(in)) != true)
					cout << "ERROR: Reading network design parameters file failed.\n";
				break;

			case FileIO::CAMERA:
				if ((rtn = fio.read_cameras(in, _cameras)) != true)
					cout << "ERROR: Reading camera file failed.\n";
				break;

			case FileIO::IMAGE:
				if ((rtn = fio.read_images(in, _cameras, _images)) != true)
					cout << "ERROR: Reading image file failed.\n";
				break;

			case FileIO::IMAGE_COORDS:
				if ((rtn = fio.read_image_coords(in, _images)) != true)
					cout << "ERROR: Reading image coords. file failed.\n";
				break;

			case FileIO::OBJECT_COORDS:
				if ((rtn = fio.read_object_coords(in, _obj_points)) != true)
					cout << "ERROR: Reading object coords. file failed.\n";
				break;

			case FileIO::CONTROL_SUPP:
				if ((rtn = fio.read_ctrl_supp(in, _obj_points)) != true)
					cout << "ERROR: Reading CS file failed.\n";
				break;

			case FileIO::EXT_ORI:
				if ((rtn = fio.read_stations(in, _stations)) != true)
					cout << "ERROR: Reading station file failed.\n";
				break;

			case FileIO::APSET:
				if ((rtn = fio.read_ap_sets(in, _ap_sets)) != true)
					cout << "ERROR: Reading AP set file failed.\n";
				break;

			case FileIO::POS_COR:
				if ((rtn = fio.read_pos_correction(in, _pos_corrs)) != true)
					cout << "ERROR: Reading position correction file failed.\n";
				break;

			case FileIO::ATT_COR:
				if ((rtn = fio.read_att_correction(in, _att_corrs)) != true)
					cout << "ERROR: Reading attitude correction file failed.\n";
				break;

			case FileIO::IMU_MISALIGNMENT:
				if ((rtn = fio.read_imu_misalignment(in, _imu_mis)) != true)
					cout << "ERROR: Reading IMU misalignment data file failed.\n";
				break;
				
			case FileIO::ORIENTATION_DATA_PARAMS:
				if ((rtn = OrientationData::read_params(in)) != true)
					cout << "ERROR: Reading orient. data params./options file failed.\n";
				break;
				
			case FileIO::ORIENTATION_DATA_ODF:
				cfg_in >> id;
			case FileIO::ORIENTATION_DATA:
				if ((rtn = fio.read_orientation_data(fname.c_str(), type,
						id.c_str(), _ori_data)) != true)
					cout << "ERROR: Reading orientation data file failed.\n";
				break;
				
			case FileIO::CAMERA_ADS:
				cfg_in >> id;
				if (fio.read_ads_camera(fname.c_str(), type, id.c_str(), _cameras) != true)
					cout << "ERROR: Reading ADS camera file failed.\n";
				break;
				
			//case OC_TRANSFORM:
				//if ((rtn = AdjObject::lspars.read(in)) != true)
			//	_oc_transf = new OCTransform;
			//	assert(_oc_transf != 0);
			//	if (_oc_transf->read(in) != true)
			//		{
			//		cout << "ERROR: Reading OC transformation file failed.\n";
			//		rtn = ERROR;
			//		}
			//	else
			//		rtn = true;
			//	break;

			default:
				cout << "type = " << type << endl;
				assert(false);
			}
		in.close();

		if (rtn != true)
			{
			cout << "ERROR: Reading input file failed: " << fname << endl;
			return false;
			}
		}

	return true;
}


// print cameras

void BundleAdjustment::print_cameras(ostream& os) const
{
	os << "Cameras\n";
	os << "=======\n\n";

	vector<Camera*>::const_iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		if (*cam != 0) (*cam)->print(os);

	os << endl;
}


// print image configuration

void BundleAdjustment::print_image_config(ostream& os) const
{
	os << "Image configuration\n";
	os << "===================\n\n";

	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) (*image)->print(os);

	os << endl;
}


// print image coordinates

void BundleAdjustment::print_image_coords(ostream& os) const
{
	os << "Image coordinates\n";
	os << "=================\n\n";

	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) (*image)->print_image_coords(os);

	os << endl;
}


// print object points

void BundleAdjustment::print_object_points(ostream& os) const
{
	int num_nu = 0, num_co = 0, num_ch = 0, num_in = 0 ;

	os << "Object point coordinates\n";
	os << "========================\n";
	os << "Object point type:\n";
	os << "in=intersection, co=control point, ch=check point\n\n";

	os << "Point ID       X/m         Y/m         Z/m"
		"      sX/m      sY/m      sZ/m\n\n";
		
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;

		// print single point
		(*op)->print(os);

		// object point type
		ObjectPoint::op_type type = (*op)->get_type();

		switch (type)
			{
			case ObjectPoint::invalid:	num_nu++; break;
			case ObjectPoint::inter:	num_in++; break;
			case ObjectPoint::ctrl:		num_co++; break;
			case ObjectPoint::check:	num_ch++; break;
			default:
				assert(false); break;
			}
		}
	os << endl;

	os << "Number of object points with type\n";
	os << "  intersection  :  " << num_in << endl;
	os << "  control       :  " << num_co << endl;
	os << "  check         :  " << num_ch << endl;

	int num_tot = num_in + num_co + num_ch;
	os << "  total         :  " << num_tot << endl;
	//os << "  inuse         :  " << num_tot - num_un << endl;
	if (num_nu != 0)
		os << "WARNING: " << num_nu << " invalid points found!\n";
	os << endl << endl;
}


// print control support
/*
void BundleAdjustment::print_ctrl_supp(ostream& os) const
{
	os << "Control support\n";
	os << "===============\n\n";

	//ObjectPoint tmp;
	//tmp.print_cs_info(os);
	os << "Point ID   CS\n\n";

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		if (*op != 0) (*op)->print_ctrl_supp(os);

	os << endl << endl;
}
*/


// print stations

void BundleAdjustment::print_stations(ostream& os) const
{
	os << "Stations\n";
	os << "========\n\n";
	
	vector<Station*>::const_iterator p;
	for (p = _stations.begin(); p != _stations.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print AP sets

void BundleAdjustment::print_ap_sets(ostream& os) const
{
	os << "Additional parameter sets\n";
	os << "=========================\n\n";
	
	vector<APSet*>::const_iterator p;
	for (p = _ap_sets.begin(); p != _ap_sets.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print position corrections

void BundleAdjustment::print_pos_corrs(ostream& os) const
{
	os << "Position corrections\n";
	os << "====================\n\n";
	
	vector<PosCorrection*>::const_iterator p;
	for (p = _pos_corrs.begin(); p != _pos_corrs.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print attitude corrections

void BundleAdjustment::print_att_corrs(ostream& os) const
{
	os << "Attitude corrections\n";
	os << "====================\n\n";
	
	vector<AttCorrection*>::const_iterator p;
	for (p = _att_corrs.begin(); p != _att_corrs.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print IMU misalignment

void BundleAdjustment::print_imu_misalignment(ostream& os) const
{
	os << "IMU misalignments\n";
	os << "=================\n\n";
	
	vector<IMUMisalignment*>::const_iterator p;
	for (p = _imu_mis.begin(); p != _imu_mis.end(); ++p)
		if (*p != 0) (*p)->print(os);
}


// print orientation data

void BundleAdjustment::print_orientation_data(ostream& os) const
{
	os << "Orientation data\n";
	os << "================\n\n";
	
// 	if (_ori_data.size() > 0)
// 		_ori_data[0]->print_params(os);
	OrientationData::print_params(os);
	
	vector<OrientationData*>::const_iterator it;
	for (it = _ori_data.begin(); it != _ori_data.end(); ++it)
		if (*it != 0) (*it)->print(os);
}


// print object coord. transformation
/*
void BundleAdjustment::print_oc_transf(ostream& os) const
{
	if (_oc_transf == 0) return; // escape
		
	os << "Object coordinate transformation\n";
	os << "================================\n";
	os << endl;

	_oc_transf->print(os);
	os << endl;
}
*/

// initialize bundle block adjustment:
// Check data, EOs and APs, allocate memory, etc.
// Attention: Cameras are attached previously!

bool BundleAdjustment::initialize(ostream& os)
{
	Timer timer; timer.start();
	os << "*** Start initialize adjustment" << endl << endl;
	
	// check data (still missing)
	//check_image_points(os);

	// attach image objects
	{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		assert(*image != 0);
		
		if (ScannerImage* p = dynamic_cast<ScannerImage*>(*image))
			{
			if (p->attach(_ori_data) != true)
				{
				os << "ERROR: Attach orientation data set failed.\n";
				return false;
				}
			}
		else
			{
			if ((*image)->attach(_stations) != true)
				{
				os << "ERROR: Attach station failed.\n";
				return false;
				}
			}
			
		(*image)->inuse(true); // use this image
		}
	}
	
	
	// attach AP sets
	{
	vector<APSet*>::const_iterator it;
	for (it = _ap_sets.begin(); it != _ap_sets.end(); ++it)
		{
		if (*it == 0) continue;
		/*
		if ((*it)->attach(_images) != true)
			{
			os << "ERROR: Attach AP set " << (*it)->get_id()
				<< " failed.\n";
			return false;
			}
		*/
		if ((*it)->attach(_cameras) != true)
			{
			os << "ERROR: Attach AP set " << (*it)->get_id()
				<< " failed.\n";
			return false;
			}
		}
	}
	
	// attach position correction
	{
	vector<PosCorrection*>::const_iterator it;
	for (it = _pos_corrs.begin(); it != _pos_corrs.end(); ++it)
		{
		if (*it == 0) continue;
		
		if ((*it)->attach(_images) != true)
			{
			os << "ERROR: Attach position correction " << (*it)->get_id()
				<< " failed.\n";
			return false;
			}
		}
	}
	
	// attach attitude correction
	{
	vector<AttCorrection*>::const_iterator it;
	for (it = _att_corrs.begin(); it != _att_corrs.end(); ++it)
		{
		if (*it == 0) continue;
		
		if ((*it)->attach(_images) != true)
			{
			os << "ERROR: Attach attitude correction " << (*it)->get_id()
				<< " failed.\n";
			return false;
			}
		}
	}
	
	// attach IMU misalignment
	{
	vector<IMUMisalignment*>::const_iterator it;
	for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
		{
		if (*it == 0) continue;
		
		if ((*it)->attach(_images) != true)
			{
			os << "ERROR: Attach IMU misalignment " << (*it)->get_id()
				<< " failed.\n";
			return false;
			}
		}
	}
	
	// attach orientation data sets
	{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image == 0) continue;
		
		if (ScannerImage* p = dynamic_cast<ScannerImage*>(*image))
			{
			if (p->attach(_ori_data) != true)
				{
				os << "ERROR: Attach orientation data set failed.\n";
				return false;
				}
			}
		}
	}
	
	// check object point approximations
	//if (check_approx((verb) ? &os : 0) != 0)
	if (check_approx((_pars.pr_chk_appr ? &os : 0)) != true)
		{
		os << "ERROR: Check of object point approximations failed.\n";
		return false;
		}

	// check number of rays per object point
	check_rays_per_point((_pars.pr_chk_rays ? &os : 0));

	// check number of rays per image
	check_rays_per_image((_pars.pr_chk_rays ? &os : 0));

	// calculate station approximations
	// (approximations for transformation parameters)
	// add flags for printing and debugging class params./options!!!
	if (_pars.resec_flg == true)
		{
		if (calc_station_approx(os, true) != true)
			os << "ERROR: Calculation of station approximations failed.\n";
		}
		
	// calculate EO, including position and attitude correction
	//vector<ImageGeometry*>::iterator stat;
	//for (stat = _stations.begin(); stat != _stations.end(); ++stat)
	//	if (*stat != 0) (*stat)->calc_orientation();
	
	// calculate object point approximations
	if (_pars.intersec_flg == true)
		{
		if (calc_object_point_approx(os, _pars.db_inter) != true)
			os << "ERROR: Calculation of object point approximations failed.\n";
		}
		
	// remove unused object points, EOs, AP sets etc.
	clean_up(&os);

	// prepare data
	// find time interval and set reference time
	{
	Form fix4(4); fix4.fixed().width(12);
	os << "Time interval\n";
	os << "-------------\n";
	double t_min = 0., t_max = 0.;
	time_interval(t_min, t_max);
	os << "t_min        = " << fix4(t_min) << " s\n";
	os << "t_max        = " << fix4(t_max) << " s\n";
	os << "duration dt  = " << fix4(t_max-t_min) << " s\n";
	os << endl;

	// set reference time
	set_reference_time(t_min);
	os << "Reference time t0 (= t_min) = " << fix4(t_min) << " s\n";
	os << endl;
	}
	
	// count absolute frequency of image rays
	if (_pn_pars.pr_freq_rays)
		{
		const int nbins = 20;	// number of bins
		int histo[nbins];		// histogram
		int i = 0;
		for (; i < nbins; i++) histo[i] = 0; // initilize histogram
	
		vector<ObjectPoint*>::iterator op;
		for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
			{
			// for all object points
			if(*op == 0) continue;
			int rays = 0; // count number of rays
		
			vector<Image*>::const_iterator image;
			for (image = _images.begin(); image != _images.end(); ++image)
				{
				assert(*image != 0);

				if ((*image)->image_point_exist((*op)->get_id())) rays++;
				}
			
			// frequency
			(rays < nbins) ? (++histo[rays]) : (++histo[nbins-1]);
			}
		
		os << "Absolute frequency of rays" << endl;
		os << "--------------------------" << endl;
	
		int n = 0;
		for (i = 0; i < nbins - 1; n += histo[i], i++)
			os << "  " << i << "-fold : " << histo[i] << endl;
		os << "  " << i << "-fold and more : " << histo[i] << endl;
		os << "  Sum = " << n << " (= number of object points)" << endl;
		os << endl;
		}
	
	// count number of image points/observations
	int num_ipts = 0;
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) num_ipts += (*image)->count_image_points();

	int num_obs = 2 * num_ipts;

	// create unknown parameter information (upari)
	// ... for object points
	{
	vector<ObjectPoint*>::iterator it;
	for (it = _obj_points.begin(); it != _obj_points.end(); ++it)
		if (*it != 0) (*it)->set_upari();
	}

	// ... for stations
	{
	vector<Station*>::iterator it;
	for (it = _stations.begin(); it != _stations.end(); ++it)
		if (*it != 0) (*it)->set_upari();
	}
	
	// ... for AP sets
	{
	vector<APSet*>::const_iterator it;
	for (it = _ap_sets.begin(); it != _ap_sets.end(); ++it)
		if (*it != 0) (*it)->set_upari();
	}
	
	// ... for position correction
	{
    vector<PosCorrection*>::const_iterator it;
	for (it = _pos_corrs.begin(); it != _pos_corrs.end(); ++it)
		if (*it != 0) (*it)->set_upari();
	}
	
	// ... for attitude correction
	{
	vector<AttCorrection*>::const_iterator it;
	for (it = _att_corrs.begin(); it != _att_corrs.end(); ++it)
		if (*it != 0) (*it)->set_upari();
	}
	
	// ... for IMU misalignment
	{
	vector<IMUMisalignment*>::const_iterator it;
	for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
		if (*it != 0) (*it)->set_upari();
	}
	
	// build list of adjustment objects
	// ... for object points
	{
    vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		if (*op != 0) _list.setup(*op);
	}

	// ... for stations
	{
    vector<Station*>::iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		if (*stat != 0) _list.setup(*stat);
	}

	// ... for AP sets
	{
	vector<APSet*>::const_iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		if (*aps != 0) _list.setup(*aps);
	}

	// ... for position correction
	{
    vector<PosCorrection*>::const_iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		if (*pcor != 0) _list.setup(*pcor);
	}

	// ... for attitude correction
	{
	vector<AttCorrection*>::const_iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		if (*acor != 0) _list.setup(*acor);
	}

	// ... for IMU misalignment
	{
	vector<IMUMisalignment*>::const_iterator it;
	for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
		if (*it != 0) _list.setup(*it);
	}

	// number of unknowns: observed and eliminated
	int nu = _list.num_unk();
	int nu_obs = _list.num(obs);
	int nu_eli = _list.num(eli);

	// number of contraints
	int num_con = _list.num_con();

	// degree of freedom
	int f = num_obs - nu + nu_obs + nu_eli + num_con;

	// print the number of observations, unknowns etc.
	Form int6; int6.width(6);
	
	os << "Data defined for adjustment\n";
	os << "---------------------------\n";
	os << "Images                      :  " << int6(count(_images)) << endl;
	os << "Cameras                     :  " << int6(count(_cameras)) << endl;
	os << "Stations                    :  " << int6(count(_stations)) << endl;
	os << "Additional parameter sets   :  " << int6(count(_ap_sets)) << endl;
	os << "Position corrections        :  " << int6(count(_pos_corrs)) << endl;
	os << "Attitude corrections        :  " << int6(count(_att_corrs)) << endl;
	os << "IMU misalignments           :  " << int6(count(_imu_mis)) << endl;	os << "Object points (total)       :  " << int6(num_obpts()) << endl;
	os << "Control points              :  " << int6(num_copts()) << endl;
	os << "Check points                :  " << int6(num_chpts()) << endl;
	os << endl;
	os << "Number of observations, unknowns and redundancy\n";
	os << "-----------------------------------------------\n";
	os << "Observations:\n";
	os << "   Image point coordinates      :  " << int6(num_ipts) << endl;
	os << "   Total number of observations :  " << int6(num_obs) << endl;
	os << endl;
	os << "Unknowns:                           total        thereof\n";
	os << "                                              obs.     eli.\n";
	// Hierfuer Template-Funktionen schreiben!!!
	os << "   Object point coordinates     :  "
		<< int6(_list.num_obptc()) << "   "
		<< int6(_list.num_obptc(obs)) << "   "
		<< int6(_list.num_obptc(eli)) << endl;
	os << "   Station parameters           :  "
		<< int6(_list.num_eop()) << "   "
		<< int6(_list.num_eop(obs)) << "   "
		<< int6(_list.num_eop(eli)) << endl;
	os << "   Additional parameters        :  "
		<< int6(_list.num_aps()) << "   "
		<< int6(_list.num_aps(obs)) << "   "
		<< int6(_list.num_aps(eli)) << endl;
	os << "   Position corrections         :  "
		<< int6(_list.num_pos_cor()) << "   "
		<< int6(_list.num_pos_cor(obs)) << "   "
		<< int6(_list.num_pos_cor(eli)) << endl;
	os << "   Attitude corrections         :  "
		<< int6(_list.num_att_cor()) << "   "
		<< int6(_list.num_att_cor(obs)) << "   "
		<< int6(_list.num_att_cor(eli)) << endl;
	os << "   IMU misalignments            :  "
		<< int6(_list.num_imu_mis()) << "   "
		<< int6(_list.num_imu_mis(obs)) << "   "
		<< int6(_list.num_imu_mis(eli)) << endl;
	os << "   Total number of unknowns     :  "
		<< int6(nu) << "   "
		<< int6(nu_obs) << "   "
		<< int6(nu_eli) << endl;
	os << endl;
	os << "Constraints                     :  " << int6(num_con) << endl;
	os << endl;
	os << "Redundancy                      :  " << int6(f) << endl;
	os << endl;

	// DO THIS OUTSIDE!!!
	// print unknown parameter infomation
	//upari.print(os);

	// debug upari-table
	if (_pars.db_upari)
		{
		//_list.print(cout, _ls_pars);
		_list.print(cout);
		cout.flush();
		}

	// create the desired adjustment interface
	switch(AdjObject::lspars.adj_interface())
		{
		case LSParams::NATIVE:
			_adj_interface = new AdjInterfaceNative;
			break;

#ifdef GENLIB2_HAVE_LAPACK
		case LSParams::LAPACK:
			_adj_interface = new AdjInterfaceLapack;
			break;
#endif

		default:
			os << "ERROR: Unkown adjustment interface.\n";
			return false;
			
			break;
		}

	// allocate memory for normal equations
	_atpa.newsize(nu, nu);
	_atpl.newsize(nu);
	_dec_vec.newsize(nu);
	_dx.newsize(nu);

	Form fix2(2); fix2.fixed();
	os << "Elapsed cpu time to initialize adjustment: " << fix2(timer.stop())
		<< " s\n\n";	

	return true;
}


// perform adjustment

bool BundleAdjustment::adjustment(ostream& os)
{
	Form int3; int3.width(3);
	Form fix2(2); fix2.fixed();

	Timer timer; timer.start();
	bool cv1 = false;	// convergence flag 1
	bool cv2 = false;	// convergence flag 2
	int nit = 0;		// iteration counter

	//cout << _atpa.rows() << " " << _atpa.cols() << endl;
	os << "*** Start iteration loop" << endl << endl;

	// iteration loop
	while (nit++ < AdjObject::lspars.max_iter() && !(cv1 && cv2))
		{
		Timer timer1, timer2; timer1.start();
		int num_obs = 0;	// number of observations

		os << "Iteration number " << int3(nit) << "\n";
		os << "--------------------\n\n";

		// initialize normal equation matrices, etc.
		_atpa.clear();
		_atpl.clear();
		_dx.clear();
		_ltpl = 0.;
		
		// add contributions to normal equations for ...
		// all images
		vector<Image*>::const_iterator image;
		for (image = _images.begin(); image != _images.end(); ++image)
			{
			if (*image == 0) continue;

			// add contribution to normal equations
			// for all image point observations
			if ((*image)->normal_eqns(_obj_points, _atpa, _atpl, _ltpl,
				num_obs) != true)
				{
				os << "ERROR: Setup of normal equations failed.\n";
				return false;
				}
			}

		// all image point observations
// 		vector<Station*>::iterator stat;
// 		for (stat = _stations.begin(); stat != _stations.end(); ++stat)
// 			if (*stat != 0) (*stat)->setup();
		
		// add weighted constraints
		_list.add_constraints(_atpa, _atpl, _ltpl);

		// stochastic weighting
		// all unknown parameters are treated as weighted direct observations
		_list.stoch_weighting(_atpa, _atpl, _ltpl);
/*
		// add inner contraints
		//add_inner_contraints();
		vector<ObjectPoint*>::const_iterator op;
		for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
			{
			if (*op == 0) continue;
			
			// three observation equation
			LinObsEqn eqn(3);

			// setup sub-matrices of derivatives
			DerivSub deriv(3, 7);

			// set index
			deriv.index((*op)->get_index());

			double x = (*op)->crds.x;
			double y = (*op)->crds.y;
			double z = (*op)->crds.z;
			
			// design matrix A
			deriv(0,0) = 1.;
			deriv(0,1) = 0.;
			deriv(0,2) = 0.;
			deriv(0,3) = 0.;
			deriv(0,4) =  z;
			deriv(0,5) = -y;
			deriv(0,6) =  x;
			
			deriv(1,0) = 0.;
			deriv(1,1) = 1.;
			deriv(1,2) = 0.;
			deriv(1,3) = -z;
			deriv(1,4) = 0.;
			deriv(1,5) =  x;
			deriv(1,6) =  y;
			
			deriv(2,0) = 0.;
			deriv(2,1) = 0.;
			deriv(2,2) = 1.;
			deriv(2,3) =  y;
			deriv(2,4) = -x;
			deriv(2,5) = 0.;
			deriv(2,6) =  z;
			eqn.add_a(&deriv);

			// weight
			eqn.set_p(0, 1.);

			// misclosure vector
			eqn.set_l(0, 0.);
			eqn.set_l(1, 0.);
			eqn.set_l(2, 0.);

			// add contributions to normal equation
			eqn.add_contr(_atpa, _atpl, _ltpl);
			}
*/
		// solve normal equations by Cholesky
		//atpa.print_symb(os);
		//os << endl;
		timer2.start();
		
		// 1. factorization using Cholesky decomposition
		//if (chol_dec(_atpa, _dec_vec) != true)
		//	{
		//	//sta.status = ERRNUM;
		//	os << "*** ATTENTION: Solving of normal equations failed.\n";
		//	return ERROR;
		//	}

		// 2. solving by backsubstitution
		//chol_sol(_atpa, _dec_vec, _atpl, _dx);
		
		if (solve() != true)
			{
			os << "*** ERROR: Solving of normal equations failed.\n";
			return false;
			}
			
		/*
		{
		Timer timer;
		int n = _atpa.rows();
		int p = upari.num_obptc();
		int r = n - p;
		cout << "n=" << n << ", p=" << p << ", r=" << r << endl;

		// inversion of N11
		timer.start();
		SubMatrix<double> Nii(_atpa, 0, 0, 3, 3);
		Vector<double> d(3);
		for (int i = 0; i < p; i += 3)
			{
			Nii.newpos(i, i);

			if (chol_dec(Nii, d) != true)
				{
				os << "ERROR: Cholesky decomposition failed: " << i << endl;
				return false;
				}
			chol_inv(Nii, d);
			Nii.make_upper_symmetric();
			}
		
		SubMatrix<double> Q11(_atpa, 0, 0, p, p);
		Q11.make_upper_symmetric();

		os << "Q11: " << timer.lap() << endl;

		// reduced normal equations 
		SubMatrix<double> N21(_atpa, p, 0, r, p);
		SubMatrix<double> N12(_atpa, 0, p, p, r);
		SubMatrix<double> N22(_atpa, p, p, r, r);
		SubVector<double> n1(_atpl, 0, p);
		SubVector<double> n2(_atpl, p, r);

		Matrix<double> T(r, p, 0.);
		//Matrix<double> T = N21 * Q11;
		
		SubMatrix<double> N21ij(N21, 0, 0, 3, 3);
		SubMatrix<double> Q11jj(Q11, 0, 0, 3, 3);
		SubMatrix<double> Tij(T, 0, 0, 3, 3);
		for (i = 0; i < r; i += 3)
			{
			for (int j = 0; j < p; j += 3)
				{
				N21ij.newpos(i, j);
				Q11jj.newpos(j, j);
				Tij.newpos(i, j);
				mult(Tij, N21ij, Q11jj);
				}
			}
		
		//{
		//ofstream os("NpsQp.pbm");
		//NpsQp.print_pbm(os);
		//}

		Matrix<double> S = T * N12;
		Vector<double> s = T * n1;

		Matrix<double> Nr = N22 - S;
		Vector<double> nr = n2 - s;

		{
		ofstream os("Nr.pbm");
		Nr.make_upper_symmetric();
		Nr.print_pbm(os);
		}

		os << "Nr, nr: " << timer.lap() << endl;

		//Matrix<double> d;
		if (chol_dec(Nr, d) != true)
			{
			os << "ERROR: Cholesky decomposition failed.\n";
			return false;
			}
		chol_inv(Nr, d);
		Nr.make_upper_symmetric();

		SubVector<double> dx1(dx, 0, p);
		SubVector<double> dx2(dx, p, r);
		mult(dx2, Nr, nr);

		//chol_sol(Nr, d, nr, dxs);

		Vector<double> u = N12 * dx2;
		Vector<double> v = n1 - u;
		mult(dx1, Q11, v);

		//os << "dx1:" << endl << dx1 << endl;
		//os << "dx2:" << endl << dx2 << endl;
		os << "dx: " << timer.lap() << endl;
		}
		*/
		//os << "dx:" << endl << _dx << endl;
		timer2.stop();

		// output during iteration loop
		if (_pars.pr_obj_crds)
			{
			os << "Object point coordinates\n";
			os << "------------------------\n\n";
			//os << "  Param.     State"
			//	"       Approx.         Change       Adjusted\n\n";
			vector<ObjectPoint*>::iterator op;
			for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
				if (*op != 0) (*op)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_ext_ori)
			{
			os << "Stations\n";
			os << "--------\n\n";
			vector<Station*>::iterator stat;
			for (stat = _stations.begin(); stat != _stations.end(); ++stat)
				if (*stat != 0) (*stat)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_ap_sets && !is_empty(_ap_sets))
			{
			os << "Additional parameters\n";
			os << "---------------------\n\n";
			vector<APSet*>::iterator ap;
			for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
				if (*ap != 0) (*ap)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_pos_cor && !is_empty(_pos_corrs))
			{
			os << "Position corrections\n";
			os << "--------------------\n\n";
			vector<PosCorrection*>::iterator pcor;
			for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
				if (*pcor != 0) (*pcor)->print_changes(os, _dx);
			os << endl;
			}

		if (_pars.pr_att_cor && !is_empty(_att_corrs))
			{
			os << "Attitude corrections\n";
			os << "--------------------\n\n";
			vector<AttCorrection*>::iterator acor;
			for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
				if (*acor != 0) (*acor)->print_changes(os, _dx);
			os << endl;
			}
			
		if (_pars.pr_imu_mis && !is_empty(_imu_mis))
			{
			os << "IMU misalignments\n";
			os << "-----------------\n\n";
			vector<IMUMisalignment*>::iterator it;
			for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
				if (*it != 0) (*it)->print_changes(os, _dx);
			os << endl;
			}

		// average change of unknowns
		average_change_unknowns(os);
		
		// v'Pv
		double xtatpl = dot_prod(_dx, _atpl);
		_vtpv = _ltpl - xtatpl;
		//os << vtpv << endl;
		
		// number of unknowns: observed and eliminated
		int nu = _list.num_unk();
		int nu_obs = _list.num(obs);
		int nu_eli = _list.num(eli);

		// number of contraints
		int num_con = _list.num_con();

		// redundancy
		_f = num_obs - nu + nu_obs + nu_eli + num_con;
		if (_f < 0)
			{
			os << "*** ERROR: Negative redundancy.\n";
			return false;
			}

		// sigma of unit weight aposteriori
		// Questions:
		//  (a) What is the sigma node when f = 0?
		//  (b) What is the sigma node when f < 0?
		_sigma0 = (_f > 0) ? sqrt(_vtpv / double(_f)) : (0.);
		
		Form sci3(3); sci3.scientific();
		
		os << "Number of observations  :  " << num_obs << endl;
		os << "Number of unknowns      :  " << nu << " (thereof: "
			<< nu_obs << " obs. " << nu_eli << " elim.)\n";
		os << "Number of constraints   :  " << num_con << endl;
		os << "Redundancy              :  " << _f << endl;
		os << "Sigma 0 a posteriori    :  ";
		if (_sigma0 == 0.)
			os << "undefined, no redundancy!\n";
		else
			os << fix2(_sigma0 * 1e6) << " um\n";
		os << endl;
		os.flush();

		// update unknown parameters
		update_unknowns(os);
		
		// test on divergence (v'Pv)[i-1] = (v'Pv)[i]
		double vtpv_prev = 0.;
		if (nit > 1)
			{
			if (!(vtpv_prev < _vtpv))
				os << "\n*** WARNING: Iteration loop do not converge.\n";
			}
		vtpv_prev = _vtpv;
		
		os << "v'Pv                    :  " << sci3(_vtpv) << " m^2\n";
		os << "l'Pl                    :  " << sci3(_ltpl) << " m^2\n";
		os << "x'A'Pl                  :  " << sci3(xtatpl) << " m^2\n";
		os << endl;
		
		// 1st (general) check on convergence
		double eps = AdjObject::lspars.conv_eps();
		if (AdjObject::lspars.conv_chk() == true)
			{
			os << "Convergence check\n";
			os << "eps                     :  " << sci3(eps) << " m^2\n";
			os << "x'A'Pl < eps            :  " << (xtatpl < eps ? "true" : "false");
			os << " -> convergence limit ";
			if (xtatpl < eps)
				{
				cv1 = true;
				os << "reached\n";
				}
			else
				{
				cv1 = false;
				os << "not reached\n";
				}
			os << endl;
			}
			else
				cv1 = true;
		
		// 2nd (specific) check on convergence
		if (AdjObject::lspars.chk_obj() == true
		    || AdjObject::lspars.chk_pcc() == true
			|| AdjObject::lspars.chk_rot() == true)
				cv2 = _list.check_convergence(_dx);
		else
			cv2 = true;

		os << "Elapsed cpu time per iteration: " << fix2(timer1.stop())
			<< " s, solution only: " << fix2(timer2.time()) << " s\n\n";
		}
		// *** END OF ITERATION LOOP ***

	nit--;

	if (nit >= AdjObject::lspars.max_iter())
		{
		os << "\n*** WARNING: Maximum number of iterations "
			"has been reached.\n";
		}
	else
		{
		os << "\n*** Iteration criterion limit has been reached.\n";
		}

	os << "\nNumber of iterations: " << nit << endl;
	os << "Total cpu time for iterations : " << fix2(timer.stop()) << " s\n\n";

	// DON'T make the normal matrix symmetric! The inversion will fail
	// because the Cholesky factor is stored in the lower triangle!
	//_atpa.make_upper_symmetric();

	// write structure of normal matrix in PBM-format
	if (_pars.fname_normal_matrix_struct.compare("none") != 0)
		{
		ofstream os(_pars.fname_normal_matrix_struct.c_str());
		if (!os)
			cout << "ERROR: Cannot write structure of normal matrix file.\n";
		else
			write_normal_matrix_struct(os);
		}

	// write elements of normal matrix in ASCII-format
	if (_pars.fname_normal_matrix.compare("none") != 0)
		{
		ofstream os(_pars.fname_normal_matrix.c_str());
		if (!os)
			cout << "ERROR: Cannot write normal matrix file.\n";
		else
			write_normal_matrix(os);
		}
				
	// write elements of ride hand side vector in ASCII-format
	if (_pars.fname_rhs_vector.compare("none") != 0)
		{
		ofstream os(_pars.fname_rhs_vector.c_str());
		if (!os)
			cout << "ERROR: Cannot write right side vector file.\n";
		else
			write_rhs_vector(os);
		}
				
	// write elements of solution vector in ASCII-format
	if (_pars.fname_solution_vector.compare("none") != 0)
		{
		ofstream os(_pars.fname_solution_vector.c_str());
		if (!os)
			cout << "ERROR: Cannot write solution vector file.\n";
		else
			write_solution_vector(os);
		}
	
	// check: A'Pv = 0
	/* This is a test on the adjustment.
	 * The condition A'Pv = 0 can also be tested on A'Pl
	 * since l = -v after iterations.
	 * ATTENTION: For eliminated unknowns and constraints
	 * there might be a large number added! Due to the multiplication
	 * with a large weight pii.
	 */
	check_atpv(os);

	// Hauptprobe missing!!!!

	// print adjusted unknown parameters
	// and image coord. residuals outside!

	return true;
}


// print unknown parameters

void BundleAdjustment::print_unknowns(ostream& os) const
{
	// print object points
	if (_obj_points.size() > 0) print_object_points(os);

	// print estimated parameters
	if (_stations.size() > 0)
		{
		os << "Stations\n";
		os << "========\n\n";
		vector<Station*>::const_iterator stat;
		for (stat = _stations.begin(); stat != _stations.end(); ++stat)
			if (*stat != 0) (*stat)->print_unknowns(os);
		}

	if (_ap_sets.size() > 0)
		{
		os << "Additional parameter sets\n";
		os << "=========================\n\n";
		vector<APSet*>::const_iterator ap;
		for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
			if (*ap != 0) (*ap)->print_unknowns(os);
		}

	if (_pos_corrs.size() > 0)
		{
		os << "Position corrections\n";
		os << "====================\n\n";
		vector<PosCorrection*>::const_iterator pcor;
		for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
			if (*pcor != 0) (*pcor)->print_unknowns(os);
		}

	if (_att_corrs.size() > 0)
		{
		os << "Attitude corrections\n";
		os << "====================\n\n";
		vector<AttCorrection*>::const_iterator acor;
		for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
			if (*acor != 0) (*acor)->print_unknowns(os);
		}
		
	if (_imu_mis.size() > 0)
		{
		os << "IMU Misalignments\n";
		os << "=================\n\n";
		vector<IMUMisalignment*>::const_iterator it;
		for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
			if (*it != 0) (*it)->print_unknowns(os);
		}
	os << endl;
}


// check point analysis
// The practical precision estimates/empirical accuracy are computed
// form all check point coordinate differences.
// The accuracy will be derived ONLY from NEW object points.

void BundleAdjustment::check_point_analysis(ostream& os) const
{
	Form str8; str8.width(8);
	Form int9; int9.width(9);
	Form fix9(3); fix9.fixed().width(9);
	Form fix12(3); fix12.fixed().width(12);

	DPoint3 d, d_max, sum_d, sum_dd;
	int num_x = 0, num_y = 0, num_z = 0;

	os << "Check point analysis\n";
	os << "====================\n";
	os << endl;
	os << "Point ID    X-new/m      Y-new/m      Z-new/m "
		"       DX/m      DY/m      DZ/m\n\n";

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;

		// only for check points!!!
		if (!(*op)->is_check()) continue;

		os << str8((*op)->get_id()) << " "
			<< fix12((*op)->crds.x) << " "
			<< fix12((*op)->crds.y) << " "
			<< fix12((*op)->crds.z) << "  ";

		if ((*op)->is_chx())
			{
			//d.x = upari.get_obs(upari_x) - (*op)->crds.x;
			d.x = (*op)->ival(0) - (*op)->crds.x;
			sum_d.x += d.x;
			sum_dd.x += d.x * d.x;
			num_x++;
			if (fabs(d.x) > d_max.x) d_max.x = fabs(d.x); 
			os << " " << fix9(d.x);
			}
		else
			os << "     ---  ";

		if ((*op)->is_chy())
			{
			//d.y = upari.get_obs(upari_y) - (*op)->crds.y;
			d.y = (*op)->ival(1) - (*op)->crds.y;
			sum_d.y += d.y;
			sum_dd.y += d.y * d.y;
			num_y++;
			if (fabs(d.y) > d_max.y) d_max.y = fabs(d.y); 
			os << " " << fix9(d.y);
			}
		else
			os << "     ---  ";

		if ((*op)->is_chz())
			{
			//d.z = upari.get_obs(upari_z) - (*op)->crds.z;
			d.z = (*op)->ival(2) - (*op)->crds.z;
			sum_d.z += d.z;
			sum_dd.z += d.z * d.z;
			num_z++;
			if (fabs(d.z) > d_max.z) d_max.z = fabs(d.z); 
			os << " " << fix9(d.z);
			}
		else
			os << "     ---  ";

		os << endl;
		}
	os << endl;
	
	double mean_x = num_x == 0 ? 0. : sum_d.x / double(num_x);
	double mean_y = num_y == 0 ? 0. : sum_d.y / double(num_y);
	double mean_z = num_z == 0 ? 0. : sum_d.z / double(num_z);

	double rmse_x = num_x == 0 ? 0. : sqrt(sum_dd.x / double(num_x));
	double rmse_y = num_y == 0 ? 0. : sqrt(sum_dd.y / double(num_y));
	double rmse_z = num_z == 0 ? 0. : sqrt(sum_dd.z / double(num_z));

	os << "Summary of check point analysis\n";
	os << "-------------------------------\n";
	os << "Empirical accuracy in object space.\n\n";
	os << "                                  DX/m      DY/m      DZ/m\n\n";
	os << "Number of differences       :  "
		<< int9(num_x) << " " << int9(num_y) << " " << int9(num_z) << endl;
	os << "Maximum of abs. differences :  "
		<< fix9(d_max.x) << " " << fix9(d_max.y) << " " << fix9(d_max.z)
		<< endl;
	os << "Average of differences      :  "
		<< fix9(mean_x) << " " << fix9(mean_y) << " " << fix9(mean_z) << endl;
	os << "RMS of differences          :  "
		<< fix9(rmse_x) << " " << fix9(rmse_y) << " " << fix9(rmse_z) << endl;
	
	double sdev_x = num_x <= 1 ? 0. : sqrt((sum_dd.x - 2. * mean_x * sum_d.x
		+ double(num_x) * mean_x * mean_x) / double(num_x - 1));
	double sdev_y = num_y <= 1 ? 0. : sqrt((sum_dd.y - 2. * mean_y * sum_d.y
		+ double(num_y) * mean_y * mean_y) / double(num_y - 1));
	double sdev_z = num_z <= 1 ? 0. : sqrt((sum_dd.z - 2. * mean_z * sum_d.z
		+ double(num_z) * mean_z * mean_z) / double(num_z - 1));
	
	os << "Std. devs. of differences   :  "
		<< fix9(sdev_x) << " " << fix9(sdev_y) << " " << fix9(sdev_z) << endl;
	os << endl;
}


// control point residuals

void BundleAdjustment::ctrl_point_residuals(ostream& os) const
{
	Form str8; str8.width(8);
	Form int9; int9.width(9);
	Form fix9(3); fix9.fixed().width(9);
	Form fix12(3); fix12.fixed().width(12);

	//ObjectPoint* op = 0;
	DPoint3 v, v_max, sum_v, sum_vv;
	//double sum_vxvx = 0., sum_vyvy = 0., sum_vzvz = 0.;
	//int num_vxvx = 0, num_vyvy = 0, num_vzvz = 0;
	int num_x = 0, num_y = 0, num_z = 0;

	//os << endl;
	os << "Control point residuals\n";
	os << "=======================\n\n";
	os << "Point ID        X/m          Y/m          Z/m"
		"      vX/m      vY/m      vZ/m\n\n";

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		//assert(*op != 0);
		if(*op == 0 || !(*op)->is_ctrl()) continue;

		os << str8((*op)->get_id()) << " "
			<< fix12((*op)->crds.x) << " "
			<< fix12((*op)->crds.y) << " "
			<< fix12((*op)->crds.z);

		if (AdjObject::lspars.is_obs((*op)->sdev.x))
			{
			v.x = (*op)->crds.x - (*op)->ival(0);
			sum_v.x += v.x;
			sum_vv.x += v.x * v.x;
			num_x++;
			if (fabs(v.x) > v_max.x) v_max.x = fabs(v.x); 
			os << " " << fix9(v.x);
			}
		else
			os << "     ---  ";

		if (AdjObject::lspars.is_obs((*op)->sdev.y))
			{
			v.y = (*op)->crds.y - (*op)->ival(1);
			sum_v.y += v.y;
			sum_vv.y += v.y * v.y;
			num_y++;
			if (fabs(v.y) > v_max.y) v_max.y = fabs(v.y); 
			os << " " << fix9(v.y);
			}
		else
			os << "     ---  ";

		if (AdjObject::lspars.is_obs((*op)->sdev.z))
			{
			v.z = (*op)->crds.z - (*op)->ival(2);
			sum_v.z += v.z;
			sum_vv.z += v.z * v.z;
			num_z++;
			if (fabs(v.z) > v_max.z) v_max.z = fabs(v.z); 
			os << " " << fix9(v.z);
			}
		else
			os << "     ---  ";

		os << endl;
		}
	os << endl;

	double mean_x = num_x == 0 ? 0. : sum_v.x / double(num_x);
	double mean_y = num_y == 0 ? 0. : sum_v.y / double(num_y);
	double mean_z = num_z == 0 ? 0. : sum_v.z / double(num_z);

	double rmse_x = num_x == 0 ? 0. : sqrt(sum_vv.x / double(num_x));
	double rmse_y = num_y == 0 ? 0. : sqrt(sum_vv.y / double(num_y));
	double rmse_z = num_z == 0 ? 0. : sqrt(sum_vv.z / double(num_z));
	
	os << "Std. dev. from control point residuals\n";
	os << "--------------------------------------\n";
	os << "                                 vX/m      vY/m      vZ/m\n\n";
	os << "Number of residuals        :  "
		<< int9(num_x) << " " << int9(num_y) << " " << int9(num_z) << endl;
	os << "Maximum of abs. residuals  :  "
		<< fix9(v_max.x) << " " << fix9(v_max.y) << " " << fix9(v_max.z)
		<< endl;
	os << "Average of residuals       :  "
		<< fix9(mean_x) << " " << fix9(mean_y) << " " << fix9(mean_z) << endl;
	os << "RMS of residuals           :  "
		<< fix9(rmse_x) << " " << fix9(rmse_y) << " " << fix9(rmse_z) << endl;
	os << endl;
}


// calculate precision

bool BundleAdjustment::precision(ostream* os)
{
	if (os != 0)
		{
		(*os) << "*** Matrix inversion ... ";
		(*os).flush();
		}

	Form fix2(2); fix2.fixed();
	Timer timer; timer.start();

	if (inversion() != true)
		{
		cout << "*** ERROR: Inversion of normal matrix failed.\n";
		return false;
		}

	if (os != 0)
		{
		(*os) << "Done\n\n";
		(*os) << "Elapsed cpu time for inversion: " << fix2(timer.stop())
			<< " s\n\n";	
		}

	// update std. devs.
	//cout << _list.update_sdev(_sigma0, _qxx) << endl;
	_list.update_sdev(_sigma0, _qxx);

	return true;
}


// precision for estimated parameters (partially with significance test)

void BundleAdjustment::precision(ostream& os) const
{
	os << "*** Precision from Qxx and Sigma 0 a posteriori ***" << endl;
	os << endl;

	// print object points
	if (_obj_points.size() > 0) print_object_points(os);

	// average precision of object coords.
	average_precision(os);
	os << endl;

	// print stations
	if (_stations.size() > 0)
		{
		os << "Stations\n";
		os << "========\n\n";
		vector<Station*>::const_iterator it;
		for (it = _stations.begin(); it != _stations.end(); ++it)
			if (*it != 0) (*it)->print_unknowns(os);
		}

	// precision for estimated parameters with significance test
	if (_ap_sets.size() > 0)
		{
		os << "Additional parameter sets\n";
		os << "=========================\n\n";
		vector<APSet*>::const_iterator it;
		for (it = _ap_sets.begin(); it != _ap_sets.end(); ++it)
			if (*it != 0) (*it)->print_precision(os);
		}

	if (_pos_corrs.size() > 0)
		{
		os << "Position corrections\n";
		os << "====================\n\n";
		vector<PosCorrection*>::const_iterator it;
		for (it = _pos_corrs.begin(); it != _pos_corrs.end(); ++it)
			if (*it != 0) (*it)->print_precision(os);
		}

	if (_att_corrs.size() > 0)
		{
		os << "Attitude corrections\n";
		os << "====================\n\n";
		vector<AttCorrection*>::const_iterator it;
		for (it = _att_corrs.begin(); it != _att_corrs.end(); ++it)
			if (*it != 0) (*it)->print_precision(os);
		}

	if (_imu_mis.size() > 0)
		{
		os << "IMU misalignments\n";
		os << "=================\n\n";
		vector<IMUMisalignment*>::const_iterator it;
		for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
			if (*it != 0) (*it)->print_precision(os);
		}
		
	os << endl;
}


// print correlations in list form

void BundleAdjustment::correlations_list(ostream& os) const
{
	Form fix(2); fix.fixed();
	double ccoef_lim = AdjObject::lspars.ccoef_lim();

	os << "Correlations\n";
	os << "============\n";
	os << "Output limit of correlation coeffs.:  " << fix(ccoef_lim) << endl;
	os << endl;
	os << "      i      j    Type      ID                Type      ID                 rij\n";
	os << endl;
	_list.print_correlations(os, _qxx, ccoef_lim);
	os << endl;
}


// print correlations in matrix form

void BundleAdjustment::correlations_matrix(ostream& os) const
{
	Form fix(2); fix.fixed();
	double ccoef_lim = AdjObject::lspars.ccoef_lim();

	os << "Correlations\n";
	os << "============\n";
	os << "Correlation coeffs. larger " << fix(ccoef_lim)
		<< " marked with *\n\n";
	
	_list.print_corr_matrix(os, _qxx, ccoef_lim);
	os << endl;
}


// print covariances

void BundleAdjustment::print_covariances(ostream& os) const
{
	os << "Corvariances\n";
	os << "============\n";
	os << endl;

	_list.print_covar_matrix(os, _qxx, _sigma0);
	os << endl;
}


// print influence of APs on image points

void BundleAdjustment::aps_influence(ostream& os) const
{
	os << "Influence of APs on image points\n";
	os << "================================\n";
	os << "Without the influence of principal point changes.\n";
	os << "Number of grid points  :  nx = " << _pars.pr_aps_nx
		<< ",  ny = " << _pars.pr_aps_ny << endl;
	os << "(x,y)-corrections in um\n";
	os << endl;
/*
	if (is_empty(_ap_sets))
		{
		os << "*** ATTENTION: List of AP sets is empty.\n";
		}
	else
		{
		vector<APSet*>::const_iterator aps;
		for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
			{
			if (*aps == 0) continue;
			(*aps)->print_aps_influence(os, _pars);
			}
		}
*/
	vector<Camera*>::const_iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		{
		if (*cam == 0) continue;
			
		(*cam)->print_aps_influence(os, _pars);
		}
	os << endl;
}


// remove image points of unused object points

void BundleAdjustment::clean_image_points(ostream *os)
{
	int num = 0;

	if (os != 0)
		{
		(*os) << "Clean image points\n";
		(*os) << "==================\n";
		(*os) << "(Remove image coords. of unused object points.)\n\n";
		}

	// loop over all object points
	vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if(*op == 0 || !(*op)->is_unused()) continue;

		num += remove_image_point((*op)->get_id(), os);
		}

	if (os != 0)
		{
		(*os) << endl;
		(*os) << "Number of removed image points: " << num << endl;
		(*os) << endl;
		}
}


// check approximations of object coords.
// Loop over all image points and create new object points
// when corresponding point don't exist in the object point list.

bool BundleAdjustment::check_approx(ostream* os)
{
	int num = 0;

	if (os != 0)
		{
		(*os) << "Check object point approximations\n";
		(*os) << "=================================\n\n";
		}

	// loop over all images
	vector<Image*>::const_iterator img;
	for (img = _images.begin(); img != _images.end(); ++img)
		{
			
		// loop over all image points
		vector<ImagePoint*>::iterator ip;
		for (ip = (*img)->_image_points.begin();
			ip != (*img)->_image_points.end(); ++ip)
			{
			if (*ip == 0) continue;
			string id = (*ip)->get_id();

			// search object points
			ObjectPoint* op = find(_obj_points, id);
			
			if (op == 0)
				{
				if (os != 0)
					(*os) << "Point " << id << " has no approximations.\n";

				// create new object point.
				// default: intersection point without approximations.
    			ObjectPoint* op = new ObjectPoint;
				assert(op != 0);
				op->set_id(id);
				//op->sdev = op->smaxu();
				//op->sdev = AdjObject::lspars.smaxu();
				//op->type(ObjectPoint);
				_obj_points.push_back(op);

				num++;
				}
			}
        }

	if (os != 0)
		{
		(*os) << endl;
		(*os) << "Number of additional object points: " << num << endl;
		(*os) << endl;
		}
	
	return true;
}


// check number of rays per object point

void BundleAdjustment::check_rays_per_point(ostream* os)
{
	int removed = 0;
	
	if (os != 0)
		{
		(*os) << "Check number of rays per object point\n";
		(*os) << "=====================================\n";
		(*os) << "(Count number of rays per object point)\n";
		(*os) << "Number of points to check : " << _obj_points.size() << endl;
		(*os) << "Minimal number of rays per object point : "
			<< _pn_pars.min_rays_point() << "\n\n";
		(*os) << "<point ID>: <image ID> ... [<number of rays>]\n\n";
		}

	vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		// for all object points
		if(*op == 0) continue;

		int rays = 0; // count number of rays
		string id = (*op)->get_id();

		if (os) (*os) << id << ":";

		vector<Image*>::const_iterator image;
		for (image = _images.begin(); image != _images.end(); ++image)
			{
			assert(*image != 0);

			if ((*image)->image_point_exist(id))
				{
				rays++;
				if (os) (*os) << " " << (*image)->get_id();
				}
			}

		if (os != 0) (*os) << " [" << rays << "]";
		(*op)->num_rays(rays); // save number of rays
		
		// check the number of rays
		// as a course measure of reliability
		//if (rays < 2)
		if (rays < _pn_pars.min_rays_point())
			{
			// single ray option for full control points only!!!
			//if (pr.sray_opt && rays == 1 && (*op)->is_ctrl()
			//	&& (*op)->is_full())
			//if (_pars.sray_opt && rays == 1 && (*op)->is_csx(_ls_pars)
			//	&& (*op)->is_csy(_ls_pars) && (*op)->is_csz(_ls_pars))
			//if (_pars.sray_opt && rays == 1 && (*op)->is_csx()
			//	&& (*op)->is_csy() && (*op)->is_csz())
			if (rays == 1 && (*op)->is_csx()
				&& (*op)->is_csy() && (*op)->is_csz())
				{
				(*op)->inuse(true);
				if (os) (*os) << " -> single ray";
				}
			else
				{
				// delete it really
				delete *op;
				*op = 0;
				removed++;

				if (os) (*os) << " -> removed";

				// remove image point
				//num += remove_ip(id, os);
				vector<Image*>::const_iterator image;
				for (image = _images.begin(); image != _images.end(); ++image)
					{
					if (*image == 0) continue;
					string img_id = (*image)->get_id(); // save image ID

					if ((*image)->remove_image_point(id))
						{
						if (os)
							(*os) << "\nATTENTION: Image " << img_id
								<< " point " << id << " removed.\n";
						}
					}
				}
			}
		else
			{
			(*op)->inuse(true);
			//(*op)->num_rays(rays); // save number of rays
			}
		if (os) (*os) << endl;
		}

	if (os)
		(*os) << "Number of removed object points: " << removed << "\n\n";
}


// check number of rays per image

void BundleAdjustment::check_rays_per_image(ostream* os)
{
	int removed = 0;

	if (os != 0)
		{
		(*os) << "Check number of rays per image\n";
		(*os) << "==============================\n";
		(*os) << "(Count number of rays per image)\n";
		(*os) << "Number of images to check : " << _images.size() << endl;
		(*os) << "Minimal number of rays per image : "
			<< _pn_pars.min_rays_image() << "\n\n";
		(*os) << "<image ID>: <number of rays>\n\n";
		}

	vector<Image*>::iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if(*image == 0) continue;

		int rays = (*image)->count_image_points();

		if (os) (*os) << (*image)->get_id() << ": " << rays;

		if (rays < _pn_pars.min_rays_image())
			{
			delete *image;
			*image = 0;
			removed++;
			if (os) (*os) << " -> removed";
			}
		if (os) (*os) << endl;
		}

	if (os)
		(*os) << "Number of removed images: " << removed << "\n\n";
}


// calculate station approximations

bool BundleAdjustment::calc_station_approx(ostream& os, bool dbg)
{
	os << "Calculate station approximations\n";
	os << "================================\n";
	os << "(By spatial resection)\n";
	os << endl;
	
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image == 0) continue;
		
		if ((*image)->resection(os, _obj_points, dbg) != true)
			os << "ATTENTION: Resection failed." << endl;
		os << endl;
		}
	os << endl;
	
	return true;
}


// calculate object point approximations by spatial intersection

bool BundleAdjustment::calc_object_point_approx(ostream& os, bool dbg)
{
	os << "Calculate object point approximations\n";
	os << "=====================================\n";
	os << "(By spatial intersection)\n";
	os << endl;
	
	int nu = 3;
	DMatrix atpa(nu, nu);
	DVector atpl(nu);
	
	// submatrices for ...
	DerivSub	deriv_oc(2,3);	// object point coords., size always 2x3
	DerivSub	deriv_obs(2,2);	// observations, size always 2x2

	// B-matrix elements of observations
	deriv_obs.identity();
	
	// set index for object coords.
	deriv_oc.index(0);

	// always 2 condition equations for a single imaging ray
	LinConditionEqn con_eqn;

	// assign A-submatrix
	con_eqn.add_a(&deriv_oc);
	
	// assign B-submatrix
	con_eqn.set_b(deriv_obs);
	
	// co-factors: set all weights to be equal!
	DMatrix Qll(2,2);
	Qll(0,0) = 1.;
	Qll(1,1) = 1.;
	con_eqn.set_q(Qll);
	
	DVector	wt(2);			// misclosure vector
	int		num_pts = 0;	// count number of intersection points
	
	Form	fix12(3); fix12.fixed().width(12);
	
	// loop over all object points
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if(*op == 0 || (*op)->is_approx()) continue;
		
		os << "Point ID : " << (*op)->get_id() << endl;

		bool found = false;
		int num_rays = 0;
		double ltpl = 0.;
		
		// clear normal equation system
		atpa.clear();
		atpl.clear();
	
		os << "Images with this point:";
		vector<Image*>::const_iterator image;
		for (image = _images.begin(); image != _images.end(); ++image)
			{
			if (*image == 0) continue;
		
			if ((*image)->observ_eqn_intersec((*op)->get_id(), deriv_oc,
				wt, found) == false)
				{
				os << "ATTENTION: Calculation of observation equations failed." << endl;
				return false;
				}
				
			if (!found) continue;
				
			if (num_rays % 4 == 0) os << endl;
			os << "  " << (*image)->get_id();
			
			con_eqn.set_w(wt);

			// add contributions to normal equation
			if (!con_eqn.add_contr(atpa, atpl, ltpl))
				assert(false);
			
			num_rays++;
			}
		os << endl << endl;
		
		int num_obs = 2 * num_rays;
		os << "Number of rays : " << num_rays << endl;
		if (num_obs < nu)
			{
			os << "ATTENTION: Too few rays to compute object coordinates." << endl;
			return false;
			}
		os << endl;
	
		// solve normal equation sytem
		DVector dec;
		int info = chol_dec(atpa, dec);
		if (info != 0)
			{
			os << "Decomposition of normal matrix failed in row "
				<< info << endl;
			return false;
			}

		DVector x;
		info = chol_sol(atpa, dec, atpl, x);
		if (info != 0)
			{
			os << "Backsubstitution of normal equation system failed: "
			<< "info = " << info << endl;
			return false;
			}

		(*op)->crds.x = x(0);
		(*op)->crds.y = x(1);
		(*op)->crds.z = x(2);
		(*op)->set_approx(true);

		os << " X = " << fix12(x(0))
			<< "   Y = " << fix12(x(1))
			<< "   Z = " << fix12(x(2)) << endl << endl;

		num_pts++;
		}
	os << endl;

	os << "Number of intersection points : " << num_pts << endl;
	os << endl << endl;

	// precision of intersection point coords. missing!!!

	return true;
}


// remove unused object points, EOs, Apsets, etc.
/*
			os << "ATTENTION: Image " << img->get_id()
				<< " without image points: image removed.\n";
*/

void BundleAdjustment::clean_up(ostream* os)
{
	if (os)
		{
		(*os) << "Clean up\n";
		(*os) << "--------\n";
		(*os) << "(Remove unused object points, stations, AP sets, etc.)\n";
		(*os) << endl;
		}

	// remove images
	int num_imgs = 0;
	vector<Image*>::iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0 && (*image)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Image " << (*image)->get_id()
					<< " removed." << endl;
			delete *image;
			*image = 0;
			num_imgs++;
			}
		}

	// remove unused cameras
	int num_cams = 0;
	vector<Camera*>::iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		{
		if (*cam != 0 && (*cam)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Camera " << (*cam)->get_id()
					<< " removed." << endl;
			delete *cam;
			*cam = 0;
			num_cams++;
			}
		}

	// remove unused object points
	int num_opts = 0;
	vector<ObjectPoint*>::iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;
		
		if ((*op)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Object point " << (*op)->get_id()
					<< " removed." << endl;
			delete *op;
			*op = 0;
			num_opts++;
			}
		//else if ((*op)->is_approx())
		//	{
			// change type from approx to inter
		//	(*op)->type(ObjectPoint::inter);
		//	}
		}

	// remove unused stations
	int num_eos = 0;
	vector<Station*>::iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		{
		if (*stat != 0 && (*stat)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Station " << (*stat)->get_id()
					<< " removed." << endl;
			delete *stat;
			*stat = 0;
			num_eos++;
			}
		}

	// remove unused AP sets
	int num_aps = 0;
	vector<APSet*>::iterator aps;
	for (aps = _ap_sets.begin(); aps != _ap_sets.end(); ++aps)
		{
		if (*aps != 0 && (*aps)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: AP set " << (*aps)->get_id()
					<< " removed." << endl;
			delete *aps;
			*aps = 0;
			num_aps++;
			}
		}

	// remove unused position corrections
	int num_pcor = 0;
	vector<PosCorrection*>::iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		{
		if (*pcor != 0 && (*pcor)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: position correction " << (*pcor)->get_id()
					<< " removed." << endl;
			delete *pcor;
			*pcor = 0;
			num_pcor++;
			}
		}

	// remove unused attitude corrections
	int num_acor = 0;
	vector<AttCorrection*>::iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		{
		if (*acor != 0 && (*acor)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: attitude correction " << (*acor)->get_id()
					<< " removed." << endl;
			delete *acor;
			*acor = 0;
			num_acor++;
			}
		}
		
	// remove unused IMU misalignments
	int num_imu = 0;
	vector<IMUMisalignment*>::iterator p;
	for (p = _imu_mis.begin(); p != _imu_mis.end(); ++p)
		{
		if (*p != 0 && (*p)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: IMU misalignment " << (*p)->get_id()
					<< " removed." << endl;
			delete *p;
			*p = 0;
			num_imu++;
			}
		}

	// remove unused orientation data
	int num_ods = 0;
	vector<OrientationData*>::iterator od;
	for (od = _ori_data.begin(); od != _ori_data.end(); ++od)
		{
		if (*od != 0 && (*od)->is_unused())
			{
			if (os)
				(*os) << "ATTENTION: Orientation data " << (*od)->get_id()
					<< " removed." << endl;
			delete *od;
			*od = 0;
			num_ods++;
			}
		}

	if (os)
		{
		(*os) << endl;
		(*os) << "Number of removed ..." << endl;
		(*os) << "  - images               : " << num_imgs << endl;
		(*os) << "  - cameras              : " << num_cams << endl;
		(*os) << "  - object points        : " << num_opts << endl;
		(*os) << "  - stations             : " << num_eos << endl;
		(*os) << "  - AP sets              : " << num_aps << endl;
		(*os) << "  - position corrs.      : " << num_pcor << endl;
		(*os) << "  - attitude corrs.      : " << num_acor << endl;
		(*os) << "  - IMU misalignments    : " << num_imu << endl;
		(*os) << "  - orientation data set : " << num_ods << endl;
		(*os) << endl;
		}
}


/* v-check is a test on the adjustment.
 * The condition A'Pv = 0 can also be tested on A'Pl
 * since l = -v after iterations.
 */

void BundleAdjustment::check_atpv(ostream& os) const
{
	Form sci3(1); sci3.scientific().width(6);

	os << "Check adjustment on A'Pv = 0" << endl;
	os << "----------------------------" << endl;
	os << "A'Pv limit:  " << sci3(AdjObject::lspars.atpv_lim()) << endl;

	int num_err = 0;
	for (int i = 0; i < _atpl.size(); i++)
		{
		if (AdjObject::lspars.chk_atpv(_atpl(i)))
			{
			AdjObject* p = _list.find(i);
			assert(p != 0);
			
			os << "A'Pv[" << i << "] = " << sci3(_atpl(i)) << " is large:"
				<< " class: " << p->class_str()
				<< ", subtype: " << p->type_str()
				<< ", ID: " << p->get_id()
				<< ", param.: " << p->uname_idx(i) << endl;
			num_err++;
			}
		}
	os << endl;

	os << "Number of A'Pv-values > limit : " << num_err << endl;
	os << endl;
}


// average change of unknowns

void BundleAdjustment::average_change_unknowns(ostream& os) const
{
	Form sci3(3); sci3.scientific().width(10);

	os << "Average change of unknown parameters\n";
	os << " - object point coords.             :  ";

	double av = 0.;
	av = _list.average_change_obj(_dx);
	if (av < 0.)
		os << "    ---   \n";
	else
		os << sci3(av) << " m\n";

	os << " - perspective centre coords.       :  ";
	av = _list.average_change_pcc(_dx);
	if (av < 0.)
		os << "    ---   \n";
	else
		os << sci3(av) << " m\n";

	os << " - rotation parameters              :  ";
	av = _list.average_change_rot(_dx);
	if (av < 0.)
		os << "    ---   \n";
	else
		os << sci3(av) << "\n";

	os << endl;
}


// update unknowns and EO of the images

void BundleAdjustment::update_unknowns(ostream& os)
{
	// update all unknowns with the solution vector
	//cout << "Number of updated unknown parameters: "
	//	<< _list.update_unknowns(_dx) << endl;
	_list.update_unknowns(_dx);

	// calculate EO, including position and attitude correction
	//vector<ImageGeometry*>::iterator stat;
	//for (stat = _stations.begin(); stat != _stations.end(); ++stat)
	//	if (*stat != 0) (*stat)->calc_orientation();
}


// average precision of object coords. derived from all
// standard deviations of coordinates of non-control points.

void BundleAdjustment::average_precision(ostream& os) const
{
	Form int9; int9.width(9);
	Form fix9(3); fix9.fixed().width(9);

	DPoint3 sum_var, sd_max;
	int num_x = 0, num_y = 0, num_z = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;
		/*
		// check index
		if ((k = (*op)->index()) < 0)
			{
			cout << "ATTENTION: average_precision: Object point #" <<
				(*op)->get_id() << " has invalid index.\n";
			continue;
			}
		*/

		//if ((*op)->is_ctrl()) continue; // only for non-control points!!!
		
		if (!AdjObject::lspars.is_eli((*op)->sdev.x))
			{
			sum_var.x += (*op)->sdev.x * (*op)->sdev.x;
			num_x++;
			if ((*op)->sdev.x > sd_max.x) sd_max.x = (*op)->sdev.x;
			}

		if (!AdjObject::lspars.is_eli((*op)->sdev.y))
			{
			sum_var.y += (*op)->sdev.y * (*op)->sdev.y;
			num_y++;
			if ((*op)->sdev.y > sd_max.y) sd_max.y = (*op)->sdev.y;
			}

		if (!AdjObject::lspars.is_eli((*op)->sdev.z))
			{
			sum_var.z += (*op)->sdev.z * (*op)->sdev.z;
			num_z++;
			if ((*op)->sdev.z > sd_max.z) sd_max.z = (*op)->sdev.z;
			}
		}
	os << endl;

	double mean_x = num_x == 0 ? 0. : sqrt(sum_var.x / double(num_x));
	double mean_y = num_y == 0 ? 0. : sqrt(sum_var.y / double(num_y));
	double mean_z = num_z == 0 ? 0. : sqrt(sum_var.z / double(num_z));
	
	os << "Average precision of object coords.\n";
	os << "===================================\n";
	os << "Derived from std. dev. of coordinates of non-control points.\n\n";
	os << "                                   sX/m      sY/m      sZ/m\n\n";
	os << "Number of std. devs.         :  "
		<< int9(num_x) << " " << int9(num_y) << " " << int9(num_z) << endl;
	//os << "Number of non-control points :  " << num << endl;
	os << "Maximum of std. devs.        :  "
		<< fix9(sd_max.x) << " " << fix9(sd_max.y) << " " << fix9(sd_max.z)
		<< endl;
	os << "Average of std. devs.        :  "
		<< fix9(mean_x) << " " << fix9(mean_y) << " " << fix9(mean_z) << endl;
	os << endl;
}


// write residual image coords.

void BundleAdjustment::write_residuals(ostream& os) const
{
	// loop over all images
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0) (*image)->write_residuals(_obj_points, os);
		}
	os << "stop-dep\n";
}


// print residual image coords.

void BundleAdjustment::print_residuals(ostream& os) const
{
	Form fix2(2); fix2.fixed();

	DPoint2 tot_v, tot_vv, tot_pvv, tot_p;
	int tot_num = 0;
	//double vtpv = 0.;

	//os << endl;
	os << "Residuals of image coords.\n";
	os << "==========================\n";
	os << "Limit for printing residuals:  "
		<< fix2(AdjObject::lspars.res_lim() * 1e6) << " um\n";
	os << endl;

	// loop over all images
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image == 0) continue;
	
		DPoint2 sum_v, sum_vv, sum_pvv, sum_p;
    	int num = 0;
	
		(*image)->print_residuals(_obj_points,
			sum_v, sum_vv, sum_pvv, sum_p, num, os);
		
		tot_v += sum_v;
		tot_vv += sum_vv;
		tot_pvv += sum_pvv;
		tot_p += sum_p;
		tot_num += num;
		}
	os << endl;

	fix2.width(10);

	os << "Summary for all image points of all images\n";
	os << "------------------------------------------\n";
	os << "                                   vx/um        vy/um\n\n";
	os << "Total number of points    :   " << tot_num << endl;
	os << "Sum of weights (e'Pe)     :   "
		<< fix2(tot_p.x) << "   "
		<< fix2(tot_p.y) << endl;
	os << "RMS of weighted residuals :   "
		<< fix2(sqrt(tot_pvv.x / tot_p.x) * 1e6) << "   "
		<< fix2(sqrt(tot_pvv.y / tot_p.y) * 1e6) << endl;
	os << "RMS of residuals          :   "
		<< fix2(sqrt(tot_vv.x / tot_num) * 1e6) << "   "
		<< fix2(sqrt(tot_vv.y / tot_num) * 1e6) << endl;
	os << "Average of residuals      :   "
		<< fix2(tot_v.x * 1e6 / tot_num) << "   "
		<< fix2(tot_v.y * 1e6 / tot_num) << endl;
	os << endl;

	fix2.width(0);
	Form sci3(3); sci3.scientific();

	double vtpv = tot_pvv.x + tot_pvv.y;
	double sigma0 = (_f > 0) ? sqrt(vtpv / double(_f)) : (0.);
	os << "Contribution to v'pv and sigma 0:   "
		<< sci3(vtpv) << " m^2   ";
	if (sigma0 == 0.)
		os << "undefined, no redundancy!\n";
	else
		os << fix2(sigma0 * 1e6) << " um\n";
	os << "Redundancy :   " << _f << endl;
	os << endl;
}


// update cameras

void BundleAdjustment::update_cameras(ostream* os)
{
	vector<Camera*>::const_iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		{
		if (*cam == 0) continue;
		
		bool rtn = (*cam)->update_io();
		if (os)
			{
			(*os) << "Camera: " << (*cam)->get_id();
			if (rtn == true)
				(*os) << " updated\n";
			else
				(*os) << " no update!\n";
			}
		}
	if (os) (*os) << endl;
}


// IO and EO parameters calculated from DLT parameters

void BundleAdjustment::print_orient_from_dlt(ostream& os) const
{
	os << "IO and EO from DLT parameters\n";
	os << "=============================\n";
	os << endl;

	// loop over all images
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image == 0) continue;
		
		if (DLTImage* p = dynamic_cast<DLTImage*>(*image))
			p->print_orient_from_dlt(os);
		}
	os << endl;
}


// update image coords.
/*
void BundleAdjustment::update_image_coords(ostream& os) const
{
	// loop over all images
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0) (*image)->update_image_coords(_obj_points, os);
		}
	os << "stop-dep\n";
}
*/

// write initial direction
/*
void BundleAdjustment::write_init_direction(ostream& os)
{
	if (scan_eos.empty())
		cout << "ATTENTION: List of scanner EOs is empty!" << endl;

	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0) (*eo)->write_init_direction(os);
}


// write initial direction

void BundleAdjustment::print_init_direction(ostream& os)
{
	os << "Initial direction of scanner EO\n";
	os << "===============================\n";
	
	if (scan_eos.empty())
		os << "ATTENTION: List of scanner EOs is empty!" << endl;

	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0) (*eo)->print_init_direction(os);

	os << endl;
}
*/


// write cameras

void BundleAdjustment::write_cameras(ostream& os) const
{
	vector<Camera*>::const_iterator cam;
	for (cam = _cameras.begin(); cam != _cameras.end(); ++cam)
		if (*cam != 0) (*cam)->write(os);
	os << "stop-dep\n";
}


// write image coords.
/*
void BundleAdjustment::write_image_coords(ostream& os) const
{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) (*image)->write_image_coords(os);
	os << "stop-dep\n";
}
*/

// write photo coords.
/*
void BundleAdjustment::write_photo_coords(ostream& os) const
{
	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		if (*image != 0) (*image)->write_photo_coords(os);
	os << "stop-dep\n";
}
*/

// write object coords.

void BundleAdjustment::write_object_coords(ostream& os) const
{
	os << "indiv-sdev\n";
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		if (*op != 0) (*op)->write(os);
	os << "stop-dep\n";
}


// write stations

void BundleAdjustment::write_stations(ostream& os) const
{
	os << "indiv-type\n";
	vector<Station*>::const_iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		if (*stat != 0) (*stat)->write(os);
	os << "stop-dep\n";
}


// write AP sets

void BundleAdjustment::write_ap_sets(ostream& os) const
{
	vector<APSet*>::const_iterator ap;
	for (ap = _ap_sets.begin(); ap != _ap_sets.end(); ++ap)
		if (*ap != 0) (*ap)->write(os);
	os << "stop-dep\n";
}


// write IMU misalignment

void BundleAdjustment::write_imu_misalignment(ostream& os) const
{
	vector<IMUMisalignment*>::const_iterator it;
	for (it = _imu_mis.begin(); it != _imu_mis.end(); ++it)
		if (*it != 0) (*it)->write(os);
	os << "stop-dep\n";
}


// write object point analysis

void BundleAdjustment::write_object_point_analysis(ostream& os) const
{
	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op == 0) continue;
		
		// point ID, object point coords.
		os << form("%12s %12.3f %12.3f %12.3f",
			(*op)->get_id().c_str(), (*op)->crds.x, (*op)->crds.y, (*op)->crds.z);
				
		// std. devs.
		if (AdjObject::lspars.is_rsdev((*op)->sdev.x))
			os << form(" %8.3f", (*op)->sdev.x);
		else
			os << form(" %8.1e", (*op)->sdev.x);
				
		if (AdjObject::lspars.is_rsdev((*op)->sdev.y))
			os << form(" %8.3f", (*op)->sdev.y);
		else
			os << form(" %8.1e", (*op)->sdev.y);
				
		if (AdjObject::lspars.is_rsdev((*op)->sdev.z))
			os << form(" %8.3f", (*op)->sdev.z);
		else
			os << form(" %8.1e", (*op)->sdev.z);
		
		// type
		os << form(" %1d", (*op)->get_type());
		
		// calculate control point residuals
		DPoint3 v(0., 0., 0.);
		char cr[] = "000";
		
		if ((*op)->state(0) == obs)
			{
			v.x = (*op)->crds.x - (*op)->ival(0);
			cr[0] = '1';
			}
		
		if ((*op)->state(1) == obs)
			{
			v.y = (*op)->crds.y - (*op)->ival(1);
			cr[1] = '1';
			}
		
		if ((*op)->state(2) == obs)
			{
			v.z = (*op)->crds.z - (*op)->ival(2);
			cr[2] = '1';
			}
		
		// calculate check point differences
		DPoint3 d(0., 0., 0.);
		char cs[] = "000";
		
		if ((*op)->is_chx())
			{
			d.x = (*op)->ival(0) - (*op)->crds.x;
			cs[0] = '1';
			}
		
		if ((*op)->is_chy())
			{
			d.y = (*op)->ival(1) - (*op)->crds.y;
			cs[1] = '1';
			}
		
		if ((*op)->is_chz())
			{
			d.z = (*op)->ival(2) - (*op)->crds.z;
			cs[2] = '1';
			}
		
		os << form(" %3s %8.3f %8.3f %8.3f", cr, v.x, v.y, v.z);
		os << form(" %3s %8.3f %8.3f %8.3f\n", cs, d.x, d.y, d.z);
		}
	os << "stop-dep\n";
}


// update scanner EO (and clear INS corr. and CGI config.)
/*
void BundleAdjustment::update_scan_eo(ostream* os)
{
	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0) (*eo)->update(os);
	
	vector<INSCor*>::iterator icor;
	for (icor = ins_cors.begin(); icor != ins_cors.end(); ++icor)
		if (*icor != 0) (*icor)->clear();
	
	vector<CGICfg*>::iterator cgic;
	for (cgic = cgi_cfgs.begin(); cgic != cgi_cfgs.end(); ++cgic)
		if (*cgic != 0) (*cgic)->clear();
}
*/

// write scanner EO
/*
void BundleAdjustment::write_scan_eo(const char* stem, ostream* out)
{
	if (out)
		{
		(*out) << "Write scanner EO" << endl;
		(*out) << "================" << endl;
		(*out) << "(No replace! Fail if file exists)" << endl;
		(*out) << "stem=" << stem << endl;
		(*out) << endl;
		}
	
	vector<ScanEO*>::iterator eo;
	for (eo = scan_eos.begin(); eo != scan_eos.end(); ++eo)
		if (*eo != 0)
			{
			char fname[DEP_MAXNAME];
			sprintf(fname, "%s_%d.eo", stem, (*eo)->get_id());
			if (out)
				{
				(*out) << "Strip " << (*eo)->get_id() << " -> " << fname << endl;
				(*out).flush();
				}

			ofstream os(fname, ios::out|ios::noreplace);
			if (!os)
				cout << "ATTENTION: Cannot create output file: "
					<< fname << endl;
			else
				(*eo)->write(os);
			}
}
*/


// find time interval

int BundleAdjustment::time_interval(double& t_min, double& t_max) const
{
	int num = 0;

	vector<Station*>::const_iterator stat;
	for (stat = _stations.begin(); stat != _stations.end(); ++stat)
		if (*stat != 0)
			{
			if (ExtOrientation* p = dynamic_cast<ExtOrientation*>(*stat))
				{
				double t = p->get_time();
				if (t < 0.)
					{
					t_min = t_max = t;
					num++;
					break;
					}
				}
			}

	for (; stat != _stations.end(); ++stat)
		if (*stat != 0)
			{
			if (ExtOrientation* p = dynamic_cast<ExtOrientation*>(*stat))
				{
				double t = p->get_time();
				if (t < t_min) t_min = t;
				if (t > t_max) t_max = t;
				num++;
				}
			}
			
	vector<OrientationData*>::const_iterator ods;
	for (ods = _ori_data.begin(); ods != _ori_data.end(); ++ods)
		if (*ods != 0)
			{
			(*ods)->time_interval(t_min, t_max);
			break;
			}
	
	for (; ods != _ori_data.end(); ++ods)
		if (*ods != 0)
			{
			double t1 = 0., tn = 0.; 
			(*ods)->time_interval(t1, tn);
			
			if (t1 < t_min) t_min = t1;
			if (t1 > t_max) t_max = t1;
			if (tn < t_min) t_min = tn;
			if (tn > t_max) t_max = tn;
			}
			
	return num;
}


// set reference time for position and attitude correction

void BundleAdjustment::set_reference_time(double t0)
{
	vector<PosCorrection*>::iterator pcor;
	for (pcor = _pos_corrs.begin(); pcor != _pos_corrs.end(); ++pcor)
		if (*pcor != 0) (*pcor)->set_reference_time(t0);

	vector<AttCorrection*>::iterator acor;
	for (acor = _att_corrs.begin(); acor != _att_corrs.end(); ++acor)
		if (*acor != 0) (*acor)->set_reference_time(t0);
}


// count number of image points

int BundleAdjustment::count_image_points() const
{
	int n = 0;

	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0) n += (*image)->count_image_points();
		}

	return n;
}


// remove image point in image list

int BundleAdjustment::remove_image_point(const string& id, ostream* os)
{
	int removed = 0;

	vector<Image*>::const_iterator image;
	for (image = _images.begin(); image != _images.end(); ++image)
		{
		if (*image != 0 && (*image)->remove_image_point(id))
			{
			if (os)
				(*os) << "ATTENTION: Image " << (*image)->get_id()
					<< " point " << id << " removed.\n";
			removed++;
			}
		}

	return removed;
}


// count number of all object points

int BundleAdjustment::num_obpts() const
{
	int n = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op != 0 && !(*op)->is_unused()) n++;
		}

	return n;
}


// count number of control points

int BundleAdjustment::num_copts() const
{
	int n = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op != 0 && (*op)->is_ctrl()) n++;
		}

	return n;
}


// count number of check points

int BundleAdjustment::num_chpts() const
{
	int n = 0;

	vector<ObjectPoint*>::const_iterator op;
	for (op = _obj_points.begin(); op != _obj_points.end(); ++op)
		{
		if (*op != 0 && (*op)->is_check()) n++;
		}

	return n;
}


// solve normal equation system

bool BundleAdjustment::solve()
{
	assert(_adj_interface != 0);
	
	int info = _adj_interface->solve(_atpa, _atpl, _dx, _dec_vec);
	assert(info >= 0);
	
	if (info > 0)
		{
		int idx = info - 1;
		cout << "ATTENTION: Matrix is >>> not <<< positive definite at row/column "
			<< idx << endl;
		AdjObject* p = _list.find(idx);
		assert(p != 0);
		cout << "class/type : " << p->class_str();
		string str = p->type_str();
		if (!str.empty()) cout << ", " << str;
		cout << endl;
		cout << "ID         : " << p->get_id() << endl;
		cout << "parameter  : " << p->uname_idx(idx) << endl;
		cout << endl;
		return false;
		}
	
	return true;
}


// inversion of normal matrix

bool BundleAdjustment::inversion()
{
	assert(_adj_interface != 0);
	
	int info = _adj_interface->inversion(_atpa, _qxx, _dec_vec);
	assert(info == 0);
	
	return true;
}


// write structure of normal matrix in PBM-format

void BundleAdjustment::write_normal_matrix_struct(ostream& os) const
{
	assert(_adj_interface != 0);
	_adj_interface->write_matrix_struct(os, _atpa);
}


// write elements of normal matrix in ASCII-format

void BundleAdjustment::write_normal_matrix(ostream& os) const
{
	//Form sci(12); sci.scientific();
	os.setf(ios::scientific, ios::floatfield);
	os.precision(12);

	os << _atpa.rows() << " " << _atpa.cols() << endl;
	for (int i = 0; i < _atpa.rows(); i++)
		{
		for (int j = 0; j < _atpa.cols(); j++)
			os << _atpa(i,j) << " ";
			//os << sci(_atpa(i,j)) << " ";
		os << endl;
		}
}


// write elements of right side vector in ASCII-format

void BundleAdjustment::write_rhs_vector(ostream& os) const
{
	//Form sci(12); sci.scientific();
	os.setf(ios::scientific, ios::floatfield);
	os.precision(12);

	os << _atpl.size() << endl;
	for (int i = 0; i < _atpl.size(); i++)
		os << _atpl(i) << endl;
		//os << sci(_atpl(i)) << endl;
}


// write elements of solution vector in ASCII-format

void BundleAdjustment::write_solution_vector(ostream& os) const
{
	//Form sci(12); sci.scientific();
	os.setf(ios::scientific, ios::floatfield);
	os.precision(12);

	os << _dx.size() << endl;
	for (int i = 0; i < _dx.size(); i++)
		os << _dx(i) << endl;
		//os << sci(_dx(i)) << endl;
}

} // namespace DEP
