/* $Id: dlt.cpp,v 1.12 2008/01/08 16:37:46 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <typeinfo>
#include "dlt.h"
#include "depvec.h"
#include "camera.h"
#include "lconeqn.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "cholesky.h"

namespace DEP
{

// get type name

string DLTParameters::type_str() const
{
	ostringstream oss;
	oss << "DLT parameters";
	return oss.str();
}


// get key name

string DLTParameters::key_str() const
{
	ostringstream oss;
	oss << "DLT-params";
	return oss.str();
}


// get parameter name

string DLTParameters::uname(size_t i) const
{
	assert(i < _size);

	ostringstream oss;
	oss << "L" << (i + 1);
	
	return oss.str();
}


// object-to-photo coords. transformation

bool DLTParameters::object_to_image(const DPoint3& obj_c, DPoint2& img_c) const
{
	dlt_projection(_a, obj_c, img_c);
	
	return true;
}


// IO and EO parameters calculated from DLT parameters

bool DLTParameters::calc_orient_from_dlt(DPoint2& pp, DPoint2& ck, DPoint3& pcc, DRotMatrix3& R) const
{
	double ll = 1. / (_a[8] * _a[8] + _a[9] * _a[9] + _a[10] * _a[10]);
	double l = sqrt(ll);
	
	pp.x = (_a[0] * _a[8] + _a[1] * _a[9] + _a[2] * _a[10]) * ll;
	pp.y = (_a[4] * _a[8] + _a[5] * _a[9] + _a[6] * _a[10]) * ll;
	
	ck.x = sqrt((_a[0] * _a[0] + _a[1] * _a[1] + _a[2] * _a[2]) * ll - pp.x * pp.x);
	ck.y = sqrt((_a[4] * _a[4] + _a[5] * _a[5] + _a[6] * _a[6]) * ll - pp.y * pp.y);
	
	double r31 = -l * _a[8];
	double r32 = -l * _a[9];
	double r33 = -l * _a[10];

	// transposed required!
	R(0,0) = (l * _a[0] + pp.x * r31) / ck.x;
	R(1,0) = (l * _a[1] + pp.x * r32) / ck.x;
	R(2,0) = (l * _a[2] + pp.x * r33) / ck.x;
	
	R(0,1) = (l * _a[4] + pp.y * r31) / ck.y;
	R(1,1) = (l * _a[5] + pp.y * r32) / ck.y;
	R(2,1) = (l * _a[6] + pp.y * r33) / ck.y;
	
	R(0,2) = r31;
	R(1,2) = r32;
	R(2,2) = r33;

	double A[3][3];
	A[0][0] = _a[0]; A[0][1] = _a[1]; A[0][2] = _a[2];
	A[1][0] = _a[4]; A[1][1] = _a[5]; A[1][2] = _a[6];
	A[2][0] = _a[8]; A[2][1] = _a[9]; A[2][2] = _a[10];
	
	double det = sarrus(A);
	if (det == 0.)
		{
		cout << "ERROR: Determinate is equal zero." << endl;
		return false;
		}

	double A1[3][3];
	A1[0][0] = -_a[3]; A1[0][1] = _a[1]; A1[0][2] = _a[2];
	A1[1][0] = -_a[7]; A1[1][1] = _a[5]; A1[1][2] = _a[6];
	A1[2][0] =    -1.; A1[2][1] = _a[9]; A1[2][2] = _a[10];
	
	double det1 = sarrus(A1);
	
	double A2[3][3];
	A2[0][0] = _a[0]; A2[0][1] = -_a[3]; A2[0][2] = _a[2];
	A2[1][0] = _a[4]; A2[1][1] = -_a[7]; A2[1][2] = _a[6];
	A2[2][0] = _a[8]; A2[2][1] =    -1.; A2[2][2] = _a[10];
		
	double det2 = sarrus(A2);
	
	double A3[3][3];
	A3[0][0] = _a[0]; A3[0][1] = _a[1]; A3[0][2] = -_a[3];
	A3[1][0] = _a[4]; A3[1][1] = _a[5]; A3[1][2] = -_a[7];
	A3[2][0] = _a[8]; A3[2][1] = _a[9]; A3[2][2] =    -1.;
	
	double det3 = sarrus(A3);
	
	pcc.x = det1 /det;
	pcc.y = det2 /det;
	pcc.z = det3 /det;
	
	return true;
}


// Sarrus formular

double DLTParameters::sarrus(double a[3][3]) const
{
	double det = a[0][0] * a[1][1] * a[2][2];
	
	det += a[0][1] * a[1][2] * a[2][0];
	det += a[0][2] * a[1][0] * a[2][1];
	det -= a[0][2] * a[1][1] * a[2][0];
	det -= a[0][0] * a[1][2] * a[2][1];
	det -= a[0][1] * a[1][0] * a[2][2];
	
	return det;
}


// constructor

DLTImage::DLTImage() : Image()
{
	// empty
}


// get image type

string DLTImage::type_str() const
{
	ostringstream oss;
	oss << "DLT image";
	return oss.str();
}


// object-to-image coords. transformation (UNUSED ARGUMENTS: r)

bool DLTImage::object_to_image(const DPoint3& obj_c, DPoint2& img_c, double r) const
{
	assert(_station != 0);
	assert(_camera != 0);
	
	const TransfParams* trf = dynamic_cast<const TransfParams*>(_station);
	assert(trf != 0);
	
	DPoint2 tmp_c;
	trf->object_to_image(obj_c, tmp_c);
	
	DPoint3 pho_c(tmp_c.x, tmp_c.y, _camera->ck());
	_camera->photo_to_image(pho_c, img_c);

	return true;
}


// resection using the direct solution approach

bool DLTImage::resection(ostream& os, const vector<ObjectPoint*>& obpts, bool dbg)
{
	assert(_station != 0);	// necessary
	assert(_camera != 0);	// necessary
	
	os << "Image ID   : " << _image_id << endl;
	os << "Station ID : " << _station->get_id() << endl;
	os << endl;
	os << "Point ID :" << endl;

	DLTParameters* trf = dynamic_cast<DLTParameters*>(_station);
	assert(trf != 0);
	
	int nu = trf->num_unk();
	DMatrix atpa(nu, nu);
	DVector atpl(nu);
	double ltpl = 0.;

	// submatrices for ...
	DerivSub	deriv_eo(2,nu);		// transf. params, size always 2x11
	//vector<DerivSub> deriv_ap(_camera->num_ap_sets());	// APs, variable size
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2

	// derivatives of observations
	deriv_obs.identity();
	
	// set index for transf. params.
	deriv_eo.index(0);

	// always 2 condition equations for a single imaging ray
	LinConditionEqn con_eqn;

	// assign A-submatrix
	con_eqn.add_a(&deriv_eo);
	//con_eqn.add_a(deriv_ap);	// optional
	
	// assign B-submatrix
	con_eqn.set_b(deriv_obs);
	
	// co-factors: set all weights to be equal!
	DMatrix Qll(2,2);
	Qll(0,0) = 1.;
	Qll(1,1) = 1.;
	con_eqn.set_q(Qll);
	
	DVector	wt(2);				// misclosure vector
	int num_rays = 0;
	vector<ImagePoint*> imgpts; // image points in use
	
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			os << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}
			
		if (op->is_approx() == false) continue;
		
		if (num_rays % 4 == 0) os << endl;
		os << "  " << op->get_id();
		imgpts.push_back(*ip);

		// convert sensor into (observed) image coords. 
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;
		
		// convert sensor to image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

// 		// co-factors
// 		Matrix<double> Qll(2,2);
// 		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
// 		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
// 		con_eqn.set_q(Qll);
// 		//cout << "Qll: " << Qll << endl;
		
		// coefficients
		deriv_eo(0, 0) = op->crds.x;
		deriv_eo(0, 1) = op->crds.y;
		deriv_eo(0, 2) = op->crds.z;
		deriv_eo(0, 3) = 1.;
		deriv_eo(0, 4) = 0.;
		deriv_eo(0, 5) = 0.;
		deriv_eo(0, 6) = 0.;
		deriv_eo(0, 7) = 0.;
		deriv_eo(0, 8) = -img_c.x * op->crds.x;
		deriv_eo(0, 9) = -img_c.x * op->crds.y;
		deriv_eo(0,10) = -img_c.x * op->crds.z;
	
		deriv_eo(1, 0) = 0.;
		deriv_eo(1, 1) = 0.;
		deriv_eo(1, 2) = 0.;
		deriv_eo(1, 3) = 0.;
		deriv_eo(1, 4) = op->crds.x;
		deriv_eo(1, 5) = op->crds.y;
		deriv_eo(1, 6) = op->crds.z;
		deriv_eo(1, 7) = 1.;
		deriv_eo(1, 8) = -img_c.y * op->crds.x;
		deriv_eo(1, 9) = -img_c.y * op->crds.y;
		deriv_eo(1,10) = -img_c.y * op->crds.z;

// 		// partial derivatives of APs
// 		{
// 			DPoint3 pho_c;
// 			_camera->image_to_photo(img_c, pho_c, false);
// 			_camera->get_derivs(pho_c, deriv_ap);
// 		}
		
		// calculate shorten observations (observed - computed term)
		wt(0) = img_c.x;
		wt(1) = img_c.y;
		con_eqn.set_w(wt);
		//cout << "w: " << wt << endl; cout.flush();

		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return false;
		//con_eqn.print(cout); cout << endl;
		
		num_rays++;
		}
	os << endl << endl;

	int num_obs = 2 * num_rays;
	os << "Number of rays : " << num_rays << endl;
	if (num_obs < nu)
		{
		os << "ATTENTION: Insufficient control points!" << endl;
		return false;
		}
	os << endl;
	
	// solve normal equation sytem
	DVector dec;
	int info = chol_dec(atpa, dec);
	if (info != 0)
		{
		os << "Decomposition of normal matrix failed in row "
			<< info << endl;
		return false;
		}

	DVector x;
	info = chol_sol(atpa, dec, atpl, x);
	if (info != 0)
		{
		os << "Backsubstitution of normal equation system failed: "
			<< "info = " << info << endl;
		return false;
		}
		
	double xtatpl = dot_prod(x, atpl);
	double vtpv = ltpl - xtatpl;
	
	int f = num_obs - nu;
	assert(f > 0);
	
	double sigma0 = sqrt(vtpv / double(f));
	
	Form fix2(2); fix2.fixed();
	os << "Number of observations  :  " << num_obs << endl;
	os << "Number of unknowns      :  " << nu << endl;
	os << "Redundancy              :  " << f << endl;
	os << "Sigma 0 a posteriori    :  " << fix2(sigma0 * 1e6) << " um\n";
	os << endl;
	
	// set DLT parameters
	for (int i = 0; i < nu; i++)
		{
		trf->par(i, x(i));
		//cout << "L" << i << " = " << x(i) << endl;
		}
	trf->print(os);
	
	os << "Differences in image points\n\n";
	os << "  Point ID       dx/um      dy/um     |d|/um\n\n";
	
	Form str8; str8.width(8);
	Form fix10(2); fix10.fixed().width(10);
	
	DPoint2 sum_d;
	DPoint2 sum_dd;
	int num = 0;
	
	DPoint2 img_c0;
	DPoint2 d, dd;
	double d_abs = 0.;
	
	// loop over used image points
	for (ip = imgpts.begin(); ip != imgpts.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			os << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		// Attention: The scanline coordinate is only used by scanner images!!!
		DPoint2 img_c0;
		object_to_image(op->crds, img_c0);

		// convert sensor into image coords.
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);

		// difference
		d = img_c0 - img_c;
		d_abs = abs(d);
		dd = d * d;

		// convert sensor into image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		
		sum_d += d;
		sum_dd += dd;
		num++;

		// print differences
		if (AdjObject::lspars.is_small(d.x) || AdjObject::lspars.is_small(d.y))
			{
			os << "  " << str8((*ip)->get_id()) << "  "
				<< fix10(d.x * 1e6) << " " << fix10(d.y * 1e6) << " "
				<< fix10(d_abs * 1e6);
			os << endl;
			}
		}
	os << endl;

	os << "Number of points          :   " << num << endl;
	os << "RMS of differences        :   "
		<< fix10(sqrt(sum_dd.x / num) * 1e6) << "   "
		<< fix10(sqrt(sum_dd.y / num) * 1e6) << endl;
	os << "Average of differences    :   "
		<< fix10(sum_d.x * 1e6 / num) << "   "
		<< fix10(sum_d.y * 1e6 / num) << endl;
	os << endl;
	
	// precision of transformation parameters missing!!!
	
	return true;
}


// observation equation for intersection

bool DLTImage::observ_eqn_intersec(const string& id, DerivSub& a, DVector& wt,
bool& found) const
{
	ImagePoint* ip = get_image_point(id);
	if (ip == 0)
		{
		found = false;
		return true; // escape, point not found
		}
	found = true;

	DPoint3 pho_c;
	sensor_to_photo(ip->crds, pho_c);
	
	const TransfParams* trf = dynamic_cast<const TransfParams*>(_station);
	assert(trf != 0);
	const double* l = trf->a();
	
	a(0,0) = pho_c.x * l[8] - l[0];
	a(0,1) = pho_c.x * l[9] - l[1];
	a(0,2) = pho_c.x * l[10] - l[2];

	a(1,0) = pho_c.y * l[8] - l[4];
	a(1,1) = pho_c.y * l[9] - l[5];
	a(1,2) = pho_c.y * l[10] - l[6];

	wt(0) = l[3] - pho_c.x;
	wt(1) = l[7] - pho_c.y;

	return true;
}


// add contribution to normal equations for all image point observations

bool DLTImage::normal_eqns(const vector<ObjectPoint*>& obpts,
Matrix<double>& atpa, Vector<double>& atpl, double& ltpl, int& num_obs) const
{
	assert(_station != 0);	// necessary
	assert(_camera != 0);	// necessary
	
	const TransfParams* trf = dynamic_cast<const TransfParams*>(_station);
	assert(trf != 0);

	// Setup sub-matrices of derivatives
	// Always 2 observation equations for each image point.
	// The number of columns might be variable.
	DerivSub	deriv_oc(2,3);		// object points, size always 2x3
	DerivSub	deriv_eo(2,11);		// transf. params, size always 2x11
	vector<DerivSub> deriv_ap(_camera->num_ap_sets());	// APs, variable size
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2

	// derivatives of observations
	deriv_obs.identity();
	
	// set index for EO (valid for the whole image)
	deriv_eo.index(trf->get_index());

	// always 2 condition equations for a single imaging ray
	LinConditionEqn con_eqn;

	// design matrix A
	con_eqn.add_a(&deriv_oc);	// always
	con_eqn.add_a(&deriv_eo);	// always
	con_eqn.add_a(deriv_ap);	// optional
	
	// design matrix B
	con_eqn.set_b(deriv_obs);
	
	// DLT parameters
	const double* l = trf->a();
	
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}

		//cout << _image_id << ":" << (*ip)->get_id() << ":\n"; cout.flush();
		
		// set current index of object point
		deriv_oc.index(op->get_index());

		// convert sensor to image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

		// co-factors
		Matrix<double> Qll(2,2);
		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
		con_eqn.set_q(Qll);
		//cout << "Qll: " << Qll << endl;
		
		// denominator
		double q = l[8] * op->crds.x + l[9] * op->crds.y + l[10] * op->crds.z + 1.;
		if (fabs(q) == 0.)
			{
			cout << "ERROR: DLT image::normal_eqns: Zero denominator.\n";
			return false;
			}
		
		// nominators
		double m = l[0] * op->crds.x + l[1] * op->crds.y + l[2] * op->crds.z + l[3];
		double n = l[4] * op->crds.x + l[5] * op->crds.y + l[6] * op->crds.z + l[7];

		// image coords. derived from unknowns.
		// Use these values not the observations!!!
		// See personal notes.
		double x = m / q;
		double y = n / q;

		// derivatives
		deriv_eo(0, 0) = op->crds.x / q;
		deriv_eo(0, 1) = op->crds.y / q;
		deriv_eo(0, 2) = op->crds.z / q;
		deriv_eo(0, 3) = 1. / q;
		deriv_eo(0, 4) = 0.;
		deriv_eo(0, 5) = 0.;
		deriv_eo(0, 6) = 0.;
		deriv_eo(0, 7) = 0.;
		deriv_eo(0, 8) = -x * op->crds.x / q;
		deriv_eo(0, 9) = -x * op->crds.y / q;
		deriv_eo(0,10) = -x * op->crds.z / q;

		deriv_eo(1, 0) = 0.;
		deriv_eo(1, 1) = 0.;
		deriv_eo(1, 2) = 0.;
		deriv_eo(1, 3) = 0.;
		deriv_eo(1, 4) = op->crds.x / q;
		deriv_eo(1, 5) = op->crds.y / q;
		deriv_eo(1, 6) = op->crds.z / q;
		deriv_eo(1, 7) = 1. / q;
		deriv_eo(1, 8) = -y * op->crds.x / q;
		deriv_eo(1, 9) = -y * op->crds.y / q;
		deriv_eo(1,10) = -y * op->crds.z / q;

		deriv_oc(0,0) = (l[0] - x * l[8]) / q;
		deriv_oc(0,1) = (l[1] - x * l[9]) / q;
		deriv_oc(0,2) = (l[2] - x * l[10]) / q;
	
		deriv_oc(1,0) = (l[4] - y * l[8]) / q;
		deriv_oc(1,1) = (l[5] - y * l[9]) / q;
		deriv_oc(1,2) = (l[6] - y * l[10]) / q;

		// image coords. derived from the unknown parameters
		DPoint2 tmp_c(x, y);
		DPoint3 pho_c;
		_camera->image_to_photo(tmp_c, pho_c, false);

		// partial derivatives of APs
		_camera->get_derivs(pho_c, deriv_ap);

		// convert sensor into (observed) image coords. 
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;

		// AP correction
		DPoint2 corr = _camera->ap_set_corr(pho_c);
		
		// computed observations
		DPoint2 img_c0(x + corr.x, y + corr.y);
		//cout << "img_c0: " << img_c0 << endl;

		// calculate shorten observations (observed - computed term)
		// misclosure vector
		Vector<double> wt(2);
		wt(0) = img_c.x - img_c0.x;
		wt(1) = img_c.y - img_c0.y;
		con_eqn.set_w(wt);
		//cout << "w: " << wt << endl; cout.flush();

		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return false;
		//con_eqn.print(cout); cout << endl;
		
		num_obs += 2;
		}
		
	return true;
}


// IO and EO parameters calculated from DLT parameters

void DLTImage::print_orient_from_dlt(ostream& os) const
{
	const DLTParameters* p = dynamic_cast<const DLTParameters*>(_station);
	assert(p != 0);
	
	DPoint2 pp, ck;
	DPoint3 pcc;
	DRotMatrix3 R;
	p->calc_orient_from_dlt(pp, ck, pcc, R);
	
	Form fix4(4); fix4.fixed().width(10);
	Form fix7(7); fix7.fixed().width(12);

	os << "Image ID           :  " << _image_id << endl;
	os << "Principal distance :  cx = " << fix4(ck.x * 1e3) << " mm"
		"      cy = " << fix4(ck.y * 1e3) << " mm\n";
	os << "Principal point    :  xp = " << fix4(pp.x * 1e3) << " mm"
		"      yp = " << fix4(pp.y * 1e3) << " mm\n";
	os << endl;
	os << "Perspective center :  Xo = " << fix4(pcc.x) << " m "
		"      Yo = " << fix4(pcc.y) << " m "
		"      Zo = " << fix4(pcc.z) << " m\n";
	os << "Rotation matrix: R =\n";
	os << " " << fix7(R(0,0)) << fix7(R(0,1)) << fix7(R(0,2)) << endl;
	os << " " << fix7(R(1,0)) << fix7(R(1,1)) << fix7(R(1,2)) << endl;
	os << " " << fix7(R(2,0)) << fix7(R(2,1)) << fix7(R(2,2)) << endl;
	os << endl;
}


// non-member functions ...

// project object point onto image plane using DLT

void dlt_projection(const double* l, const DPoint3& obj_c, DPoint2& img_c)
{
	assert(l != 0);

	// denominator
	double q = l[8] * obj_c.x + l[9] * obj_c.y + l[10] * obj_c.z + 1.;
	if (fabs(q) == 0.) cout << "ERROR: dlt_projection: zero denominator.\n";

	// nominators
	double m = l[0] * obj_c.x + l[1] * obj_c.y + l[2] * obj_c.z + l[3];
	double n = l[4] * obj_c.x + l[5] * obj_c.y + l[6] * obj_c.z + l[7];

	// image coords.
	img_c.x = m / q;
	img_c.y = n / q;
}

} // namespace DEP
