/* $Id: eopokrot.cpp,v 1.16 2007/10/11 13:10:52 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "eopokrot.h"

namespace DEP
{

// get type name

string EOAnglesPOKRot::type_str() const
{
	ostringstream oss;
	oss << "phi-omega-kappa, rotated, ccw";
	return oss.str();
}


// get key name

string EOAnglesPOKRot::key_str() const
{
	ostringstream oss;
	oss << "ext-ori-pok-rot";
	return oss.str();
}


// get parameter name

string EOAnglesPOKRot::uname(size_t i) const
{
	ostringstream oss;
	
	switch (i)
		{
		case 0: oss << "Xo"; break;
		case 1: oss << "Yo"; break;
		case 2: oss << "Zo"; break;
		case 3: oss << "phi"; break;
		case 4: oss << "omega"; break;
		case 5: oss << "kappa"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}


// read EO parameters

bool EOAnglesPOKRot::read(istream& is)
{
	// station ID
	is >> _id;
	if (!is)
		{
		cout << "ERROR: Reading ID failed.\n";
		return false;
		}

	// unit of rotation angles and/or standard deviations
	string str;
	is >> str;
	if (!is)
		{
		cout << "ERROR: Reading unit of rotation angles failed.\n";
		return false;
		}

	UnitAngle u_angle;
	if (!u_angle.set(str.c_str()))
		{
		cout << "ERROR: Unknown unit of rotation angles:"
			<< str << endl;
		return false;
		}
	double rho = u_angle.rho();

	// exposure time
	is >> _time;
	if (!is)
		{
		cout << "ERROR: Cannot read exposure time.\n";
		return false;
		}

	// projection centre coords.
	is >> _per_centre;
	if (!is)
		{
		cout << "ERROR: Reading projection centre coords. failed.\n";
		return false;
		}

	// std. dev. projection centre coords.
	is >> _sdev_pcc;
	if (!is)
		{
		cout << "ERROR: Reading std. dev. of proj. centre coords. failed.\n";
		return false;
		}

	// rotation parameters
	is >> _a[0] >> _a[1] >> _a[2];
	if (!is)
		{
		cout << "ERROR: Reading rotation parameters failed.\n";
		return false;
		}

	// conversion into radian
	_a[0] /= rho;
	_a[1] /= rho;
	_a[2] /= rho;

	// std. dev. of rotation parameters
	is >> _sdev_a[0] >> _sdev_a[1] >> _sdev_a[2];
	if (!is)
		{
		cout << "ERROR: Reading std. dev. of rotation angles failed.\n";
		return false;
		}

	// only for real std. devs. conversion into radians
	if (is_rsdev(_sdev_a[0])) _sdev_a[0] /= rho;
	if (is_rsdev(_sdev_a[1])) _sdev_a[1] /= rho;
	if (is_rsdev(_sdev_a[2])) _sdev_a[2] /= rho;
	
	_approx_flag = true;

	return true;
}


// write EO parameters

void EOAnglesPOKRot::write(ostream& os) const
{
	//double sc = AdjObject::lspars.scale_obj();
	const double sc = 1.;
	const double rho = AdjObject::lspars.rho();

	Form sci(6); sci.scientific().width(15);
	Form fix(1); fix.fixed().width(15);

	// type
	os << key_str() << endl;

	// ID
	os << _id << " ";

	// unit of rotation angles
	os << AdjObject::lspars.u_angle_cstr() << " ";

	// exposure time
	fix.precision(4);
	os << fix(_time) << endl;

	// projection centre
	os << fix(_per_centre.x * sc) << " "
		<< fix(_per_centre.y * sc) << " "
		<< fix(_per_centre.z * sc) << endl;

	if (is_rsdev(_sdev_pcc.x))
		os << fix(_sdev_pcc.x * sc);
	else
		os << sci(_sdev_pcc.x);
	os << " ";

	if (is_rsdev(_sdev_pcc.y))
		os << fix(_sdev_pcc.y * sc);
	else
		os << sci(_sdev_pcc.y);
	os << " ";
	
	if (is_rsdev(_sdev_pcc.z))
		os << fix(_sdev_pcc.z * sc);
	else
		os << sci(_sdev_pcc.z);
	os << endl;

	// rotation parameters
	os << sci(_a[0] * rho) << " "
		<< sci(_a[1] * rho) << " "
		<< sci(_a[2] * rho) << endl;
		
	if (is_rsdev(_sdev_a[0]))
		os << sci(_sdev_a[0] * rho);
	else
		os << sci(_sdev_a[0]);
	os << " ";
	
	if (is_rsdev(_sdev_a[1]))
		os << sci(_sdev_a[1] * rho);
	else
		os << sci(_sdev_a[1]);
	os << " ";
	
	if (is_rsdev(_sdev_a[2]))
		os << sci(_sdev_a[2] * rho);
	else
		os << sci(_sdev_a[2]);
	os << endl;
		
	os << endl;
}


// print parameters

void EOAnglesPOKRot::print(ostream& os) const
{
	//double sc = AdjObject::lspars.scale_obj();
	const double sc = 1.;
	const double rho = AdjObject::lspars.rho();

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(4); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Exposure time :  " << fix(_time) << " s\n";
	os << "Subtype       :  " << type_str() << endl;
	//os << "Angular unit  :  " << AdjObject::lspars.u_angle_cstr() << endl;
	os << endl;

	os << "  Param.     State         Value       Std.dev.\n\n";

	fix.precision(3);
	os << "  " << str8(uname(0)) << "   "
		<< str4(state_str(_sdev_pcc.x)) << "   "
		<< fix(_per_centre.x * sc) << "   ";
	if (is_rsdev(_sdev_pcc.x))
		os << fix(_sdev_pcc.x * sc);
	else
		os << sci(_sdev_pcc.x);
	os << endl;
	
	os << "  " << str8(uname(1)) << "   "
		<< str4(state_str(_sdev_pcc.y)) << "   "
		<< fix(_per_centre.y * sc) << "   ";
	if (is_rsdev(_sdev_pcc.y))
		os << fix(_sdev_pcc.y * sc);
	else
		os << sci(_sdev_pcc.y);
	os << endl;
	
	os << "  " << str8(uname(2)) << "   "
		<< str4(state_str(_sdev_pcc.z)) << "   "
		<< fix(_per_centre.z * sc) << "   ";
	if (is_rsdev(_sdev_pcc.z))
		os << fix(_sdev_pcc.z * sc);
	else
		os << sci(_sdev_pcc.z);
	os << endl;

	fix.precision(5);
	for (size_t i = 0, k = 3; i < _a.size(); i++, k++)
		{
		os << "  " << str8(uname(k)) << "   "
			<< str4(state_str(_sdev_a[i])) << "   "
			<< fix(_a[i] * rho) << "   ";
		if (is_rsdev(_sdev_a[i]))
			os << fix(_sdev_a[i] * rho);
		else
			os << sci(_sdev_a[i]);
		os << endl;
		}
	os << endl;

	DRotMatrix3 R;
	calc_rot_matrix(R, _a);

	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// print changes of unknowns

void EOAnglesPOKRot::print_changes(ostream& os,
const Vector<double>& dx) const
{
	//double sc = AdjObject::lspars.scale_obj();
	const double sc = 1.;
	const double rho = AdjObject::lspars.rho();

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(6); sci.scientific().width(12);

	os << "Station ID :  " << _id << endl;
	os << "Subtype    :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	//os << "Angle mode :  " << AdjObject::lspars.u_angle_str() << endl;
	os << endl;
 
	os << "  Param.     State"
		"       Approx.         Change       Adjusted\n\n";
	
	size_t i = 0;
	int k = _index;

	for (; i < 3; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i) * sc) << "   "
			<< fix(dx(k) * sc) << "   "
			<< fix((par(i) + dx(k)) * sc) << endl;
		}

	fix.precision(5);
	for (; i < _size; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i) * rho) << "   "
			<< fix(dx(k) * rho) << "   "
			<< fix(((par(i) + dx(k)) * rho)) << endl;
		}
	os << endl;
}


// print unknown parameters

void EOAnglesPOKRot::print_unknowns(ostream& os) const 
{
	//double sc = AdjObject::lspars.scale_obj();
	const double sc = 1.;
	const double rho = AdjObject::lspars.rho();
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Exposure time :  " << fix(_time) << " s\n";
	os << "Subtype       :  " << type_str() << endl;
	os << "Index         :  " << _index << endl;
	os << endl;

	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.\n\n";

	size_t i = 0;
	
	for (; i < 3; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i) * sc) << "  "
			<< fix(par(i) * sc) << "  "
			<< fix(change(i) * sc) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i) * sc);
		else
			os << sci(sdev(i));
		os << endl;
		}
	
	fix.precision(5);
	for (; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i) * rho) << "  "
			<< fix(par(i) * rho) << "  "
			<< fix(change(i) * rho) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i) * rho);
		else
			os << sci(sdev(i));
		os << endl;
		}
	os << endl;

	DRotMatrix3 R;
	calc_rot_matrix(R, _a);

	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// calculate partial derivatives of rotation parameters

void EOAnglesPOKRot::calc_derivs_rot(DerivSub& deriv,
double h, double u, double v, double w,
const DPoint3& d, const vector<double>& a) const
{
	assert(a.size() == 3);
	
	DRotMatrix3 dR_a1;
	DRotMatrix3 dR_a2;
	DRotMatrix3 dR_a3;
	dr(a[0], a[1], a[2], dR_a1, dR_a2, dR_a3);
	
	DPoint3 dd_a1; mult_inv(dd_a1, dR_a1, d);
	DPoint3 dd_a2; mult_inv(dd_a2, dR_a2, d);
	DPoint3 dd_a3; mult_inv(dd_a3, dR_a3, d);
	
	deriv(0,0) = h * (w * dd_a1.x - u * dd_a1.z);
	deriv(1,0) = h * (w * dd_a1.y - v * dd_a1.z);
	
	deriv(0,1) = h * (w * dd_a2.x - u * dd_a2.z);
	deriv(1,1) = h * (w * dd_a2.y - v * dd_a2.z);
	
	deriv(0,2) = h * (w * dd_a3.x - u * dd_a3.z);
	deriv(1,2) = h * (w * dd_a3.y - v * dd_a3.z);
}


// calculate partial derivatives of rotation parameters

void EOAnglesPOKRot::calc_derivs_rot(DerivSub& deriv,
double h, double u, double v, double w,
const DPoint3& d, const DRotMatrix3& M, const vector<double>& a) const
{
	assert(a.size() == 3);
	
	DRotMatrix3 dR_a1;
	DRotMatrix3 dR_a2;
	DRotMatrix3 dR_a3;
	dr(a[0], a[1], a[2], dR_a1, dR_a2, dR_a3);
	
	DRotMatrix3 dN_a1 = dR_a1 * M;
	DRotMatrix3 dN_a2 = dR_a2 * M;
	DRotMatrix3 dN_a3 = dR_a3 * M;
	
	DPoint3 dd_a1; mult_inv(dd_a1, dN_a1, d);
	DPoint3 dd_a2; mult_inv(dd_a2, dN_a2, d);
	DPoint3 dd_a3; mult_inv(dd_a3, dN_a3, d);
	
	deriv(0,0) = h * (w * dd_a1.x - u * dd_a1.z);
	deriv(1,0) = h * (w * dd_a1.y - v * dd_a1.z);
	
	deriv(0,1) = h * (w * dd_a2.x - u * dd_a2.z);
	deriv(1,1) = h * (w * dd_a2.y - v * dd_a2.z);
	
	deriv(0,2) = h * (w * dd_a3.x - u * dd_a3.z);
	deriv(1,2) = h * (w * dd_a3.y - v * dd_a3.z);
}


// non-member functions ...

} // namespace DEP
