/* $Id: eoquaternion.cpp,v 1.11 2007/10/11 13:20:55 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "eoquaternion.h"
#include "lobseqn.h"

namespace DEP
{

// get type name

string EOQuaternion::type_str() const
{
	ostringstream oss;
	oss << "unit quaternion";
	return oss.str();
}


// get key name

string EOQuaternion::key_str() const
{
	ostringstream oss;
	oss << "ext-ori-quaternion";
	return oss.str();
}


// get parameter name

string EOQuaternion::uname(size_t i) const
{
	ostringstream oss;
	
	switch (i)
		{
		case 0: oss << "Xo"; break;
		case 1: oss << "Yo"; break;
		case 2: oss << "Zo"; break;
		case 3: oss << "q0"; break;
		case 4: oss << "q1"; break;
		case 5: oss << "q2"; break;
		case 6: oss << "q3"; break;
		default:
			assert(false);
			break;
		}

	return oss.str();
}

/*
// 

void EOQuaternion::init()
{
	// quaternion norm
	double n = _a[0] * _a[0] + _a[1] * _a[1]
		- _a[2] * _a[2] + _a[3] * _a[3];
		
	double b = sqrt(n);
	cout << "|q| - 1 = " << (b - 1.) << endl;
	
// 	_a[0] /= b; _a[1] /= b; _a[2] /= b; _a[3] /= b;
// 	
// 	n = _a[0] * _a[0] + _a[1] * _a[1]
// 		- _a[2] * _a[2] + _a[3] * _a[3];
// 		
// 	b = sqrt(n);
// 	cout << "|q| - 1 = " << (b - 1.) << endl;
}*/


// add weighted constraints

void EOQuaternion::add_constraints(Matrix<double>& atpa,
Vector<double>& atpl, double& ltpl) const
{
	// one observation equation
	LinObsEqn eqn(1);

	// setup sub-matrices of derivatives
	DerivSub deriv(1, _size);

	// set index
	deriv.index(_index);
	
	// design matrix A
	deriv(0,0) = 0.;
	deriv(0,1) = 0.;
	deriv(0,2) = 0.;
	deriv(0,3) = 2. * _a[0];
	deriv(0,4) = 2. * _a[1];
	deriv(0,5) = 2. * _a[2];
	deriv(0,6) = 2. * _a[3];
	eqn.add_a(&deriv);

	// weight
	eqn.set_p(0, lspars.constr_wt());

	// misclosure vector
	double l = (_a[0] * _a[0] + _a[1] * _a[1]
		+ _a[2] * _a[2] + _a[3] * _a[3]) - 1.;
	eqn.set_l(0, -l);

	// add contributions to normal equation
	eqn.add_contr(atpa, atpl, ltpl);
}


// get derivatives of the quaternion

void EOQuaternion::calc_derivs_rot(DerivSub& deriv,
double h, double u, double v, double w,
const DPoint3& d, const vector<double>& a) const
{
	assert(a.size() == 4);
	
	DRotMatrix3 dR_q0;
	DRotMatrix3 dR_q1;
	DRotMatrix3 dR_q2;
	DRotMatrix3 dR_q3;
	
	dr(a[0], a[1], a[2], a[3], dR_q0, dR_q1, dR_q2, dR_q3);
	
	DPoint3 dd_q0; mult_inv(dd_q0, dR_q0, d);
	DPoint3 dd_q1; mult_inv(dd_q1, dR_q1, d);
	DPoint3 dd_q2; mult_inv(dd_q2, dR_q2, d);
	DPoint3 dd_q3; mult_inv(dd_q3, dR_q3, d);
	
	deriv(0,0) = h * (w * dd_q0.x - u * dd_q0.z);
	deriv(1,0) = h * (w * dd_q0.y - v * dd_q0.z);
	
	deriv(0,1) = h * (w * dd_q1.x - u * dd_q1.z);
	deriv(1,1) = h * (w * dd_q1.y - v * dd_q1.z);
	
	deriv(0,2) = h * (w * dd_q2.x - u * dd_q2.z);
	deriv(1,2) = h * (w * dd_q2.y - v * dd_q2.z);
	
	deriv(0,3) = h * (w * dd_q3.x - u * dd_q3.z);
	deriv(1,3) = h * (w * dd_q3.y - v * dd_q3.z);
}


// get derivatives of the quaternion

void EOQuaternion::calc_derivs_rot(DerivSub& deriv,
double h, double u, double v, double w,
const DPoint3& d, const DRotMatrix3& M, const vector<double>& a) const
{
	assert(a.size() == 4);
	
	DRotMatrix3 dR_q0;
	DRotMatrix3 dR_q1;
	DRotMatrix3 dR_q2;
	DRotMatrix3 dR_q3;
	
	dr(a[0], a[1], a[2], a[3], dR_q0, dR_q1, dR_q2, dR_q3);
	
	DRotMatrix3 dN_q0 = dR_q0 * M;
	DRotMatrix3 dN_q1 = dR_q1 * M;
	DRotMatrix3 dN_q2 = dR_q2 * M;
	DRotMatrix3 dN_q3 = dR_q3 * M;
	
	DPoint3 dd_q0; mult_inv(dd_q0, dN_q0, d);
	DPoint3 dd_q1; mult_inv(dd_q1, dN_q1, d);
	DPoint3 dd_q2; mult_inv(dd_q2, dN_q2, d);
	DPoint3 dd_q3; mult_inv(dd_q3, dN_q3, d);
	
	deriv(0,0) = h * (w * dd_q0.x - u * dd_q0.z);
	deriv(1,0) = h * (w * dd_q0.y - v * dd_q0.z);
	
	deriv(0,1) = h * (w * dd_q1.x - u * dd_q1.z);
	deriv(1,1) = h * (w * dd_q1.y - v * dd_q1.z);
	
	deriv(0,2) = h * (w * dd_q2.x - u * dd_q2.z);
	deriv(1,2) = h * (w * dd_q2.y - v * dd_q2.z);
	
	deriv(0,3) = h * (w * dd_q3.x - u * dd_q3.z);
	deriv(1,3) = h * (w * dd_q3.y - v * dd_q3.z);
}


// non-member functions ...

} // namespace DEP
