/* $Id: extori.cpp,v 1.22 2008/01/08 16:37:27 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <typeinfo>
#include "extori.h"

namespace DEP
{

// constructor

ExtOrientation::ExtOrientation(size_t sz) : Station(sz+3),
	_time(-1.), // this indicates a not valied time!!!
	_per_centre(),
	_sdev_pcc()
{
	assert(3 <= sz && sz <= 9); // check number of rotation params.
	_a.resize(sz);
	_sdev_a.resize(sz);
}


// destructor

ExtOrientation::~ExtOrientation()
{
	// empty
}


// read

bool ExtOrientation::read(istream& is)
{
	// station ID
	is >> _id;
	if (!is)
		{
		cout << "ERROR: Reading ID failed.\n";
		return false;
		}

	// exposure time
	is >> _time;
	if (!is)
		{
		cout << "ERROR: Cannot read exposure time.\n";
		return false;
		}

	// projection centre coords.
	is >> _per_centre;
	if (!is)
		{
		cout << "ERROR: Reading projection centre coords. failed.\n";
		return false;
		}

	// std. dev. projection centre coords.
	is >> _sdev_pcc;
	if (!is)
		{
		cout << "ERROR: Reading std. dev. of proj. centre coords. failed.\n";
		return false;
		}

	// rotation parameters
	size_t i = 0;
	for (; i < _a.size(); i++)
		{
		is >> _a[i];
		if (!is)
			{
			cout << "ERROR: Reading rotation parameters failed.\n";
			return false;
			}
		}

	// std. dev. of rotation parameters
	for (i = 0; i < _sdev_a.size(); i++)
		{
		is >> _sdev_a[i];
		if (!is)
			{
			cout << "ERROR: Reading std. dev. of rotation parameters failed.\n";
			return false;
			}
		}

	return true;
}


// write

void ExtOrientation::write(ostream& os) const
{
	//double sc = AdjObject::lspars.scale_obj();
	double sc = 1.;

	Form sci(6); sci.scientific().width(15);
	Form fix(5); fix.fixed().width(15);

	// type
	os << key_str() << " ";

	// ID
	os << _id << " ";

	// exposure time
	os << fix(_time) << endl;

	// projection centre
	os << fix(_per_centre.x * sc) << " "
		<< fix(_per_centre.y * sc) << " "
		<< fix(_per_centre.z * sc) << endl;
		
	if (is_rsdev(_sdev_pcc.x))
		os << fix(_sdev_pcc.x * sc);
	else
		os << sci(_sdev_pcc.x);
	os << " ";
	if (is_rsdev(_sdev_pcc.y))
		os << fix(_sdev_pcc.y * sc);
	else
		os << sci(_sdev_pcc.y);
	os << " ";
	if (is_rsdev(_sdev_pcc.z))
		os << fix(_sdev_pcc.z * sc);
	else
		os << sci(_sdev_pcc.z);
	os << endl;
	
	// rotation parameters
	size_t i = 0;
	for (; i < _a.size(); i++)
		os << " " << sci(_a[i]);
	os << endl;

	for (i = 0; i < _sdev_a.size(); i++)
		os << " " << sci(_sdev_a[i]);
	os << endl;
}


// print

void ExtOrientation::print(ostream& os) const
{
	//double sc = AdjObject::lspars.scale_obj();
	double sc = 1.;

	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(4); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Exposure time :  " << fix(_time) << " s\n";
	os << "Subtype       :  " << type_str() << endl;
	os << endl;

	os << "  Param.     State         Value       Std.dev.\n\n";

	fix.precision(3);
	os << "  " << str8(uname(0)) << "   "
		<< str4(state_str(_sdev_pcc.x)) << "   "
		<< fix(_per_centre.x * sc) << "   ";
	if (is_rsdev(_sdev_pcc.x))
		os << fix(_sdev_pcc.x * sc);
	else
		os << sci(_sdev_pcc.x);
	os << endl;
	
	os << "  " << str8(uname(1)) << "   "
		<< str4(state_str(_sdev_pcc.y)) << "   "
		<< fix(_per_centre.y * sc) << "   ";
	if (is_rsdev(_sdev_pcc.y))
		os << fix(_sdev_pcc.y * sc);
	else
		os << sci(_sdev_pcc.y);
	os << endl;
	
	os << "  " << str8(uname(2)) << "   "
		<< str4(state_str(_sdev_pcc.z)) << "   "
		<< fix(_per_centre.z * sc) << "   ";
	if (is_rsdev(_sdev_pcc.z))
		os << fix(_sdev_pcc.z * sc);
	else
		os << sci(_sdev_pcc.z);
	os << endl;
	
	fix.precision(6);
	for (size_t i = 0, k = 3; i < _a.size(); i++, k++)
		{
		os << "  " << str8(uname(k)) << "   "
			<< str4(state_str(_sdev_a[i])) << "   "
			<< fix(_a[i]) << "   ";
		if (is_rsdev(_sdev_a[i]))
			os << fix(_sdev_a[i]);
		else
			os << sci(_sdev_a[i]);
		os << endl;
		}
	os << endl;

	DRotMatrix3 R;
	calc_rot_matrix(R, _a);
	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// get adjustment object class name

string ExtOrientation::class_str() const
{
	ostringstream oss;
	oss << "EO";
	return oss.str();
}


// set upari

bool ExtOrientation::set_upari()
{
	//cout << _size << endl;
	if (create() != true) return false;
	
	int k = 0;
	set(k, &_per_centre.x, &_sdev_pcc.x); k++;
	set(k, &_per_centre.y, &_sdev_pcc.y); k++;
	set(k, &_per_centre.z, &_sdev_pcc.z); k++;
	
	assert(_a.size() == _sdev_a.size());
	for (size_t i = 0; i < _a.size(); i++, k++)
		set(k, &_a[i], &_sdev_a[i]);

	return true;
}


// print changes of unknowns

void ExtOrientation::print_changes(ostream& os, const Vector<double>& dx) const
{
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(6); sci.scientific().width(12);

	os << "Station ID :  " << _id << endl;
	os << "Subtype    :  " << type_str() << endl;
	os << "Index      :  " << _index << endl;
	os << endl;
	
	os << "  Param.     State"
		"       Approx.         Change       Adjusted\n\n";
	
	size_t i = 0;
	int k = _index;
	for (; i < 3; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i)) << "   "
			<< fix(dx(k)) << "   "
			<< fix(par(i) + dx(k)) << endl;
		}

	fix.precision(6);
	for (; i < _size; i++, k++)
		{
		os << "  " << str8(uname(i)) << "   "
			<< str4(state_str(i)) << "   "
			<< fix(par(i)) << "   "
			<< fix(dx(k)) << "   "
			<< fix((par(i) + dx(k))) << endl;
		}
	os << endl;
}


// print unknown parameters

void ExtOrientation::print_unknowns(ostream& os) const 
{
	//double sc = AdjObject::lspars.scale_obj();
	double sc = 1.;
	
	Form str4; str4.left().width(4);
	Form str8; str8.left().width(8);
	Form fix(3); fix.fixed().width(12);
	Form sci(1); sci.scientific().width(12);

	os << "Station ID    :  " << _id << endl;
	os << "Exposure time :  " << fix(_time) << " s\n";
	os << "Subtype       :  " << type_str() << endl;
	os << "Index         :  " << _index << endl;
	os << endl;

	os << "  Param.    State      Initial      Adjusted        Change      "
		"Std.dev.\n\n";

	size_t i = 0;
	for (; i < 3; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i)) << "  "
			<< fix(par(i)) << "  "
			<< fix(change(i)) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i) * sc);
		else
			os << sci(sdev(i));
		os << endl;
		}
	
	fix.precision(6);
	for (; i < _size; i++)
		{
		os << "  " << str8(uname(i)) << "  " << str4(state_str(i)) << "  "
			<< fix(ival(i)) << "  "
			<< fix(par(i)) << "  "
			<< fix(change(i)) << "  ";
		if (is_rsdev(sdev(i)))
			os << fix(sdev(i));
		else
			os << sci(sdev(i));
		os << endl;
		}
	os << endl;
	
	DRotMatrix3 R;
	calc_rot_matrix(R, _a);
	
	fix.precision(7);
	os << "Rotation matrix:\n";
	os << " " << fix(R(0,0)) << fix(R(0,1)) << fix(R(0,2)) << endl;
	os << " " << fix(R(1,0)) << fix(R(1,1)) << fix(R(1,2)) << endl;
	os << " " << fix(R(2,0)) << fix(R(2,1)) << fix(R(2,2)) << endl;
	os << endl;
}


// check on convergence

int ExtOrientation::check_convergence(const Vector<double>& dx) const
{
	if (lspars.chk_pcc())
		{
		int k = _index;
		for (size_t i = 0; i < 3; i++, k++)
			{
			if (fabs(dx(k)) > lspars.conv_pcc()) return i;
			}
		}

	if (lspars.chk_rot())
		{
		int k = _index + 3;
		for (size_t i = 3; i < _size; i++, k++)
			{
			if (fabs(dx(k)) > lspars.conv_rot()) return i;
			}
		}

	return -1;
}


// spatial intersection

// bool ExtOrientation::intersection(const ImageGeometry& g2,
// const DPoint3& pho_c1, const DPoint3& pho_c2,
// DPoint3& obj_c, double& e) const
// {
// 	assert(false);
// 	// using method by Albertz/Kreiling
// 	//e = intersection3(this->_per_centre, this->_rot_matrix, pho_c1,
// 	//	eo2._per_centre, eo2._rot_matrix, pho_c2, obj_c);
// 	//return true;
// 
// 	bool rtn = false;
// 	/*
// 	try
// 	{
// 		const ExtOrientation& eo2 = dynamic_cast<const ExtOrientation&>(g2);
// 		rtn = intersection(eo2, pho_c1, pho_c2, obj_c, e);
// 	}
// 	catch (bad_cast) 
// 	{
// 		cerr << "ExtOrientation::intersection: "
// 			"Incompatible image geometries!\n";
// 		rtn = false;
// 	}
// 	*/
// 	return rtn;
// }


// sum of absolute changes

void ExtOrientation::sum_changes_pcc(const Vector<double>& dx,
double& sum, size_t& num) const
{
	int k = _index;

	for (size_t i = 0; i < 3; i++, k++)
		if (state(i) != eli) sum += fabs(dx(k)), num++;
}


// sum of absolute changes

void ExtOrientation::sum_changes_rot(const Vector<double>& dx,
double& sum, size_t& num) const
{
	int k = _index + 3;

	for (size_t i = 3; i < _size; i++, k++)
		if (state(i) != eli) sum += fabs(dx(k)), num++;
}


// non-member functions ...

// project object point onto photo plane

bool projection(const DPoint3& pc, const DRotMatrix3& R,
const DPoint3& obj_c, DPoint3& pho_c)
{
	// reduce coordinates
	double dx = obj_c.x - pc.x;
	double dy = obj_c.y - pc.y;
	double dz = obj_c.z - pc.z;
	//DPoint3 d = obj_c - pc;

	// rotate system
	//DPoint3 trf_c = R * d;
	//if (fabs(trf_c.z) == 0.) cerr << "DEP::projection: zero denominator.\n";

	// reciprocal scale factor
	//double reci_lam = pho_c.z / trf_c.z;

	// denominator
	double q = R(0,2) * dx + R(1,2) * dy + R(2,2) * dz;
	if (fabs(q) == 0.)
		{
		cout << "ERROR: projection: zero denominator.\n";
		return false;
		}

	// nominators
	double m = R(0,0) * dx + R(1,0) * dy + R(2,0) * dz;
	double n = R(0,1) * dx + R(1,1) * dy + R(2,1) * dz;

	// photo coords.
	pho_c.x = pho_c.z * m / q;
	pho_c.y = pho_c.z * n / q;
	// scale into image plane
	//pho_c.x = trf_c.x * reci_lam;
	//pho_c.y = trf_c.y * reci_lam;
	
	return true;
}


// Sarrus formular

static double sarrus(double a[3][3])
{
	double det = a[0][0] * a[1][1] * a[2][2];
	
	det += a[0][1] * a[1][2] * a[2][0];
	det += a[0][2] * a[1][0] * a[2][1];
	det -= a[0][2] * a[1][1] * a[2][0];
	det -= a[0][0] * a[1][2] * a[2][1];
	det -= a[0][1] * a[1][0] * a[2][2];
	
	return det;
}


// spacial intersection

bool intersection(
const DPoint3& pc1, const DRotMatrix3& rm1, const DPoint3& pho_c1,
const DPoint3& pc2, const DRotMatrix3& rm2, const DPoint3& pho_c2,
DPoint3& obj_c, double& e)
{
	//cout << "intersec3()\n";
	//cout << pc1 << endl;
	//cout << rm1 << endl;
	//cout << pho_c1 << endl;
	//cout << pc2 << endl;
	//cout << rm2 << endl;
	//cout << pho_c2 << endl;
	//cout << endl;

	// transform image vector into object space
	DPoint3 q1;
	q1.x = rm1(0,0) * pho_c1.x + rm1(0,1) * pho_c1.y + rm1(0,2) * pho_c1.z;
	q1.y = rm1(1,0) * pho_c1.x + rm1(1,1) * pho_c1.y + rm1(1,2) * pho_c1.z;
	q1.z = rm1(2,0) * pho_c1.x + rm1(2,1) * pho_c1.y + rm1(2,2) * pho_c1.z;
	q1 += pc1;

	DPoint3 q2;
	q2.x = rm2(0,0) * pho_c2.x + rm2(0,1) * pho_c2.y + rm2(0,2) * pho_c2.z;
	q2.y = rm2(1,0) * pho_c2.x + rm2(1,1) * pho_c2.y + rm2(1,2) * pho_c2.z;
	q2.z = rm2(2,0) * pho_c2.x + rm2(2,1) * pho_c2.y + rm2(2,2) * pho_c2.z;
	q2 += pc2;
	
	double dx = q1.x - pc1.x;
	double dy = q1.y - pc1.y;
	double dz = q1.z - pc1.z;
	double d1 = sqrt(dx * dx + dy * dy + dz * dz);

	double l1 = dx / d1;
	double m1 = dy / d1;
	double n1 = dz / d1;

	dx = q2.x - pc2.x;
	dy = q2.y - pc2.y;
	dz = q2.z - pc2.z;
	double d2 = sqrt(dx * dx + dy * dy + dz * dz);

	double l2 = dx / d2;
	double m2 = dy / d2;
	double n2 = dz / d2;

	double l = m1 * n2 - n1 * m2;
	double m = - (l1 * n2 - n1 * l2);
	double n = l1 * m2 - m1 * l2;

	double bx = pc2.x - pc1.x;
	double by = pc2.y - pc1.y;
	double bz = pc2.z - pc1.z;

	double D_lam[3][3];
	D_lam[0][0] = bx; D_lam[1][0] = l; D_lam[2][0] = l2;
	D_lam[0][1] = by; D_lam[1][1] = m; D_lam[2][1] = m2;
	D_lam[0][2] = bz; D_lam[1][2] = n; D_lam[2][2] = n2;

	double D_mue[3][3];
	D_mue[0][0] = bx; D_mue[1][0] = l; D_mue[2][0] = l1;
	D_mue[0][1] = by; D_mue[1][1] = m; D_mue[2][1] = m1;
	D_mue[0][2] = bz; D_mue[1][2] = n; D_mue[2][2] = n1;

	double D[3][3];
	D[0][0] = l1; D[1][0] = l2; D[2][0] = l;
	D[0][1] = m1; D[1][1] = m2; D[2][1] = m;
	D[0][2] = n1; D[1][2] = n2; D[2][2] = n;

	double det_lam = sarrus(D_lam);
	double det_mue = sarrus(D_mue);
	double det = sarrus(D);
	if (det == 0.)
		{
		cout << "ERROR: intersection: rays are parallel.\n";
		return false;
		}

	double lam = - det_lam / det;
	double mue = - det_mue / det;

	double x1 = pc1.x + lam * l1;
	double y1 = pc1.y + lam * m1;
	double z1 = pc1.z + lam * n1;

	double x2 = pc2.x + mue * l2;
	double y2 = pc2.y + mue * m2;
	double z2 = pc2.z + mue * n2;

	obj_c.x = (x1 + x2) / 2.;
	obj_c.y = (y1 + y2) / 2.;
	obj_c.z = (z1 + z2) / 2.;

	dx = x1 - x2;
	dy = y1 - y2;
	dz = z1 - z2;

	e = sqrt(dx * dx + dy * dy + dz * dz);

	return true;
}


// ray-z-plane intersection

void intersection(const DPoint3& pc, const DRotMatrix3& R,
DPoint3& obj_c, const DPoint3& pho_c)
{
	// transformed photo coords.
	DPoint3 trf_c = R * pho_c;

	// scale factor
	double lam = (obj_c.z - pc.z) / trf_c.z;

	// get X,Y-object coords.
	obj_c.x = pc.x + lam * trf_c.x;
	obj_c.y = pc.y + lam * trf_c.y;
}

} // namespace DEP
