/* $Id: fiodep.h,v 1.16 2007/12/10 06:56:44 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_FIODEP_INCLUDED
#define DEP_FIODEP_INCLUDED

#include "depdef.h"

namespace DEP
{
class LSParams;
class Camera;
class Image;
class ObjectPoint;
class Station;
class APSet;
class PosCorrection;
class AttCorrection;
class OrientationData;
class CamINSConfig;
class IMUMisalignment;

/** DEP file I/O class.
 *  This class provides some functions for input and output
 *  of DEP-files.
 *  @name file I/O
 *  @author Dirk Stallmann
 */
class FileIO
{
public:
	/// file I/O codes
	enum ecode
	{
		/// unknown
		UNKNOWN,

		/// parameters/options
		PARAMS,

		/// LS parameters/options
		LS_PARAMS,

		/// photogrametric network design parameters/options
		NETWORK_DESIGN_PARAMS,

		/// cameras
		CAMERA,
			/// frame camera
			CAMERA_FRAME,
			/// CCD frame camera
			CAMERA_CCD,
			/// CCD frame camera with top-right orientation
			CAMERA_CCD_TR,
			/// pushbroom camera
			CAMERA_PUSHBROOM,
			/// pushbroom camera calibrated
			CAMERA_PUSHBROOM_CALIB,
			/// ADS camera
			CAMERA_ADS,

		/// image types
		IMAGE,
			/// frame image
			IMAGE_FRAME,
			/// frame image extended
			IMAGE_FRAME_EXT,
			/// scanner image
			IMAGE_SCANNER,
			/// DLT image
			IMAGE_DLT,
			/// affine transformation image
			IMAGE_AFFINE_TRANSF,
			
		/// station/exterior orientation
		STATION,
			EXT_ORI,
			EXT_ORI_OPK_ROT,
			EXT_ORI_POK_ROT,
			EXT_ORI_OPK_FIX,
			EXT_ORI_AUSTRALIS,
			EXT_ORI_QUATERNION,
			DLT_PARAMS,
			AFFINE_TRANSF_PARAMS,

		/// image coords.
		IMAGE_COORDS,

		/// object coords.
		OBJECT_COORDS,

		/// control support
		CONTROL_SUPP,

		/// additional parameter set
		APSET,
			/// implemented in GAP
			APSET_GAP,
			/// implemented in Australis
			APSET_AUSTRALIS,
			/// implemented in NSE
			APSET_NSE,
			/// orthogonal polynom by Ebner
			APSET_P12,
			/// orthogonal polynom by Gruen
			APSET_P44,
			/// modelling IO
			APSET_INNER_OR,
			/// modelling radial distortion
			APSET_RADIAL,
			/// modelling decentering distortion
			APSET_DECENTERING,
			/// modelling in-plane distortion
			APSET_INPLANE,
			/// modelling out-of-plane distortion
			APSET_OUTOFPLANE,
			/// Brown's eqn. 22
			APSET_BROWN_EQN22,
			/// Brown's eqn. 23
			APSET_BROWN_EQN23,
			
		/// position correction
		POS_COR,

		/// attitude correction
		ATT_COR,

		/// orientation data parameters/options
		ORIENTATION_DATA_PARAMS,
			
		/// orientation data
		ORIENTATION_DATA,
			/// orientation data in ODF form
			ORIENTATION_DATA_ODF,

		/// IMU misalignment
		IMU_MISALIGNMENT,

		/// camera/INS configuration
		CAM_INS_CONFIG,

		/// PAT-B parameters/options
		PATB_PARAMS,
		
		/// PAT-B photographs
		PATB_PHOTOGRAPHS,
		
		/// PAT-B control points
		PATB_CONTROL_POINTS,
		
		/// PAT-B orientation parameters
		PATB_ORIENT_PARAMS,

		/// object coordinate transformation
		OC_TRANSFORM,

		/// stop-dep
		STOP_DEP,

		/// end-of-map
		END_OF_MAP
	};

	// type of rotation
	enum
	{
		ROT_MATRIX,		// rotation matrix
		ROT_ANGLES 		// angle of rotation
	};

	/// constructor
	FileIO() { }

	/// destructor
	virtual ~FileIO() { }

	/**@name get keyword.
	 *
	 *  @param type
	 *  @return code
	 */
	static const char* keyword(ecode type)
	{
		return _map[type].keyword;
	}

	/**@name get code.
	 *
	 *  @param symbol
	 *  @return code
	 */
	static ecode code(const char* keyword);

	/**@name determine file type.
	 *
	 *  @param is (IN) input stream
	 *  @param type (OUT) file type
	 *  @return code
	 */
	bool file_type(const char* fname, ecode& type) const;
	
	/// read cameras
	bool read_cameras(istream& is, vector<Camera*>& cams) const;

	/// read images
	bool read_images(istream& is, const vector<Camera*>& cams,
		vector<Image*>& images);

	/// read stations
	bool read_stations(istream& is, vector<Station*>& stats);

	/// read stations have individual type and std. devs.
	bool read_stations_indiv(istream& is,
		vector<Station*>& stats);

	/// read stations have common type and std. devs.
	bool read_stations_common(istream& is,
		vector<Station*>& stats);

	/// read image coords.
	bool read_image_coords(istream& is, vector<Image*>& images);

	/// read object point coords.
	bool read_object_coords(istream& is, vector<ObjectPoint*>& obpts);

	/// read control support definition
	bool read_ctrl_supp(istream& is, vector<ObjectPoint*>& obpts);

	/// read AP sets
	bool read_ap_sets(istream& is, vector<APSet*>& aps);

	/// read position corrections
	bool read_pos_correction(istream& is,
		vector<PosCorrection*>& pcors);

	/// read attitude corrections
	bool read_att_correction(istream& is,
		vector<AttCorrection*>& acors);

	/// read IMU misaligment data
	bool read_imu_misalignment(istream& is,
		vector<IMUMisalignment*>& v);

	/// read orientation data
	bool read_orientation_data(const char* fname, ecode type,
		const char* id, vector<OrientationData*>& ods) const;

	/// read ADS camera
	bool read_ads_camera(const char* fname, ecode type,
	                 const char* id, vector<Camera*>& cams) const;
		
protected:
	/// map structure
	struct map_t
	{
		/// code
		int				code;

		/// keyword
		const char*		keyword;
	};

	/// map
	static map_t _map[];
};

} // namespace DEP

#endif
