/* $Id: lconeqn.h,v 1.5 2007/07/23 09:40:39 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef DEP_LCONEQN_INCLUDED
#define DEP_LCONEQN_INCLUDED

#include "depdef.h"

namespace DEP
{
class DerivSub;

/** Linearised condition equation class.
 *  The linearised condition equation the functional model in the
 *  general least-squares adjustment.
 *  The linear condition equations has the form
 *
 *    Bv + Ax + u = 0
 *
 *  where
 *    B       design matrix, dim(B)=(r,n)
 *    v       residual vector dim(v)=(n)
 *    A       design matrix, dim(A)=(r,u)
 *    x       unknown parameter vector, dim(x)=(u)
 *    u       constant vector, dim(u)=(r)
 *
 * Togehter with the stochatic model
 *
 *    Qll = sigma0**(-2) Vll
 *
 *  where
 *    Qll     cofactor matrix
 *    sigma0  standard deviation of unit weight
 *    Vll     covariance matrix, dim(Vll)=(r,r)
 *
 *  This class holds only the observation equations consist a vector of
 *  sub-matrices, the vector of shorten observation (l)
 *  and the weight vector (p).
 *  The sub-matrices holding the partial derivatives.
 *  Notice, the number of rows must be equal for all sub-matrices
 *  and vectors! 
 *  @author Dirk Stallmann
 */
class LinConditionEqn
{
public:
	/// constructor
	LinConditionEqn();

	/// print equations
	void print(ostream& os) const;

	/// set design matrix B
	void set_b(const Matrix<double>& B) { m_B = B; }
	//void set_b(Matrix<double>* B) { _B = B; }

	/// add design sub-matrix A
	void add_a(DerivSub* A) { _A.push_back(A); }

	/// add vector of design sub-matrix A
	void add_a(vector<DerivSub>& A) {
		vector<DerivSub>::iterator pa;
	    for (pa = A.begin(); pa != A.end(); ++pa)
			_A.push_back(&*pa);
	}

	/// set misclosure vector w
	void set_w(const Vector<double>& w) { _w = w; }

	/// set cofactor matrix Qll
	void set_q(const Matrix<double>& Qll) { _Qll = Qll; }

	/// add contributions of the linearized observation equation
	/// to the normal equations.
	/// By convention the normal matrix is handled in upper triangular form.
	bool add_contr(Matrix<double>& AtPA, Vector<double>& AtPw,
		double& wtPw) const;

private:
	/// design matrix B
	Matrix<double> m_B;	// compiler problem workaround: start member
                                // name not with an underscore! 

	/// submatrices of the design matrix A
	vector<DerivSub*> _A;

	/// misclosure vector
	Vector<double>	_w;

	/// co-factor matrix
	Matrix<double>	_Qll;
};

} // namespace DEP

#endif
