/* $Id: lspars.cpp,v 1.12 2007/07/23 21:17:55 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "lspars.h"

namespace DEP
{

// constructor

LSParams::LSParams() :
	_unit_objc("m"),
	_unit_angle("rad")
{
	_sigma0 = 1e-6;			// 1 um
	_max_iter = 10;			// 10 iterations
	_conv_chk = true;
	_conv_eps = 1e-16;		// 10^-16
	_chk_obj = false;
	_chk_pcc = false;
	_chk_rot = false;
	_conv_obj = 0.001;		// 1 mm
	_conv_pcc = 0.001;		// 1 mm
	_conv_rot = 1e-6;		// 10^-6 rad
	_smin = SMIN;
	_smax = SMAX;
	_sminu = SMIN_UNK;
	_smaxu = SMAX_UNK;
	_unksup_wt = UNKSUP_WT;
	_constr_wt = CONSTR_WT;
	_ccoef_lim = 0.8;
	_incr_crd = 1e0;		// 1 m
	_incr_rot = 1e-3;		// 1 mrad
	_t_quantil = 1.96;		// t(inf,95%)
	_atpv_lim = 1e-4; 
	_res_lim = 10e-6;		// 10 um
	//_unit_objc(UnitLength::METRE),
	//_unit_angle(UnitAngle::RADIAN),
	_rho = 1.;
	_adj_interface = NATIVE;
	_ap_derivs = IMAGE_COORDS_PLUS_APS;
}


// read

bool LSParams::read(istream& is)
{
	string str;
	
	// std. dev. of unit weight
	is >> _sigma0;

	// maximum nuber of iterations
	is >> _max_iter;

	// convergence check and limit
	is >> _conv_chk >> _conv_eps;
	
	// unknowns used as iteration criterion
	is >> _chk_obj >> _chk_pcc >> _chk_rot;

	// convergence limits
	is >> _conv_obj >> _conv_pcc >> _conv_rot;

	// limit of std. devs.
	is >> _smin >> _smax >> _sminu >> _smaxu;
	if (_smin < 0. ||  _smax < 0. || _sminu < 0. || _smaxu < 0.)
		{
		cout << "ERROR: Limits of std. devs. must be >= than zero.\n";
		return false;
		}

	// check std. devs.
	if (_sminu >= _smin)
		{
		cout << "ERROR: sminu must be less than smin.\n";
		return false;
		}

	if (_smaxu <= _smax)
		{
		cout << "ERROR: smaxu must be greater than smax.\n";
		return false;
		}

	// weights
	is >> _unksup_wt >> _constr_wt;

	// output limit of corr. coeff.
	is >> _ccoef_lim;

	// increments for num. derivatives
	is >> _incr_crd >> _incr_rot;

	// t-quantil
	is >> _t_quantil;

	// A'Pv check
	is >> _atpv_lim;

	// residual limit for printing
	is >> _res_lim;
	_res_lim /= 1e6;

	// unit of object space coords.
	is >> str;
	if (!is)
		{
		cout << "Reading unit of object coords. failed.\n";
		return false;
		}

	UnitLength ulength;
	UnitLength::ecode ltype = ulength.code(str.c_str());
	if (ltype == UnitLength::UNKNOWN)
		{
		cout << "Unknown unit of object coords.: " << str << endl;
		return false;
		}

	// unit of rotation angles
	is >> str;
	if (!is)
		{
		cout << "Reading unit of rotation angles failed.\n";
		return false;
		}

	UnitAngle uangle;
	UnitAngle::ecode atype = uangle.code(str.c_str());
	if (atype == UnitAngle::UNKNOWN)
		{
		cout << "Unknown unit of rot. angles: " << str << endl;
		return false;
		}
	_rho = uangle.rho(atype);
	_unit_angle = uangle.symbol(atype);

	// adjustment interface
	is >> str;

	if (str.compare("native") == 0)
		_adj_interface = NATIVE;
	else if (str.compare("lapack") == 0)
		_adj_interface = LAPACK;
	else
		{
		cout << "Unknown adjustment interface: " << str << endl;
		return false;
		}

	// derivatives of APs
	is >> str;

	if (str.compare("image-coords") == 0)
		_ap_derivs = IMAGE_COORDS;
	else if (str.compare("collinear-equation") == 0)
		_ap_derivs = COLLINEAR_EQUATION;
	else if (str.compare("image-coords-plus-aps") == 0)
		_ap_derivs = IMAGE_COORDS_PLUS_APS;
	else
		{
		cout << "Unknown method to build derivatives of APs: " << str << endl;
		return false;
		}
	
	return true;
}


// print

void LSParams::print(ostream& os) const
{
	Form fix2(2); fix2.fixed().width(8);
	Form fix4(4); fix4.fixed().width(8);
	Form sci2(2); sci2.scientific().width(8);
	
	os << "Least-squares parameters/options\n";
	os << "================================\n";
	os << "Standard deviation of unit weight  :  "
		<< fix2(_sigma0 * 1e6) << " um\n";
	os << "Maximum number of iterations       :  " << _max_iter << "\n";
	os << "\n";
	os << "General convergence check          :  "
		<< (_conv_chk ? "on" : "off") << endl;
	os << "Convergence limit                  :  "
		<< sci2(_conv_eps) << endl;

	os << "Unknowns used for iteration criterion\n";
	os << "  - object point coords.           :  "
		<< (_chk_obj ? "on" : "off") << endl;
	os << "  - perspective centre coords.     :  "
		<< (_chk_pcc ? "on" : "off") << endl;
	os << "  - rotation parameters            :  "
		<< (_chk_rot ? "on" : "off") << endl;
	os << "Convergence limits\n";
	os << "  - object point coords.           :  "
		<< fix4(_conv_obj) << " m\n";
	os << "  - perspective centre coords.     :  "
		<< fix4(_conv_pcc) << " m\n";
	os << "  - rotation parameters            :  "
		<< sci2(_conv_rot) << "\n";

	os << "Limit of std. dev. to indicate that\n";
	os << "  - unknown should be eliminated   :  " << sci2(_smin) << "\n";
	os << "  - value is an unknown            :  " << sci2(_smax) << "\n";
	os << "Value of std. dev. to indicate that\n";
	os << "  - unknown is fixed               :  " << sci2(_sminu) << "\n";
	os << "  - unknown is free                :  " << sci2(_smaxu) << "\n";
	os << "Weight for suppression of unknowns :  "
		<< sci2(_unksup_wt) << "\n";
	os << "Weight for constraints             :  "
		<< sci2(_constr_wt) << "\n";
	os << "Output limit of correlation coeffs :  "
		<< fix2(_ccoef_lim) << "\n";
	os << "Increment for num. derivatives\n";
	os << "  - object point coordinates       :  "
		<< sci2(_incr_crd) << " m\n";
	os << "  - rotation angles                :  "
		<< sci2(_incr_rot) << " rad\n";
	os << "Limit for single significance test :  "
		<< fix2(_t_quantil) << "\n";
	os << "Limit for A'Pv-check               :  "
		<< sci2(_atpv_lim) << "\n";
	os << "Limit for printing residuals       :  "
		<< fix2(_res_lim * 1e6) << " um\n";
	os << endl;
	os << "Unit of object space coords.       :  "
		<< _unit_objc.symbol() << "\n";
	//os << Unit of image  space coords.       :  "
	//	_unit_image.symbol_cstr() << endl;
	os << "Unit of rotation angles            :  "
		<< _unit_angle.symbol() << "\n";
	os << "\n";

	os << "Adjustment interface               :  ";
	switch (_adj_interface)
		{
		case NATIVE:
			os << "native";
			break;

		case LAPACK:
			os << "lapack";
			break;
		}
	os << endl;

	os << "Partial derivatives of APs from    :  ";
	switch (_ap_derivs)
	{
		case IMAGE_COORDS:
			os << "image coords.";
			break;
			
		case COLLINEAR_EQUATION:
			os << "collinear equation";
			break;

		case IMAGE_COORDS_PLUS_APS:
			os << "image coords. plus APs";
			break;
	}
	os << endl;
	
	os << endl;
}

} // namespace DEP
