/* $Id: pmflin.cpp,v 1.7 2007/12/10 06:58:02 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2005 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <typeinfo>
#include "pmflin.h"
#include "depvec.h"
#include "camera.h"
#include "lconeqn.h"
#include "imgpnt.h"
#include "objpnt.h"
#include "cholesky.h"

namespace DEP
{

// get type name

string AffineTransfParams::type_str() const
{
	ostringstream oss;
	oss << "affine transf.";
	return oss.str();
}


// get key name

string AffineTransfParams::key_str() const
{
	ostringstream oss;
	oss << "affine-transf";
	return oss.str();
}


// get parameter name

string AffineTransfParams::uname(size_t i) const
{
	ostringstream oss;
	assert(i < _size);

	switch (i)
		{
		case 0: oss << "a0"; break;
		case 1: oss << "a1"; break;
		case 2: oss << "a2"; break;
		case 3: oss << "a3"; break;
		case 4: oss << "b0"; break;
		case 5: oss << "b1"; break;
		case 6: oss << "b2"; break;
		case 7: oss << "b3"; break;
		}

	return oss.str();
}


// object-to-photo coords. transformation

bool AffineTransfParams::object_to_image(const DPoint3& obj_c, DPoint2& img_c) const
{
	affine_transform(_a, obj_c, img_c);
	
	return true;
}


// constructor

AffineTransfImage::AffineTransfImage() : Image()
{
	// empty
}


// get image type

string AffineTransfImage::type_str() const
{
	ostringstream oss;
	oss << "affine transf.";
	return oss.str();
}


// object-to-image coords. transformation (UNUSED ARGUMENTS: r)

bool AffineTransfImage::object_to_image(const DPoint3& obj_c, DPoint2& img_c, double r) const
{
	assert(_station != 0);
	assert(_camera != 0);
	
	const TransfParams* trf = dynamic_cast<const TransfParams*>(_station);
	assert(trf != 0);
	
	DPoint2 tmp_c;
	trf->object_to_image(obj_c, tmp_c);
	
	DPoint3 pho_c(tmp_c.x, tmp_c.y, _camera->ck());
	_camera->photo_to_image(pho_c, img_c);

	return true;
}


// resection

bool AffineTransfImage::resection(ostream& os, const vector<ObjectPoint*>& obpts, bool dbg)
{
	assert(_station != 0);	// necessary
	assert(_camera != 0);	// necessary
	
	os << "Image ID   : " << _image_id << endl;
	os << "Station ID : " << _station->get_id() << endl;
	os << endl;
	os << "Point ID :" << endl;

	TransfParams* trf = dynamic_cast<TransfParams*>(_station);
	assert(trf != 0);
	
	int nu = trf->num_unk();
	DMatrix atpa(nu, nu);
	DVector atpl(nu);
	double ltpl = 0.;

	// submatrices for ...
	DerivSub	deriv_eo(2,nu);		// transf. params.
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2

	// derivatives of observations
	deriv_obs.identity();
	
	// set index for transf. params.
	deriv_eo.index(0);

	// always 2 condition equations for a single imaging ray
	LinConditionEqn con_eqn;

	// assign A-submatrix
	con_eqn.add_a(&deriv_eo);
	//con_eqn.add_a(deriv_ap);	// optional
	
	// assign B-submatrix
	con_eqn.set_b(deriv_obs);
	
	// co-factors: set all weights to be equal!
	DMatrix Qll(2,2);
	Qll(0,0) = 1.;
	Qll(1,1) = 1.;
	con_eqn.set_q(Qll);
	
	DVector	wt(2);				// misclosure vector
	int num_rays = 0;
	vector<ImagePoint*> imgpts; // image points in use
	
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			os << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}
			
		if (op->is_approx() == false) continue;
		
		if (num_rays % 4 == 0) os << endl;
		os << "  " << op->get_id();
		imgpts.push_back(*ip);

		// convert sensor into (observed) image coords. 
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;
		
		// convert sensor to image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

// 		// co-factors
// 		Matrix<double> Qll(2,2);
// 		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
// 		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
// 		con_eqn.set_q(Qll);
// 		//cout << "Qll: " << Qll << endl;
		
		// coefficients
		deriv_eo(0, 0) = 1.;
		deriv_eo(0, 1) = op->crds.x;
		deriv_eo(0, 2) = op->crds.y;
		deriv_eo(0, 3) = op->crds.z;
		deriv_eo(0, 4) = 0.;
		deriv_eo(0, 5) = 0.;
		deriv_eo(0, 6) = 0.;
		deriv_eo(0, 7) = 0.;
		
		deriv_eo(1, 0) = 0.;
		deriv_eo(1, 1) = 0.;
		deriv_eo(1, 2) = 0.;
		deriv_eo(1, 3) = 0.;
		deriv_eo(1, 4) = 1.;
		deriv_eo(1, 5) = op->crds.x;
		deriv_eo(1, 6) = op->crds.y;
		deriv_eo(1, 7) = op->crds.z;
		
// 		// partial derivatives of APs
// 		{
// 			DPoint3 pho_c;
// 			_camera->image_to_photo(img_c, pho_c, false);
// 			_camera->get_derivs(pho_c, deriv_ap);
// 		}
		
		// calculate shorten observations (observed - computed term)
		wt(0) = img_c.x;
		wt(1) = img_c.y;
		con_eqn.set_w(wt);
		//cout << "w: " << wt << endl; cout.flush();

		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return false;
		//con_eqn.print(cout); cout << endl;
		
		num_rays++;
		}
	os << endl << endl;

	int num_obs = 2 * num_rays;
	os << "Number of rays : " << num_rays << endl;
	if (num_obs < nu)
		{
		os << "ATTENTION: Insufficient control points!" << endl;
		return false;
		}
	os << endl;
	
	// solve normal equation sytem
	DVector dec;
	int info = chol_dec(atpa, dec);
	if (info != 0)
		{
		os << "Decomposition of normal matrix failed in row "
			<< info << endl;
		return false;
		}

	DVector x;
	info = chol_sol(atpa, dec, atpl, x);
	if (info != 0)
		{
		os << "Backsubstitution of normal equation system failed: "
			<< "info = " << info << endl;
		return false;
		}
		
	double xtatpl = dot_prod(x, atpl);
	double vtpv = ltpl - xtatpl;
	
	int f = num_obs - nu;
	assert(f > 0);
	
	double sigma0 = sqrt(vtpv / double(f));
	
	Form fix2(2); fix2.fixed();
	os << "Number of observations  :  " << num_obs << endl;
	os << "Number of unknowns      :  " << nu << endl;
	os << "Redundancy              :  " << f << endl;
	os << "Sigma 0 a posteriori    :  " << fix2(sigma0 * 1e6) << " um\n";
	os << endl;
	
	// set tranformation parameters
	for (int i = 0; i < nu; i++)
		{
		trf->par(i, x(i));
		//cout << "a" << i << " = " << x(i) << endl;
		}
	trf->print(os);
	
	os << "Differences in image points\n\n";
	os << "  Point ID       dx/um      dy/um     |d|/um\n\n";
	
	Form str8; str8.width(8);
	Form fix10(2); fix10.fixed().width(10);
	
	DPoint2 sum_d;
	DPoint2 sum_dd;
	int num = 0;
	
	DPoint2 img_c0;
	DPoint2 d, dd;
	double d_abs = 0.;
	
	// loop over used image points
	for (ip = imgpts.begin(); ip != imgpts.end(); ++ip)
		{
		if (*ip == 0) continue;

		// find corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			os << "ATTENTION: Object point "
				<< (*ip)->get_id() << " not found" << endl;
			continue;
			}

		// collinear equation -> calculate image coords.
		// Attention: The scanline coordinate is only used by scanner images!!!
		DPoint2 img_c0;
		object_to_image(op->crds, img_c0);

		// convert sensor into image coords.
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);

		// difference
		d = img_c0 - img_c;
		d_abs = abs(d);
		dd = d * d;

		// convert sensor into image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		
		sum_d += d;
		sum_dd += dd;
		num++;

		// print differences
		if (AdjObject::lspars.is_small(d.x) || AdjObject::lspars.is_small(d.y))
			{
			os << "  " << str8((*ip)->get_id()) << "  "
				<< fix10(d.x * 1e6) << " " << fix10(d.y * 1e6) << " "
				<< fix10(d_abs * 1e6);
			os << endl;
			}
		}
	os << endl;

	os << "Number of points          :   " << num << endl;
	os << "RMS of differences        :   "
		<< fix10(sqrt(sum_dd.x / num) * 1e6) << "   "
		<< fix10(sqrt(sum_dd.y / num) * 1e6) << endl;
	os << "Average of differences    :   "
		<< fix10(sum_d.x * 1e6 / num) << "   "
		<< fix10(sum_d.y * 1e6 / num) << endl;
	os << endl;
	
	// precision of transformation parameters missing!!!
	
	return true;
}


// observation equation for intersection

bool AffineTransfImage::observ_eqn_intersec(const string& id, DerivSub& a, DVector& wt,
bool& found) const
{
	ImagePoint* ip = get_image_point(id);
	if (ip == 0)
		{
		found = false;
		return true; // escape, point not found
		}
	found = true;

	DPoint3 pho_c;
	sensor_to_photo(ip->crds, pho_c);
	
	const TransfParams* trf = dynamic_cast<const TransfParams*>(_station);
	assert(trf != 0);
	const double* c = trf->a();
	
	a(0,0) = c[1];
	a(0,1) = c[2];
	a(0,2) = c[3];

	a(1,0) = c[5];
	a(1,1) = c[6];
	a(1,2) = c[7];

	wt(0) = pho_c.x - c[0];
	wt(1) = pho_c.y - c[4];

	return true;
}


// add contribution to normal equations for all image point observations

bool AffineTransfImage::normal_eqns(const vector<ObjectPoint*>& obpts,
Matrix<double>& atpa, Vector<double>& atpl, double& ltpl, int& num_obs) const
{
	assert(_station != 0);	// necessary
	assert(_camera != 0);	// necessary
	
	const TransfParams* trf = dynamic_cast<const TransfParams*>(_station);
	assert(trf != 0);

	// Setup sub-matrices of derivatives
	// Always 2 observation equations for each image point.
	// The number of columns might be variable.
	DerivSub	deriv_oc(2,3);		// object points, size always 2x3
	DerivSub	deriv_eo(2,8);		// transf. params, size always 2x8
	vector<DerivSub> deriv_ap(_camera->num_ap_sets());	// APs, variable size
	DerivSub	deriv_obs(2,2);		// observations, size always 2x2

	// derivatives of observations
	deriv_obs.identity();
	
	// set index for EO (valid for the whole image)
	deriv_eo.index(trf->get_index());

	// always 2 condition equations for a single imaging ray
	LinConditionEqn con_eqn;

	// design matrix A
	con_eqn.add_a(&deriv_oc);	// always
	con_eqn.add_a(&deriv_eo);	// always
	con_eqn.add_a(deriv_ap);	// optional
	
	// design matrix B
	con_eqn.set_b(deriv_obs);
	
	// DLT parameters
	const double* a = trf->a();
	
	// loop over all image points
	vector<ImagePoint*>::const_iterator ip;
	for (ip = _image_points.begin(); ip != _image_points.end(); ++ip)
		{
		if (*ip == 0) continue;
		
		// search for corresponding object point
		ObjectPoint* op = find(obpts, (*ip)->get_id());
		if (op == 0)
			{
			cout << "ATTENTION: Object point " << (*ip)->get_id()
				<< " not found.\n";
			continue;
			}

		//cout << _image_id << ":" << (*ip)->get_id() << ":\n"; cout.flush();
		
		// set current index of object point
		deriv_oc.index(op->get_index());

		// convert sensor to image std. dev.
		DPoint2 sdev;
		_camera->sdev_sensor_to_image((*ip)->sdev, sdev);
		//cout << (*ip)->get_id() << ": " << sdev << endl;

		// co-factors
		Matrix<double> Qll(2,2);
		Qll(0,0) = sdev.x / AdjObject::lspars.sigma0(); Qll(0,0) *= Qll(0,0);
		Qll(1,1) = sdev.y / AdjObject::lspars.sigma0(); Qll(1,1) *= Qll(1,1);
		con_eqn.set_q(Qll);
		//cout << "Qll: " << Qll << endl;
			
		// image coords. derived from unknowns.
		// Use these values not the observations!!!
		// See personal notes.	
		double x = a[0] + a[1] * op->crds.x + a[2] * op->crds.y + a[3] * op->crds.z;
		double y = a[4] + a[5] * op->crds.x + a[6] * op->crds.y + a[7] * op->crds.z;

		// derivatives
		deriv_eo(0, 0) = 1.;
		deriv_eo(0, 1) = op->crds.x;
		deriv_eo(0, 2) = op->crds.y;
		deriv_eo(0, 3) = op->crds.z;
		deriv_eo(0, 4) = 0.;
		deriv_eo(0, 5) = 0.;
		deriv_eo(0, 6) = 0.;
		deriv_eo(0, 7) = 0.;

		deriv_eo(1, 0) = 0.;
		deriv_eo(1, 1) = 0.;
		deriv_eo(1, 2) = 0.;
		deriv_eo(1, 3) = 0.;
		deriv_eo(1, 4) = 1.;
		deriv_eo(1, 5) = op->crds.x;
		deriv_eo(1, 6) = op->crds.y;
		deriv_eo(1, 7) = op->crds.z;

		deriv_oc(0,0) = a[1];
		deriv_oc(0,1) = a[2];
		deriv_oc(0,2) = a[3];
	
		deriv_oc(1,0) = a[5];
		deriv_oc(1,1) = a[6];
		deriv_oc(1,2) = a[7];

		// image coords. derived from the unknown parameters
		DPoint2 tmp_c(x, y);
		DPoint3 pho_c;
		_camera->image_to_photo(tmp_c, pho_c, false);

		// partial derivatives of APs
		_camera->get_derivs(pho_c, deriv_ap);

		// convert sensor into (observed) image coords. 
		DPoint2 img_c;
		_camera->sensor_to_image((*ip)->crds, img_c);
		//cout << "img: " << img_c << endl;

		// AP correction
		DPoint2 corr = _camera->ap_set_corr(pho_c);
		
		// computed observations
		DPoint2 img_c0(x + corr.x, y + corr.y);
		//cout << "img_c0: " << img_c0 << endl;

		// calculate shorten observations (observed - computed term)
		// misclosure vector
		Vector<double> wt(2);
		wt(0) = img_c.x - img_c0.x;
		wt(1) = img_c.y - img_c0.y;
		con_eqn.set_w(wt);
		//cout << "w: " << wt << endl; cout.flush();

		// add contributions to normal equation
		if (!con_eqn.add_contr(atpa, atpl, ltpl))
			return false;
		//con_eqn.print(cout); cout << endl;
		
		num_obs += 2;
		}
		
	return true;
}


// non-member functions ...

// projection using affine transformation

void affine_transform(const double* a, const DPoint3& obj_c, DPoint2& img_c)
{
	assert(a != 0);

	// image coords.
	img_c.x = a[0] + a[1] * obj_c.x + a[2] * obj_c.y + a[3] * obj_c.z;
	img_c.y = a[4] + a[5] * obj_c.x + a[6] * obj_c.y + a[7] * obj_c.z;
}

} // namespace DEP
