/**********************************************************************/
//
// FILE:    goni2dgap.cpp
//          Preprocessing program for the bundle adjustment program dgap21.
//          Calculation of (a) object coordinates from the gonimeter
//          measurements, (b) image coordinates from the pixel position
//          in CCD image and (c) exterior orientation files.
//          The gonimeter measurements and the corresponding pixel coordinates
//          are read from the measurement file (mea). The parameters of
//          the CCD camera are read from the camera file (cam). The null
//          direction of the theodolit (r0) and the camera azimut (alpha)
//          are approximations for the orientation of the theodolit and
//          the camera. (Changes of this orientation parameters are
//          later calculated in the bundle adjustment to find an optimal
//          fit of the bundles. The changes of the orientation parameters 
//          are of less interest.) The input data is usually in cm2-format
//          e.g. each row contains:
//
//             <mea_id> <r1> <x1> <y1> <r2> <x2> <y2>
//          
//
// CREATED: 30 May 01  (ds)
// CHANGES: 01 Sep 01  Using libdep-2.1. (ds)
//          05 Sep 01  Add some comments and update usage information.
//                     The output files are now libdep-2.1 compatible. (ds)
//          10 Sep 01  Change program name from ZICal2dgap to goni2dgap. (ds)
//          05 Oct 01  Bug fix in parsing camera keywords and change of
//                     keywords. Output of the exterior orientation file
//                     (.eo). The gonimeter measurements are only done in
//                     the X-Z-plane (the Y-coordinate is always zero). 
//                     And therefore the camera azmiut (alpha) is applied
//                     in the exterior orientation by kappa = alpha! (ds)
//                     Using the Parametrisation omega-phi-kappa about
//                     fixed axies and omega = phi ~ 0, kappa is the angle
//                     to rotate the x-image-coordinate axis parallel to the
//                     X-object-coordinate axis positive in counter-clockwise
//                     direction. (ds)
//          18 Oct 01  Changes for dgap 2.2 (ds):
//                      - Output of pixel instead of image coords.
//                        (dgap supports now CCD cameras!). 
//                      - Image related information (image ID, camera ID,
//                        image points) are split in two files, the
//                        image file (.im) and the image coords. file (.ic).
//                      - The std. dev. of the pixel/image coords. is now
//                        an argument parameter. 
//          25 Feb 02  Image and station ID is required for the
//                     image definition (.im) (ds)
//          01 Mar 02  Change name of variable alpha into kappa.
//                     kappa is now the rotation of the camera
//                     while alpha is the azimut of the collimator. (ds)
//          26 Jul 02  Adapted to dgap 2.4.3. (ds)
//          09 Mar 04  Changes for new DEP library and C++ standard. (ds)
// REFS:    ifp-Bericht zur DMC-Kamerakalibrierung
//
// (c) 2001 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <fstream>
using namespace std;

#include "libdep.h"
using namespace DEP;


const char* const program_name = "goni2dgap";
const char* const program_version = "version 0.10.0 (10/2007)";

static void usage(int status=0)
{
	cerr << "Usage: " << program_name << " [options] "
		"cam mea image_id r0 kappa sdev_ic\n"
		<< "Arguments:\n"
		<< "  cam              Camera filename.\n"
		<< "  mea              Measurement filename.\n"
		<< "  image_id         Image ID.\n"
		<< "  r0               Null direction of the theodolit in gon.\n"
		<< "  kappa            Camera rotation in gon.\n"
		<< "  sdev_ic          Std.devs. of image/pixel coords.\n"
		<< "Options:\n"
		<< "  -az alpha        Azimut alpha in gon.\n"
		<< "  -cm2             Input data in cm2-format.\n"
		<< "  -help            Show this text.\n"
		<< "  -verbose         Gives more information.\n"
		<< "  -version         Show program version.\n";

	exit(status);
}

static void error(const char* s, const char* s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}

// initilize static data members
DEP::LSParams DEP::AdjObject::lspars;	// definition of least-squares params.


int main(int argc, char* argv[])
{
	char*	cam_name = 0;		// input camera filename
	char*	mea_name = 0;		// input measurement filename
	char*	image_id = 0;		// image ID
	bool	verb = false;		// verbose
	bool	cm2 = false;		// cm2-format
	double	sc = 1e-6;			// object coords. scale = 1 um
	double	r0 = 0.;			// 
	double	alpha = 0.;			// azimut
	double	kappa = 0.;			// camera rotation
	double	sdev_ic = 1e-6;		// 1 um

	if (argc == 1) usage(1);

	for (int i = 1, j = 0; i < argc; i++)
		{
		if (**(argv + i) == '-')
			{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-az") == 0)
				kappa = atof(argv[++i]);
			else if (strcmp(argv[i], "-verbose") == 0)
				verb = true;
			else if (strcmp(argv[i], "-version") == 0)
				{
				cerr << program_name << " " << program_version
					<< " compiled " << __DATE__ << " " << __TIME__ << endl;
				exit(0);
				}
			else if (strcmp(argv[i], "-cm2") == 0)
				cm2 = true;
			else
				error("Illegal option:", argv[i]);
			}
		else
			{
			switch (j++)
				{
				case 0:
					cam_name = argv[i];
					break;

				case 1:
					mea_name = argv[i];
					break;

				case 2:
					image_id = argv[i];
					break;

				case 3:
					r0 = atof(argv[i]);
					break;

				case 4:
					kappa = atof(argv[i]);
					break;

				case 5:
					sdev_ic = atof(argv[i]);
					break;

				default:
					error("Too many arguments given:", argv[i]);
				}
			}
		}

	//cout << r0 << endl;
	//cout << alpha << endl;

	// convert gon into rad
	r0 /= rho_gon;
	alpha /= rho_gon;
	kappa /= rho_gon;

	// create camera
	CCDCamera* camera = 0;

	{
	ifstream is(cam_name);
	if (!is) error("Cannot open input camera file:", cam_name);

	// camera type
	string str;
	is >> str;

	if (str.compare("camera-ccd") == 0)
		camera = new CCDCamera;
	else if (str.compare("camera-ccd-tr") == 0)
		camera = new CCDCameraTR;
	else
		error("Unkown camera type:", str.c_str());
	assert(camera != 0);

	if (camera->read(is) != true)
		{
		camera->print(cout);
		error("Cannot read input camera file:", cam_name);
		}

	if (verb)
		{
		camera->print(cout);
		cout << endl;
		}
	}
	
	// open input measurement file
	ifstream mea_file(mea_name);
	if (!mea_file) error("Cannot open input measurement file:", mea_name);

	// open two (of four) output files:
	//   <image_id>.ic   image coords.
	//   <image_id>.oc   object coords.
	string		str;

	str = image_id;
	str += ".ic";
	ofstream ic_file(str.c_str());
	if (!ic_file) error("Cannot open output file:", str.c_str());
	
	str = image_id;
	str += ".oc";
	ofstream oc_file(str.c_str());
	if (!oc_file) error("Cannot open output file:", str.c_str());
	
	string 		mea_id;				// measurement point ID
	double		r = 0.;				// direction 
	double		x = 0., y = 0.;		// X/Y-object coords.
	double		z = -camera->ck() / sc;		// Z-object coord.
	DPoint2		pix_c;				// pixel coords.
	DPoint2		img_c;				// image coords.
	double		sdev_oc = 1e-31;	// very very small std. dev.

	Form		fix; fix.fixed().width(8);
	Form		fix8(4); fix8.fixed().width(8);
	Form		fix12(4); fix12.fixed().width(12);

	// write ic-file header
	//ic_file << image_id << "   " << camera->id() << "   " << sc << endl;
	//ic_file << image_id << endl;
	ic_file << "common-sdev" << endl;
	ic_file << fix8(sdev_ic) << " " << fix8(sdev_ic) << endl;
	ic_file << image_id << "   0. 0." << endl;
	
	// write oc-file header
	oc_file << "common-sdev"
		<< "   " << sdev_oc << "   " << sdev_oc << "   " << sdev_oc
		<< endl;

	// read input data
	//while (mea_file >> mea_id >> r >> pix_c.x >> pix_c.y)
	while (mea_file >> mea_id)
		{
		// read cm1 or cm2 format
		if (cm2)
			{
			double		r1 = 0., r2 = 0.;	// directions
			DPoint2		pix_c1, pix_c2;		// pixel coords.
			mea_file >> r1 >> pix_c1;
			mea_file >> r2 >> pix_c2;

			// average
			r = (400. - r1 + r2) / 2.;
			pix_c = (pix_c1 + pix_c2) / 2.;
			}
		else
			{
			mea_file >> r >> pix_c;
			//r2 = 3. * r1 - 400.;
			//pix_c2 = pix_c1;
			}

		string point_id = image_id;	
		point_id += '-' + mea_id;

		r /= rho_gon;

		// transform pixel into image coords.
		//camera.pixel_to_image(pix_c, img_c);
		//camera->external_to_image(pix_c, img_c);
		//ic_file << a(point_id)
		//	<< " " << fix12(img_c.x / sc)
		//	<< " " << fix12(img_c.y / sc)
		//	<< " " << fix8(sdev_ic / sc)
		//	<< " " << fix8(sdev_ic / sc)
		//	<< endl;

		img_c = pix_c;
		ic_file << fix(point_id)
			<< " " << fix12(pix_c.x)
			<< " " << fix12(pix_c.y)
			<< endl;

		// calc. object coords.
		// In fact the goniometer measurements are always done along
		// the X-axis while the camera is rotated.
		double tau = r - r0;
		double s = z * tan(tau);
		x = s * cos(alpha);
		y = s * sin(alpha);
		//x = s;
		//y = 0.;
		oc_file << fix(point_id)
			<< " " << fix12(x)
			<< " " << fix12(y)
			<< " " << fix12(z)
			<< endl;
		}
	
	// ic/oc-file 
	ic_file << "stop-dep\nstop-dep" << endl;
	oc_file << "stop-dep" << endl;

	// create EO
	EOAnglesOPKFix* eo = new EOAnglesOPKFix;
	assert(eo != 0);

	eo->set_id(image_id);

	// camera rotation
	vector<double> a(3, 0.);
	a[2] = kappa;
	eo->set_rot_params(a);

	DPoint3 sdev_pcc(1e-31); // fixed
	vector<double> sdev_rot(3, 1e+31); // free
	eo->set_sdev_per_centre(sdev_pcc);
	eo->set_sdev_rot_params(sdev_rot);
	eo->set_approx(true);

	if (verb)
		{
		eo->print(cout);
		cout << endl;
		}

	// write image file
	str = image_id;
	str += ".im";
	ofstream im_file(str.c_str());
	if (!im_file) error("Cannot open output file:", str.c_str());

	// image ID == EO ID
	im_file << "image-frame" << "   " << image_id << "   "
		 << image_id << "   " << camera->get_id() << endl;
	im_file << "stop-dep" << endl;

	// write EO
	str = image_id;
	str += ".eo";
	ofstream eo_file(str.c_str());
	if (!eo_file) error("Cannot open output file:", str.c_str());

	eo_file << "indiv-type\n";
	eo->write(eo_file);

	return 0;
}
