/* $Id: oc2gplt.cpp,v 1.4.2.2 2008/02/28 17:01:37 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2007 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <fstream>
//#include <vector>
//#include <sstream>
#include "libdep.h"
#include "fiodep.h"

using namespace std;
using namespace DEP;


const char* const program_name = "oc2gplt";
const char* const program_version = "version 0.1.1 (02/2008)";

static void usage(int status=0)
{
	cerr << "Usage: " << program_name << " [options] "
		"oc-file [cs-file] > gp-file\n"
		<< "Arguments:\n"
		<< "  oc-file      Object coords. filename.\n"
		<< "  cs-file      CS filename (optional).\n"
		<< "  gp-file      Gnuplot filename.\n"
		<< "Options:\n"
		<< "  -help            Show this text.\n"
		<< "  -pointsize sz    Set point size. Default: 1.0.\n"
		<< "  -offset x y      Set label offset. Default: 0.5 0.5.\n"
		<< "  -term <term>     Set terminal type = {png | eps}. Default: png.\n"
		<< "  -version         Show program version.\n";

	exit(status);
}

static void error(const char* s, const char* s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}


// initialize static data members
DEP::LSParams DEP::AdjObject::lspars;	// definition of least-squares params.
double OrientationData::shift = 0.;		// global offset for orient. data
OrientationData::IpolMethod OrientationData::ipol = LINEAR;	// interpolation method

// terminale
enum term_type
{
	TERM_PNG,
	TERM_EPS
};

int main(int argc, char *argv[])
{
	if (argc == 1) usage(1);

	term_type	term = TERM_PNG;
	char*	oc_fname = 0;
	char*	cs_fname = 0;
	float	pointsize = 1.;
	float	offset_x = 0.5;
	float	offset_y = 0.5;
	int		i = 0, j = 0;
	
	for (i = 1, j = 0; i < argc; i++)
		{
		if (**(argv + i) == '-')
			{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-offset") == 0)
			{
				offset_x = atof(argv[++i]);
				offset_y = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-pointsize") == 0)
			{
				pointsize = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-term") == 0)
			{
				i++;
				if (strcmp(argv[i], "png") == 0)
					term = TERM_PNG;
				else if (strcmp(argv[i], "eps") == 0)
					term = TERM_EPS;
				else
					error("Illegal argument:", argv[i]);
			}
			else if (strcmp(argv[i], "-version") == 0)
				{
				cerr << program_name << " " << program_version
					<< " compiled " << __DATE__ << " " << __TIME__
					<< endl;
				exit(0);
				}
			else
				error("Illegal option:", argv[i]);
			}
		else
			{
			switch (j++)
				{
				case 0:
					oc_fname = argv[i];
					break;
			
				case 1:
					cs_fname = argv[i];
					break;

				default:
					error("Too many arguments given:", argv[i]);
				}
			}
		}
	
	FileIO fio;
	vector<ObjectPoint*> obpts;
	
	if (oc_fname != 0)
		{
		ifstream is(oc_fname);
		if (!is) error("Cannot open input object coords. file:", oc_fname);
		fio.read_object_coords(is, obpts);
		}
	
	if (cs_fname != 0)
		{
		ifstream is(cs_fname);
		if (!is) error("Cannot open input CS file:", cs_fname);
		fio.read_ctrl_supp(is, obpts);
		}
	
	Form fix9(3); fix9.fixed().width(9);
	vector<ObjectPoint*>::const_iterator op;
	
	switch (term)
	{
		case TERM_PNG:
			cout << "set term png size 1200, 900" << endl;
			break;
					
		case TERM_EPS:
			cout << "set term postscript eps" << endl;
			break;
		
		default:
			assert(false);
			break;
	}
	cout << "set autoscale xy" << endl;
	//cout << "show pointsize" << endl;
	cout << "set pointsize " << pointsize << endl;
	//cout << "set title 'X/Y-plane'";
	cout << "set title 'X/Y-plane - ";
	cout << oc_fname;
	if (cs_fname != 0) cout << "   " << cs_fname;
	cout << "'" << endl;
	cout << "set xlabel 'X (m)'" << endl;
	cout << "set ylabel 'Y (m)'" << endl;
	cout << "set key off" << endl;
	//cout << "set key title \"control\"" << endl;
	cout << "set grid" << endl;
	cout << "set format x '%.0f'" << endl;
	cout << "set format y '%.0f'" << endl;
	cout << "set datafile missing '?'" << endl;
	//cout << "test" << endl;
	
	cout << "set size ratio -1" << endl;

	int tag = 1;
	for (op = obpts.begin(); op != obpts.end(); ++op, ++tag)
	{
		if (*op == 0) continue;
		
		cout << "set label " << tag
			<< " '" << (*op)->get_id() << "'"
			<< " at " << fix9((*op)->crds.x) << ", " << fix9((*op)->crds.y)
			<< " point pt ";
		ObjectPoint::op_type type = (*op)->get_type();
		switch (type)
		{
			case ObjectPoint::ctrl:
				cout << "9";
				break;
				
			case ObjectPoint::inter:
				cout << "7";
				break;
				
			case ObjectPoint::check:
				cout << "11";
				break;
				
			default:
				cout << "1";
				break;
		}
		cout << " offset " << offset_x << ", " << offset_y
			<< endl;
	}

	cout << "plot '-' with points pt -1" << endl;
	//cout << "plot '-' title \"control\" with points pt 9, ";
	//cout << " '-' title \"intersec.\" with points pt 7, ";
	//cout << " '-' title \"check\" with points pt 11" << endl;
	
	// control points
	for (op = obpts.begin(); op != obpts.end(); ++op)
	{
		if (*op == 0) continue;
	
		//ObjectPoint::op_type type = (*op)->type();
		//if (type == ObjectPoint::ctrl)
			cout << fix9((*op)->crds.x) << " " << fix9((*op)->crds.y) << endl;
	}
	//cout << "? ? #dummy line" << endl;
	cout << "e" << endl;
	
// 	
// 	// intersection points
// 	for (op = obpts.begin(); op != obpts.end(); ++op)
// 	{
// 		if (*op == 0) continue;
// 
// 		ObjectPoint::op_type type = (*op)->type();
// 		if (type == ObjectPoint::inter)
// 			cout << fix9((*op)->crds.x) << fix9((*op)->crds.y) << endl;
// 	}
// 	cout << "? ? #dummy line" << endl;
// 	cout << "e" << endl;
// 	
// 	// check points
// 	for (op = obpts.begin(); op != obpts.end(); ++op)
// 	{
// 		if (*op == 0) continue;
// 
// 		ObjectPoint::op_type type = (*op)->type();
// 		if (type == ObjectPoint::check)
// 			cout << fix9((*op)->crds.x) << fix9((*op)->crds.y) << endl;
// 	}
// 	cout << "? ? #dummy line" << endl;
// 	cout << "e" << endl;
	
// 	for ()
// 		{
// 		cout << "set label " << id
// 			<< " at " << op.x << ", " << op.y
// 			<< endl;
// 		}

	
	return 0;
}
