/* $Id: opa2gplt.cpp,v 1.3.2.4 2008/03/05 18:27:15 cvsroot Exp $ */

/* DGAP - General Analytical Positioning
 * Copyright (C) 2007 Dirk Stallmann
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include "form.h"

using namespace std;
using namespace GenLib2;


const char* const program_name = "opa2gplt";
const char* const program_version = "version 0.3.0 (03/2008)";

static void usage(int status=0)
{
	cerr << "Usage: " << program_name << " [options] "
		"opa-file > gp-file\n"
		<< "Arguments:\n"
		<< "  opa-file     OPA filename.\n"
		<< "  gp-file      Gnuplot filename.\n"
		<< "Options:\n"
		<< "  -help            Show this text.\n"
		<< "  -pointsize sz    Set point size. Default: 1.0.\n"
		<< "  -offset x y      Set label offset. Default: 0.5 0.5.\n"
		<< "  -vscale sc       Set vector scale factor. Default: 20.\n"
		<< "  -term <term>     Set terminal type = {png | eps}. Default: png.\n"
		<< "  -version         Show program version.\n";

	exit(status);
}

static void error(const char* s, const char* s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}


// OPA data record

struct OPAData
{
	string	id;
	double	x, y, z;
	double	sx, sy, sz;
	int		type;
	char	cr[3];
	double	vx, vy, vz;
	char	cs[3];
	double	dx, dy, dz;
};

// terminale
enum term_type
{
	TERM_PNG,
	TERM_EPS
};

int main(int argc, char *argv[])
{
	if (argc == 1) usage(1);
	
	term_type	term = TERM_PNG;
	char*	opa_fname = 0;
	float	pointsize = 1.;
	float	offset_x = 0.5;
	float	offset_y = 0.5;
	double	sc = 20.;
	int		i = 0, j = 0;
	
	for (i = 1, j = 0; i < argc; i++)
		{
		if (**(argv + i) == '-')
			{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-offset") == 0)
			{
				offset_x = atof(argv[++i]);
				offset_y = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-pointsize") == 0)
			{
				pointsize = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-vscale") == 0)
			{
				sc = atof(argv[++i]);
			}
			else if (strcmp(argv[i], "-term") == 0)
			{
				i++;
				if (strcmp(argv[i], "png") == 0)
					term = TERM_PNG;
				else if (strcmp(argv[i], "eps") == 0)
					term = TERM_EPS;
				else
					error("Illegal argument:", argv[i]);
			}
			else if (strcmp(argv[i], "-version") == 0)
				{
				cerr << program_name << " " << program_version
					<< " compiled " << __DATE__ << " " << __TIME__
					<< endl;
				exit(0);
				}
			else
				error("Illegal option:", argv[i]);
			}
		else
			{
			switch (j++)
				{
				case 0:
					opa_fname = argv[i];
					break;

				default:
					error("Too many arguments given:", argv[i]);
				}
			}
		}
	
	vector<OPAData*> opa;

	ifstream is(opa_fname);
	if (!is) error("Cannot open input object coords. file:", opa_fname);

	string out_fname(opa_fname);
	size_t pos = out_fname.rfind('.');
	out_fname[pos] = '_';
	
	string id;
	OPAData *rec = 0;
	while (is >> id)
	{
		if (id.compare("stop-dep") == 0) break;

		rec = new(OPAData);
		assert(rec != 0);

		rec->id = id;
		is >> rec->x >> rec->y >> rec->z;
		is >> rec->sx >> rec->sy >> rec->sz;
		is >> rec->type;
		is >> rec->cr;
		is >> rec->vx >> rec->vy >> rec->vz;
		is >> rec->cs;
		is >> rec->dx >> rec->dy >> rec->dz;
		opa.push_back(rec);
	}

	vector<OPAData*>::const_iterator op = opa.begin();

	double xmin = (*op)->x, xmax = (*op)->x;
	double ymin = (*op)->y, ymax = (*op)->y;
	op++;
	
	for (; op != opa.end(); ++op)
	{
		double x = (*op)->x;
		double y = (*op)->y;
		if (x < xmin) xmin = x;
		if (x > xmax) xmax = x;
		if (y < ymin) ymin = y;
		if (y > ymax) ymax = y;
	}
	
	double bx = (xmax - xmin) * 0.1;
	double by = (ymax - ymin) * 0.1;

	switch (term)
	{
		case TERM_PNG:
			cout << "set term png size 1200, 900" << endl;
			out_fname.append(".png");
			break;
					
		case TERM_EPS:
			cout << "set term postscript eps" << endl;
			out_fname.append(".eps");
			break;
		
		default:
			assert(false);
			break;
	}
	cout << "set output '" << out_fname << "'" << endl;
	//cout << "set autoscale xy" << endl;
	//cout << "show pointsize" << endl;
	cout << "set pointsize " << pointsize << endl;
	cout << "set title 'X/Y-plane - " << opa_fname << "'" << endl;
	cout << "set xlabel 'X (m)'" << endl;
	cout << "set ylabel 'Y (m)'" << endl;
	cout << "set format x '%.0f'" << endl;
	cout << "set format y '%.0f'" << endl;
	cout << "set key off" << endl;
	//cout << "set key title \"control\"" << endl;
	cout << "set grid" << endl;
	cout << "set datafile missing '?'" << endl;

	// line styles
	cout << "set style line 1 lt 1 lw 1" << endl;
	//cout << "set style line 2 lt 3 lw 2" << endl;

	// arrow styles
	cout << "set style arrow 1 head filled size screen 0.012,15 ls 1" << endl;
	cout << "set style arrow 2 head size screen 0.006,90 ls 1" << endl;
	//cout << "test" << endl;
	
	cout << "set size ratio -1" << endl;
	//cout << "set clip" << endl;
// 	cout << "unset autoscale" << endl;
 	cout << "set xr [" << (xmin - bx) << ":" << (xmax + bx) << "]" << endl;
 	cout << "set yr [" << (ymin - by) << ":" << (ymax + by) << "]" << endl;
 
	Form fix9(3); fix9.fixed().width(9);

	int tag = 1;
	int num = 0; // counter
	for (op = opa.begin(); op != opa.end(); ++op, ++tag)
	{
		//if (*op == 0) continue;

		cout << "set label " << tag
			<< " '" << (*op)->id << "'"
			<< " at " << fix9((*op)->x) << ", " << fix9((*op)->y)
			<< " point pt ";

		switch ((*op)->type)
		{
			case 1:
				cout << "9";
				break;
				
			case 2:
				cout << "7";
				break;
				
			case 3:
				cout << "11";
				num++;
				break;
				
			default:
				assert(false);
				break;
		}
		cout << " offset " << offset_x << ", " << offset_y << endl;
	}

	cout << "plot '-' with points pt -1, \\" << endl;
	cout << "'-' with vectors arrowstyle 1, \\" << endl;
	cout << "'-' with vectors arrowstyle 2" << endl;
	for (op = opa.begin(); op != opa.end(); ++op)
	{
		cout << fix9((*op)->x) << " " << fix9((*op)->y) << endl;
	}
	//cout << "? ? #dummy line" << endl;
	cout << "e" << endl;
	if (num == 0) return 0;


	for (op = opa.begin(); op != opa.end(); ++op)
	{
		if ((*op)->type == 3)
		{
			cout << fix9((*op)->x) << " " << fix9((*op)->y) << " ";
			cout << fix9((*op)->dx * sc) << " " << fix9((*op)->dy * sc) << endl;
		}
	}

	//cout << "? ? #dummy line" << endl;
	cout << "e" << endl;

	for (op = opa.begin(); op != opa.end(); ++op)
	{
		//if (*op == 0) continue;

		if ((*op)->type == 3)
		{
			cout << fix9((*op)->x) << " " << fix9((*op)->y) << " ";
			cout << fix9(0.) << " " << fix9((*op)->dz * sc) << endl;
		}
	}
	//cout << "? ? #dummy line" << endl;
	cout << "e" << endl;

	return 0;
}
