/**********************************************************************/
//
// FILE:    test_proj.cpp
//          Test projection methods for scanner images.
// CREATED: 14 Apr 04  (ds)
// CHANGES:
// REFS:
//
// (c) 2004 Dirk Stallmann (ds)/ IfP Uni Stuttgart
/**********************************************************************/

#include <cstdlib>
#include <cstring>
#include <fstream>
using namespace std;

#include "form.h"
#include "osform.h"
using namespace GenLib2;

#include "extori.h"
#include "adscam.h"
#include "adsorient.h"
#include "imuopkfix.h"
#include "scanner.h"
using namespace DEP;

// name of this program
#define PROGRAM_NAME "test_proj"

// version of this program
#define PROGRAM_VERSION "0.3.1"

const char* const program_name = PROGRAM_NAME;
const char* const program_version = PROGRAM_VERSION;


// keywords

enum ekey
{
	// unknown
	UNKNOWN,
/*
	// pushbroom camera
	PB_CAMERA,

	// ADS camera
	ADS_CAMERA,

	// orientation data
	ORIENTATION_DATA,

	// ADS orientation
	ADS_ORIENTATION,

	// IMU misaligment
	IMU_MISALIGNMENT,

	// height of rectification plane
	Z_PLANE_HEIGHT,

	// digital elevation model
	DEM,
*/
	// projection by bisection
	BISECTION,

	// projection by hint
	HINT,

	// pixel coordinate generation method
	RANDOM,

	// constant column
	CONST,

	// undefined
	UNDEF,
	
	// end-of-map
	END_OF_MAP
};


// map structure

struct map_t
{
	/// code
	ekey	code;

	/// keyword (short name)
	const char *keyword;
};


// map

map_t map[] =
{
	{ UNKNOWN, "unknown" },
	/*
	{ PB_CAMERA, "pb-camera" },
	{ ADS_CAMERA, "ads-camera" },
	{ ORIENTATION_DATA, "orientation-data" },
	{ ADS_ORIENTATION, "ads-orientation" },
	{ Z_PLANE_HEIGHT, "z-plane-height" },
	{ DEM, "digital-elevation-model" },
	*/
	{ BISECTION, "bisection" },
	{ HINT, "hint" },
	{ RANDOM, "random" },
	{ CONST, "const" },
	{ UNDEF, "undefined" },
	{ END_OF_MAP,		0 }
};


ekey code(const char* keyword)
{
	for (int i = 0; map[i].keyword != 0; i++)
		{
		if (strcmp(keyword, map[i].keyword) == 0)
			return (ekey)map[i].code;
		}

	return (ekey)map[0].code;
}


static void usage(int status=0)
{
	cerr << "Usage: " << program_name << " [options] cfg_file out_file stdin stdout\n"
		<< "Arguments:\n"
		<< "  cfg_file       Configuration file.\n"
		<< "  out_file       Output file with projection results.\n"
		<< "  stdin          Pixel coords. are read from std. input.\n"
		<< "  stdout         Some informations are written to std. output.\n"
		<< "Options:\n"
		<< "  -help          Show this text.\n"
		<< "  -dbg           Debug mode.\n"
		<< "  -verbose       Gives more information.\n"
		<< "  -version       Show program version.\n";
	exit(status);
}


static void error(const char* s, const char* s2 = "")
{
	cerr << program_name << ": " << s << ' ' << s2 << endl;
	exit(1);
}


static void version(ostream& os)
{
	os << program_name << " version " << program_version
		<< " compiled " << __DATE__ << " " << __TIME__ << endl;
	os.flush();
}


// initialize static data members
DEP::LSParams DEP::AdjObject::lspars; // definition of least-squares params.
double OrientationData::shift = 0.;		// global offset for orient. data
OrientationData::IpolMethod OrientationData::ipol = LINEAR;	// interpolation method

int main(int argc, char* argv[])
{
	char*	cfg_name = 0;		// configuration file
	char*	out_name = 0;		// output file
	bool	dbg = false;		// debug flag
	bool	verb = false;		// verbose flag

	if (argc == 1) usage(1);

	int i = 0, j = 0;
	while (argc > ++i)
		{
		if (**(argv + i) == '-')
			{
			if (strcmp(argv[i], "-help") == 0)
				usage(0);
			else if (strcmp(argv[i], "-dbg") == 0)
				dbg = true;
			else if (strcmp(argv[i], "-verbose") == 0)
				verb = true;
			else if (strcmp(argv[i], "-version") == 0)
				{
				version(cerr);
				exit(0);
				}
			else
				error("Illegal option:", argv[i]);
			}
		else
			{
			switch (j++)
				{
				case 0:
					cfg_name = argv[i];
					break;

				case 1:
					out_name = argv[i];
					break;

				default:
					error("Too many arguments given:", argv[i]);
				}
			}
		}

	// open and read config file
	string	cam_fname;		// camera filename
	string	ori_fname;		// orientation data filename
	//string	dem_fname;		// DEM filename
	string	func_str;		// projection string
	//string	gen_str;		// pixel coord. generation method
	double zm = 0.;			// Z-plane height
	int nmax = 0;			// number of random points
	int itmax = 20;			// maximal number of iterations
	double eps = 1e-2;		// row-shift convergence limit (pixel)
	double lim = 1.;		// limit of row/column diff. for projection (pixel)
	
	// read configuration file
	ifstream in(cfg_name);
	if (!in) error("Cannot open configuration file:", cfg_name);

	in >> cam_fname;
	in >> ori_fname;
	in >> zm;
	in >> func_str;
	//in >> gen_str >> nmax;
	in >> nmax;
	in >> itmax >> eps >> lim;

	// projection function
	ekey func = code(func_str.c_str());
	if ((func != BISECTION) && (func != HINT))
		error("Unknown keyword:", func_str.c_str());

	
	cout << "Camera file       : " << cam_fname << endl;
	cout << "Orientation file  : " << ori_fname << endl;
	//cout << "IMU misalign. file   : " << mis_fname << endl;
	//cout << "DEM file             : " << dem_fname << endl;
	cout << form("Z-plane height               :   Zo  = %10.3f m\n", zm);
	cout << form("Projection method            :   %s\n", map[func].keyword);
	//cout << form("Pixel generation method      :   %s\n", gen_str.c_str());
	cout << form("Number of random points      :   %2d\n", nmax);
	cout << form("Maximal number of iterations :   %9d\n", itmax);
	cout << form("Row-shift convergence limit  :   %9.3f pix\n", eps);
	cout << form("Projection limit             :   %9.3f pix\n", lim);
	cout << endl;
	
	// read camera data
	ADSCamera* cam = new ADSCamera;
	{
	ifstream ifs(cam_fname.c_str());

	if (cam->read(ifs) != true)
		error("Cannot read camera file:", cam_fname.c_str());

	if (verb)
		{
		cout << "Camera" << endl;
		cout << "------" << endl;
		cam->print(cout);
		cout << endl;
		}
	}
	
	// orientation
	ADSOrientation* eo = new ADSOrientation;
	{
	if (eo->read(ori_fname.c_str()) != true)
		error("Cannot read orientation data:", ori_fname.c_str());

	in.close();

	// calculate rotation matrices
	//eo.calc_rot_matrix();

	if (verb)
		{
		cout << "Orientation data" << endl;
		cout << "----------------" << endl;
		eo->print(cout);
		cout << endl;
		}
	}

	ofstream ofs(out_name);
	if (!ofs)
		error("Cannot open primary grid point file:", out_name);

	int			nit = 0;			// number of iterations
	int			fail_intersec = 0;	// failure counter
	int			fail_project = 0;	// failure counter
	DPoint2		pix_c;
	DPoint3		obj_c(0., 0., zm);	// z-plane height
	int			num = 0;			// counter
	DPoint2		d_max(0., 0.);
	DPoint2		sum_d(0., 0.);
	DPoint2		sum_dd(0., 0.);
	int			it_max = 0;
	double		sum_it = 0;
		
	 /*
	ekey gen = code(gen_str.c_str());
	if ((gen != RANDOM) && (gen != CONST))
		error("Unknown keyword:", gen_str.c_str());
	*/
	
/*
	switch (gen)
		{
		case RANDOM:
*/
		//cout << RAND_MAX << endl;
		double sc = double(cam->ncols()) / (double(RAND_MAX) + 1.);
		double sr = double(eo->nrows() - 1) / (double(RAND_MAX) + 1.);
		
		//for (i = 0; i < eo->nrows(); i++)
		for (i = 0; i < nmax; i++)
			{
			// 1. generate random numbers
			double c = double(rand()) * sc;
			double r = double(rand()) * sr;
			//double c = nmax;
			//double r = i;
			DPoint2 pix_c0(c, r); // pixel coords.
			if (dbg) cout << "pix_c0 = " << pix_c0 << endl;

			// 2. calculate "true" relation between
			// pixel- and object coordinates by z-plane intersection
			bool rtn = intersection(*cam, *eo, pix_c0, obj_c);

			if (rtn != true)
				{
				fail_intersec++;
				cout << "*** Intersection failed: " << pix_c0 << endl;
				continue;
				}

			// 3. back projection
			switch (func)
				{
				case BISECTION:
					rtn = projection_bisection(*cam, *eo, obj_c, pix_c, nit,
						dbg);
					break;

				case HINT:
					// do back projection by search method
					rtn = projection_hint(*cam, *eo, obj_c, pix_c, nit,
						-1, eps, itmax, dbg);
					break;

				case UNKNOWN:
				default:
					assert(false);
					break;
				}

			DPoint2 d = pix_c0 - pix_c;

			char state = 'f';
			if (rtn != true)
			{
			fail_project++;
			cout << "*** Projection failed: point " << i
				<< " at " << pix_c0 << " num. of it. " << nit << endl;
			cout << "d = " << d << endl;
			}
		else
			{
			// calculate statistic only from valid values

			// test limit
			if (fabs(d.x) > lim || fabs(d.y) > lim)
				state = 'e';
			else
				state = 'o';

			if (d.x > d_max.x) d_max.x = d.x;
			if (d.y > d_max.y) d_max.y = d.y;
			sum_d += d;
			sum_dd += d * d;
			if (nit > it_max) it_max = nit;
			sum_it += nit;
			num++;
			}

		if (fabs(d.x) > lim || fabs(d.y) > lim)
			{
			// write: c0, r0, X, Y, Z, state, c, r, nit, dc, dr
			ofs << form("%10.3f %10.3f %12.3f %12.3f %12.3f ",
				pix_c0.x, pix_c0.y, obj_c.x, obj_c.y, obj_c.z);
			ofs << form("%c %10.3f %10.3f %2d %10.3f %10.3f\n",
				state, pix_c.x, pix_c.y, nit, d.x, d.y);
			}
		}
	cout << "Test finished\n";
	cout << endl;

	cout << "Number of ...\n";
	cout << "  - successfull projected points :   " << num << endl;
	cout << "  - failed by intersection       :   " << fail_intersec
		<< endl;
	cout << "  - failed by projection         :   " << fail_project
		<< endl;
	cout << endl;

	Form dec(0); dec.fixed().width(10);
	Form fix(4); fix.fixed().width(10);
	cout << "                                 dc/pix       dr/pix\n\n";
	cout << "Number of differences    :   "
		<< dec(num) << "   "
		<< dec(num) << endl;
	cout << "Maximum of differences   :   "
		<< fix(d_max.x) << "   "
		<< fix(d_max.y) << endl;
	cout << "Average of differences   :   "
		<< fix(sum_d.x / num) << "   "
		<< fix(sum_d.y / num) << endl;
	cout << "RMS of differences       :   "
		<< fix(sqrt(sum_dd.x / num)) << "   "
		<< fix(sqrt(sum_dd.y / num)) << endl;
	cout << endl;

	cout << "Maximum of iterations    :   "
		<< dec(it_max) << endl;
	cout << "Average of iterations    :   "
		<< fix(sum_it / num) << endl;
	cout << endl;

	return 0;
}
